#include "seg.h"
#include "seg_global.h"

static int *yy0, *x0; 
static int DebugFlag=0 ;



/*********************************************************/
/*     MAP segmentation of the image with Pyramid model	  */
/*     The transition prob is modeled as the joint distribution*/
/*     P(son | father, majority of neighbors of father )*/
/***********************************************************/
void MAP_Pyramid(struct BigPyramid *Pyramid, struct ModelPar *Par, char *outfn,
    int X0, int Y0) 
{
  struct CP_Trees *CPT ;
  int D, nClasses, nRow,nCol;
  int i,j,k,m ;
  unsigned char **Segment, ***SegPt ;
  char fname[80] ;
  static FILE *outfd ;
  double ****LProb, ****ll ;
  static int Counts = 0 ;    /* The number of times this program is executed */


  if(vlevel >=3) printf("\tbegining MAP_Pyramid() ......\n") ;

  Counts ++ ;    /* for multi-iteration MAP segmentation */
  CPT = &(Par->CPT) ;
  nClasses = CPT->info.nclasses ;
  D = (Pyramid->level) ;
  SegPt = (Pyramid->_Segment) ;
  LProb = (Pyramid->LProb) ;
  ll = (Pyramid->LLikeH) ;

  if( X0>=0 && X0<(Pyramid->nRow)[0] && Y0>=0 && Y0<(Pyramid->nCol)[0] ) {
 
      DebugFlag = 1 ;
   
      if(Counts==1) {        /* initialization */
          sprintf(fname,"%s%d.debug",outfn,Counts) ;
          outfd = fopen (fname, "w");  
          if (outfd == NULL) FileOpenErr(fname) ;
      }
      
      x0 = (int*) get_spc(D, sizeof(int)) ;
      yy0 = (int*) get_spc(D, sizeof(int)) ;
      
      x0[0]=X0;  yy0[0]=Y0; 
      for(i=1; i<D; i++) {
          x0[i] = (int) x0[i-1] / 2 ;
          yy0[i] = (int) yy0[i-1] / 2 ;
      }
      
      
      /* output the LProb of the coarest level */
      nRow = (Pyramid->nRow)[D-1] ;
      nCol = (Pyramid->nCol)[D-1] ;
      i=(x0[D-1]);  
      j=(yy0[D-1]); 
      if((i<nRow) && (j<nCol)) {
          if(outfd != NULL) fprintf(outfd, "%3d, %3d,   ", i,j) ;
          fprintf(stdout, "%3d, %3d,   ", i,j) ;
          for(m=0; m<nClasses; m++) {
              if(outfd!=NULL) fprintf(outfd, "%lf  ", LProb[D-1][i][j][m]) ;
              fprintf(stdout, "%lf  ", LProb[D-1][i][j][m]) ;
          }
          if(outfd!=NULL) fprintf(outfd, "\n") ;
          fprintf(stdout, "\n") ;
         
          if(outfd!=NULL) fprintf(outfd, "     final ") ;
          fprintf(stdout, "     final ") ;
          for(m=0; m<nClasses; m++) {
              if(outfd!=NULL) fprintf(outfd, "%lf  ", ll[D-1][i][j][m]) ;
              fprintf(stdout, "%lf  ", ll[D-1][i][j][m]) ;
          }
          if(outfd!=NULL) fprintf(outfd, "\n\n") ;
          fprintf(stdout, "\n\n") ;
      }
      
  } else {    /* no debug */
      
      outfd = NULL ;
  }
      
      

  /* segment the coarsest resolution */
  SegCoarsestResolution(Pyramid, nClasses) ;
  
  /* This part is only appliable to Pyramid model */
  for(k=D-2; k>=0; k--) {

    nRow = (Pyramid->nRow)[k] ;
    nCol = (Pyramid->nCol)[k] ;
    Segment = (unsigned char **) get_img(nCol,nRow,sizeof(unsigned char)) ;  
    SegPt[k] = Segment ;
    
    if(DebugFlag) {
        /* output the LProb of this level */
        if(outfd!=NULL) fprintf(outfd, "Level %d,  DataWeight %lf \n\n", k, 
            (Par->DataWeight)[k]) ;
        fprintf(stdout, "Level %d,  DataWeight %lf \n\n", k, (Par->DataWeight)[k]) ;
        i=(x0[k]); 
        j=(yy0[k]); 
        if((i<nRow) && (j<nCol)) {
          if(outfd!=NULL) fprintf(outfd, "%3d, %3d,   ", i,j) ;
          fprintf(stdout, "%3d, %3d,   ", i,j) ;
          for(m=0; m<nClasses; m++) {
              if(outfd!=NULL) fprintf(outfd, "%lf  ", LProb[k][i][j][m]) ;
              fprintf(stdout, "%lf  ", LProb[k][i][j][m]) ;
          }
          if(outfd!=NULL) fprintf(outfd, "\n") ;
          fprintf(stdout, "\n") ;
          if(outfd!=NULL) fprintf(outfd, "     LLike ") ;
          fprintf(stdout, "     LLike ") ;
          for(m=0; m<nClasses; m++) {
              if(outfd!=NULL) fprintf(outfd, "%lf  ", ll[k][i][j][m]) ;
               fprintf(stdout, "%lf  ", ll[k][i][j][m]) ;
           }
         if(outfd!=NULL) fprintf(outfd, "\n") ;
         fprintf(stdout, "\n") ;
        }
    }


    SegPyramidModel(Pyramid, Par, k, outfd);
  
    if(vlevel >=20) printf("\t\tMAP_Pyramid() level %d......\n",k) ;
  }
          
  if(vlevel >=3) printf("\tfinish MAP_Pyramid() ......\n") ;
  

  /* save the segmentation results for the finest resolution */    
  if(outfn!=NULL) OutputSeg(Pyramid, outfn, "P") ;    /* u_pyramid.c */

  
  if(DebugFlag) {
      free((void*) x0);   
      free((void*) yy0);  
      
      if(outfd!=NULL) fprintf(outfd,"\n\n") ;
      fprintf(stdout,"\n\n") ;
  }
}
 
 
 
 
  
/************************************************************************/
/*  The Pyramid Model to calculate ll[i][j]+log(Px_n|x_n+1) */
/*************************************************************************/
void SegPyramidModel(
  struct BigPyramid *Pyramid,
  struct ModelPar *Par, 
  int K,  /*  K the current level */
  FILE *outfd
)
{

  struct CP_Trees *CPT ;
  double ****ll ;
  double ***LProb ;
  int nRow,nCol, nNeighb;  
  unsigned char **SegFather, **Seg ;
  struct TreeNode **rootsDT ;
  int nClasses;
  int nRowFa,nColFa ; /* index for a father */
  int i,j,m, i1,j1, i2,j2, index ;
  static int *localSeg ;
  static int flag=0 ;
  double *Prob ;
  
  CPT = &(Par->CPT) ;
  nClasses = (CPT->info).nclasses ;
  nNeighb = (CPT->info).dimen ;
  rootsDT = ((CPT->trees)[K+1]).roots ;
  SegFather = (Pyramid->_Segment)[K+1] ;
  Seg = (Pyramid->_Segment)[K] ;
  LProb = (Pyramid->LProb)[K] ;
  ll = (Pyramid->LLikeH) ;
  nRow = (Pyramid->nRow)[K] ;
  nCol = (Pyramid->nCol)[K] ;
  nRowFa = (Pyramid->nRow)[K+1] ;
  nColFa = (Pyramid->nCol)[K+1] ;


  if(flag==0) {
    localSeg=(int*) get_spc(nNeighb,sizeof(int)) ;
      flag = 1;
  }


  for(i=0; i<nRowFa; i++)
  for(j=0; j<nColFa; j++) {
       
      CoarseNeighbors(CPT, SegFather, nRowFa,nColFa, i,j, localSeg) ;      
      
      if(DebugFlag) {
        if((i==x0[K+1])&&(j==yy0[K+1])) {
          if(outfd!=NULL) fprintf(outfd, "     father ") ;
          fprintf(stdout, "     father ") ;
          for(i1=0; i1<nNeighb; i1++) {
              if(outfd!=NULL) fprintf(outfd, "%d=%d  ", i1,localSeg[i1]) ;
              fprintf(stdout, "%d=%d  ", i1,localSeg[i1]) ;
          }
          if(outfd!=NULL) fprintf(outfd, "\n") ;
          fprintf(stdout, "\n") ;
        }
      }
        

      
      index=0 ;
      for(i1=0, i2=2*i; i1<2; i1++, i2++) 
      for(j1=0, j2=2*j; j1<2; j1++, j2++) {
      
        /*  i2=2*i+i1 ;    j2=2*j+j1 ;*/
          if((i2<nRow)&&(j2<nCol)) {
          
            Prob = ContexProb(localSeg, rootsDT[index], nClasses, nNeighb) ;    

             if(DebugFlag) {
                    if((i2==x0[K])&&(j2==yy0[K])) {


                    if(outfd!=NULL) fprintf(outfd, "     Cntx  ") ;
                    fprintf(stdout, "     Cntx  ") ;
                    for(m=0; m<nClasses; m++) {
                        if(outfd!=NULL) fprintf(outfd, "%d=%lf  ", m,
                            LogCalculate((double)(Prob[m]),"SegPyramidModel")) ;
                        fprintf(stdout, "%d=%lf  ", m,
                            LogCalculate((double)(Prob[m]),"SegPyramidModel")) ;
                    }
                    if(outfd!=NULL) fprintf(outfd, "\n") ;
                    fprintf(stdout, "\n") ;
                    if(outfd!=NULL) fprintf(outfd, "     Cntx  ") ;
                    fprintf(stdout, "     Cntx  ") ;
                    for(m=0; m<nClasses; m++) {
                        if(outfd!=NULL) fprintf(outfd, "%d=%lf  ", m, Prob[m]) ;
                        fprintf(stdout, "%d=%lf  ", m, Prob[m]) ;
                    }
                    if(outfd!=NULL) fprintf(outfd, "\n") ;
                    fprintf(stdout, "\n") ;
                  }
            }
            
             if(DebugFlag) { 
                 if((i2==x0[K])&&(j2==yy0[K])) { 

                  if(outfd!=NULL) fprintf(outfd, "     final ") ;
                  fprintf(stdout, "     final ") ;
                 
                  Seg[i2][j2] = SegMaxProb(Par,ll[K][i2][j2], Prob, outfd, K)  ;
                  Seg[i2][j2] = SegMaxProb(Par,ll[K][i2][j2], Prob, stdout, K)  ;
                  if(outfd!=NULL) fprintf(outfd, "\n\n") ;
                  fprintf(stdout, "\n\n") ;
                 } else {
                     Seg[i2][j2] = SegMaxProb(Par,ll[K][i2][j2],Prob,(FILE*)NULL,K);
  
                 }
              } else {
                  Seg[i2][j2] = SegMaxProb(Par,ll[K][i2][j2],Prob,(FILE*)NULL,K);
              } 
          }
          
          index ++ ;
      }
  }
}


/********************************************************************/
/*    Maximization at a pixel */
/*    ll={ll[i][j]}[class] is the loglikelihood at pixel i,j*/
/*    TranProb =  transition prob at pixel i,j.   */
    
/*    Confident Segmentation :  Prob(Max_ll) > .6*/
/*    OK Segmentation        :  Prob(Max) - Prob(Second) > .1*/
/*    Reject                :  O.W.*/
    
/**********************************************************************/
int SegMaxProb(
  struct ModelPar *Par,
  double *ll,
  double *TranProb,
  FILE *outfd,
  int K    /* level */
)
{
  int m, nClasses ;
  int max_index ;
  double tmp, max_ll ;
  
  nClasses = (Par->CPT).info.nclasses ;
  
  for(m=0; m<nClasses; m++) {  /* fine resolution */
          
      tmp= ll[m] + LogCalculate((double)(TranProb[m]),"SegMaxProb") ;

/*    When using a weight term
      tmp= ll[m] + (Par->DataWeight)[K]*
          LogCalculate((double)(TranProb[m]),"SegMaxProb") ;
*/

      if(outfd!=NULL) fprintf(outfd, "%lf  ", tmp) ;
      
      
      /*  Find Max(ll) */
      if(m==0) { 
            max_ll = tmp;  max_index = m; 
      }
      
      if(max_ll<tmp) { 
            max_ll = tmp; max_index = m; 
        } 
  }
  
  return(max_index) ;
}

  

/**********************************************************/
/*     MAP segmentation of the image with Quadtree model*/
/***********************************************************/
void MAP_QuadTree(struct BigPyramid *Pyramid, struct ModelPar *Par, 
    char *outfn, int X0, int Y0) 
{
  FILE *outfd ;
  struct Par_QTR *QTr_par ;
  int D, nRow,nCol, nClasses;
  int i,j, m,i1,j1 ;
  int fatherR,fatherC ;    /* index for a father */
  unsigned char ***SegPt ;
  unsigned char **Segment, **SegFather ;
  double **alpha ;
  double max_ll ;
  int max_index ;
  double ***ll;     /* log likehood at each pixel of the multispectral images*/
  double ****LProb ;
  double tmp ;
  int segFather ;
  char fname[80] ;
  static int Counts=0 ;
  
  
  Counts ++ ;
  if(vlevel>3) printf("\tsegmenting the image by MAP ......\n") ;

  QTr_par = &(Par->QTr_par) ;
  nClasses = QTr_par->nclasses ;
  D = Pyramid->level ;
  SegPt = Pyramid->_Segment ;
  ll = (Pyramid->LLikeH)[D-1] ;
  LProb = (Pyramid->LProb) ;
  
  
  if( X0>=0 && X0<(Pyramid->nRow)[0] && Y0>=0 && Y0<(Pyramid->nCol)[0] ) {
 
      DebugFlag = 1 ;
   
      if(Counts==1) {        /* initialization */
          sprintf(fname,"%s%d.Q.debug",outfn,Counts) ;
          outfd = fopen (fname, "w");  
          if (outfd == NULL) FileOpenErr(fname) ;
      }
      
      x0 = (int*) get_spc(D, sizeof(int)) ;
      yy0 = (int*) get_spc(D, sizeof(int)) ;
      
      x0[0]=X0;  yy0[0]=Y0; 
      for(i=1; i<D; i++) {
          x0[i] = (int) x0[i-1] / 2 ;
          yy0[i] = (int) yy0[i-1] / 2 ;
      }
      
      
      /* output the LProb of the coarest level */
      nRow = (Pyramid->nRow)[D-1] ;
      nCol = (Pyramid->nCol)[D-1] ;
      i=(x0[D-1]);    j=(yy0[D-1]); 
      if( (i<nRow) && (j<nCol) && (outfd != NULL) ) {
          if(outfd != NULL) fprintf(outfd, "%3d, %3d,   ", i,j) ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%lf  ", LProb[D-1][i][j][m]) ;
          }
          fprintf(outfd, "\n") ;
         
          fprintf(outfd, "     final ") ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%lf  ", ll[i][j][m]) ;
          }
          if(outfd!=NULL) fprintf(outfd, "\n\n") ;
      }
      
  } else {    /* no debug */
      
      outfd = NULL ;
  }
  
  
  
  /* segment the coarsest resolution */
  SegCoarsestResolution(Pyramid, nClasses) ;
  
  /* This part is only appliable to quadtree model */
  for(i=D-2; i>=0; i--) {
      nRow = (Pyramid->nRow)[i] ;
      nCol = (Pyramid->nCol)[i] ;
      Segment = (unsigned char **) get_img(nCol,nRow,sizeof(unsigned char)) ;
      SegPt[i] = Segment ;
      SegFather = SegPt[i+1] ;
      alpha = (Par->QTr_par.Alpha0)[i+1] ;
      ll = (Pyramid->LLikeH)[i] ;

      i1=(x0[i]);   j1=(yy0[i]); 
      if(DebugFlag && (i1<nRow) && (j1<nCol) && (outfd!=NULL) ) {
          /* output the LProb of this level */
          fprintf(outfd, "Level %d \n\n", i) ;
          fprintf(outfd, "%3d, %3d,   ", i1,j1) ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%lf  ", LProb[i][i1][j1][m]) ;
          }
          fprintf(outfd, "\n\tLLike ") ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%lf  ", ll[i1][j1][m]) ;
          }
          
          segFather = SegFather[i1/2][j1/2] ;
          fprintf(outfd, "\n\tparent :  %d\n", segFather) ;

          fprintf(outfd, "\tCntx  ") ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%d=%lf  ", m,
                  LogCalculate((double)(alpha[m][segFather]),"MAP_QuadTr")) ;
          }
          fprintf(outfd, "\n") ;

          fprintf(outfd, "\tCntx  ") ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%d=%lf  ",  m, alpha[m][segFather]) ;
          }
          fprintf(outfd, "\n") ;

          fprintf(outfd, "\tFinal  ") ;
          for(m=0; m<nClasses; m++) {
              fprintf(outfd, "%d=%lf  ", m, ll[i1][j1][m] + 
                 LogCalculate((double)(alpha[m][segFather]),"MAP_QuadTr")) ;
          }
          fprintf(outfd, "\n") ;
      }


      for(i1=0; i1<nRow; i1++)
      for(j1=0; j1<nCol; j1++) {
          fatherR = i1/2 ;
          fatherC = j1/2 ;
          
          segFather = SegFather[fatherR][fatherC] ;

          for(m=0; m<nClasses; m++) {
             tmp = ll[i1][j1][m] + 
                 LogCalculate((double)(alpha[m][segFather]),"MAP_QuadTr") ;
             
             if(m==0) {
                 max_ll = tmp ;
                 max_index = m ;
                 
             } else if(max_ll<tmp) {
                 max_ll = tmp; 
                 max_index = m ;
             }
          }
          Segment[i1][j1] = max_index ;
      }
  }
  
  if(outfd != NULL) fclose(outfd) ;
  
  /* The part above is only appliable to quadtree model */
  if(vlevel >=3) printf("\tfinish MAP_QuadTree()\n") ;
  
  /* save the segmentation results for the finest resolution */
  if(outfn!=NULL) OutputSeg(Pyramid, outfn, "Q") ;    /* u_pyramid.c */
}
  


/******************************************************************************/
/******************************************************************************/
void SegCoarsestResolution(struct BigPyramid* Pyramid, int nClasses)
{
  int  nRow,nCol ;
  int i,j, m ;
  unsigned char **Segment ;
  double max_ll ;
  int max_index ;
  double ***ll;      /* log likehood at each pixel of the multispectral images*/
  int D ;

  /* segment the coarsest resolution */
  D = Pyramid->level ;
  nRow = (Pyramid->nRow)[D-1] ;
  nCol = (Pyramid->nCol)[D-1] ;
  Segment = (unsigned char **) get_img(nCol,nRow,sizeof(unsigned char)) ;
  (Pyramid->_Segment)[D-1] = Segment ;
  ll = (Pyramid->LLikeH)[D-1] ;
  

  for(i=0; i<nRow; i++) {
	  for(j=0; j<nCol; j++) {
    	  max_ll = ll[i][j][0] ;
    	  max_index = 0 ;
    	  for(m=1; m<nClasses; m++) {
        	  if(max_ll<ll[i][j][m]) {
            	  max_ll = ll[i][j][m]; 
            	  max_index = m ;
        	  }
    	  }
    	  Segment[i][j] = max_index ;

	  }
  }
}  


/******************************************************************************/
/******************************************************************************/
double LogCalculate( double p, char *Str) 
{
    if(p<0) {
        fprintf(stderr,"LogCalculate error, %lf, %s\n", p, Str) ;
        exit(-1) ;
    } else if(p<MIN_LOG) {
        return((double) log(MIN_LOG)) ;    /* impossible event */
    } else {
        return(log(p)) ;
    }
}


