#include "seg_global.h"
#include "seg.h"


/*****************************************************************************/
/* find the coarse neighbors*/ 
/*****************************************************************************/
void CoarseNeighbors(
   struct CP_Trees *CPT,
   unsigned char **SegFather,
   int nrowF, int ncolF,
   int ii,  int jj,      /* current pixel indices */
   int *Neighbors
)
{
   int i1,j1,i2,j2, seg, index, Margin, nClasses ;
      
   Margin = (CPT->info).Margin ;
   nClasses = (CPT->info).nclasses ;
   
   /* neighbors of father and father */ 
   index = 0 ;   
   for(i1=-Margin; i1<=Margin; i1++) 
   for(j1=-Margin; j1<=Margin; j1++) {
              
       /* find neighbors of father */
       i2= ii +i1 ;  j2 = jj +j1 ;    /* k the index for localSeg */

/*  this is the new version, not using default class */
       if(i2<0) i2=0;    if(i2>=nrowF) i2=nrowF-1;
       if(j2<0) j2=0;    if(j2>=ncolF) j2=ncolF-1;

if(i2>=0 && i2<nrowF && j2>=0 && j2<ncolF) {

       seg = (int) SegFather[i2][j2] ;
       if((seg<0)||(seg>=nClasses)) {
           fprintf(stderr,"Wrong segment in CoarseNeighborsInd(): \n") ;
           fprintf(stderr,"seg %d, i%d, j%d\n",seg,i2,j2) ;
           exit(-1) ;
       }

} 
       Neighbors[index] = seg ;  
       index ++ ;
   }
   
   if(index != (CPT->info).dimen) {
          fprintf(stderr, "Too many or too few neighbors !\n") ;
          exit(-1) ;
   }
}



  


/*****************************************************************************/
/*****************************************************************************/

static int localPartition(int *XX, struct TreeNode *node, 
    int nClasses, int nNeighb)
{
   int i, j ;
   double sum ;
   
   sum = node->T0 ;
   for(i=0; i<nNeighb; i++) {
       j=XX[i] ;
       if(j<(nClasses-1)) {
           sum += (node->T)[i][j] ;
       }
   }
   
   if(sum>=0) return(1) ;
   else return(0) ;
}


/*****************************************************************************/
/* Use Decision Tree to find out the conditional probability*/ 
/* (X | coarse neighbors) which is the histogram or the conditional*/
/* probability density associated with the leaf of the decision*/
/* tree.*/
/*****************************************************************************/
double* ContexProb(
  int *X,    /* known classification of coarse neighbors */
  struct TreeNode *DecTrNode, /* DecTreeNode -- decision tree structure */
  int nClasses,
  int nNeighb
)
{
  int True ;
  struct TreeNode *child ;    /* DecTreeNode -- decision tree structure */
  
  
  if(DecTrNode != NULL) {
      True = localPartition(X,DecTrNode, nClasses, nNeighb) ;
      
      if(True) {    /* >=0 */
          if(vlevel>100) printf("L(X)>=0 ") ;
          child = DecTrNode->left ;
          if(child==NULL) return((DecTrNode->Hist)[1]) ;
      } else {
          if(vlevel>100) printf("L(X)<0  ") ;
          child = DecTrNode->right ;
          if(child==NULL) return((DecTrNode->Hist)[0]) ;
      }
          
      if(child != NULL) {
          return(ContexProb(X, child, nClasses, nNeighb)) ;
      }
  
  } else {
      fprintf(stderr,"Null Decision Tree pointer! ContexProb\n") ;
      exit(-1) ;
  }
}
