#include "seg.h"
#include "seg_global.h"

static void decimate(
 /* decimate statistics ll1 to form ll2 */
 LIKELIHOOD ***ll1,    /* log likelihood ll1[i][j][class], at fine resolution*/
 int nRow1, int nCol1, /* nRow and nCol of ll1, fine resolution */
 int M,                /* number of classes */
 LIKELIHOOD ***ll2,    /* loglikelihood ll2[i][j][class], at coarse resolution*/
 int nRow2, int nCol2, /* nRow and nCol of ll2, coarse resolution */
 double **alpha        /* transition parameter */
) ;

static void up_ll(
	LIKELIHOOD  *pt1, /* array of log likelihood values, pt1[class] */
	int     M,
	double  **alpha,	/* alpha[class][class] */
	LIKELIHOOD  *pt2 /* array of log likelihood values, pt2[class] */
);




/**********************************************************/
/*	Up propagate from the finest to the coarsest */
/*	to calculate the loglikelihood of each node. */
/***********************************************************/
void upPropagate(struct BigPyramid* Pyramid, struct ModelPar *Par)
{
  struct Par_GMM *GMM_par ;
  struct Par_QTR *QTr_par ;
  int D, nClasses, nRow,nCol;
  int *nRowPt, *nColPt ;
  int i,m,i1,j1 ;
  double ***ll;  /* log likehood at each pixel of the multispectral images*/
  double ****LLike ; /* LLikeH[level][j][k][class], log likelihood*/
  double ****LProb ; /* LProb[level][j][k][class], log likelihood*/
  double ***alpha ;
  
  if(vlevel>3) printf("\tup propagating ......\n") ;

  GMM_par = &(Par->GMM_par) ;
  QTr_par = &(Par->QTr_par) ;
  D = Pyramid->level ;
  nClasses = QTr_par->nclasses ;
  nRowPt = Pyramid->nRow ;
  nColPt = Pyramid->nCol ;
  alpha = QTr_par->Alpha0 ;
  
  LLike = Pyramid->LLikeH ;
  LProb = Pyramid->LProb ;

  /* initial */
  i=0 ;
  nRow = nRowPt[i] ;
  nCol = nColPt[i] ;

  /*  The following change may be risky
  ll = (double ***) multialloc(sizeof(double),3,nRow,nCol,nClasses) ;
  LLike[i] = ll ;
  
  for(i1=0; i1<nRow; i1++)
  for(j1=0; j1<nCol; j1++)
  for(m=0; m<nClasses; m++) {
  	  ll[i1][j1][m] = LProb[i][i1][j1][m] ;
  }
  */
  /* since at the fineset resolution, ll is same as LProb */
  LLike[0] = LProb[0] ;

    
  /* up-propagate */
  for(i=1; i<D; i++) {
    nRow = nRowPt[i] ;
    nCol = nColPt[i] ;
    ll = (double ***) multialloc(sizeof(double),3,nRow,nCol,nClasses) ;
    LLike[i] = ll ;
    
    decimate(LLike[i-1],nRowPt[i-1],nColPt[i-1],nClasses,ll,nRow,nCol,alpha[i]);
  	
    /* for levels with input images */
    if(GMM_par->existSig[i]) {
  		for(i1=0; i1<nRow; i1++)
  		for(j1=0; j1<nCol; j1++)
  		for(m=0; m<nClasses; m++) {
  			ll[i1][j1][m] += LProb[i][i1][j1][m] ;
  		}
    }
  }
  
}



static void decimate(
 /* decimate statistics ll1 to form ll2 */
 LIKELIHOOD ***ll1,    /* log likelihood ll1[i][j][class], at fine resolution*/
 int nRow1, int nCol1, /* nRow and nCol of ll1, fine resolution */
 int M,                /* number of classes */
 LIKELIHOOD ***ll2,    /* loglikelihood ll2[i][j][class], at coarse resolution*/
 int nRow2, int nCol2, /* nRow and nCol of ll2, coarse resolution */
 double **alpha        /* transition parameter */
)
{
    int nrow, ncol ;	/* nrow, ncol of coarse resolution */
    int  wflag,hflag;               /* flags indicate odd number of pixels */
    int	 i,j,m;
    int	 i1,j1;
    LIKELIHOOD *node;               /* coarse resolution point */
    LIKELIHOOD *pt1,*pt2,*pt3,*pt4; /* fine resolution neighbors */

    if(vlevel>4) printf("\t\tbegin decimating ......") ;
    nrow = nRow1 / 2 ;
    ncol = nCol1 / 2 ;
    wflag = nCol1 % 2 ;
    hflag = nRow1 % 2 ;
    if(((nrow+hflag) != nRow2)||((ncol+wflag) != nCol2)) {
    	fprintf(stderr, "Dimensions conflict in decimat().\n") ; 
    	exit(-1) ;
    }

    for(i=0,i1=0; i<nrow; i++,i1+=2)
    for(j=0,j1=0; j<ncol; j++,j1+=2) {
	
      /* i1=2*i;  j1=2*j; */
      pt1= ll1[i1][j1];
      pt2= ll1[i1][j1+1];
      pt3= ll1[i1+1][j1];
      pt4= ll1[i1+1][j1+1];

      node= ll2[i][j];
      for(m=0; m<M; m++) node[m] = 0.0;

      up_ll(pt1,M,alpha,node);
      up_ll(pt2,M,alpha,node);
      up_ll(pt3,M,alpha,node);
      up_ll(pt4,M,alpha,node);
    }

    if(wflag) {
      for(i=0; i<nrow; i++) {
        node = ll2[i][nCol2-1];
        for(m=0; m<M; m++) node[m] = 0.0;

        pt1 = ll1[2*i][nCol1-1];
        pt2 = ll1[2*i+1][nCol1-1];
        up_ll(pt1,M,alpha,node);
        up_ll(pt2,M,alpha,node);
      }
    }

    if(hflag) {
      for(j=0; j<ncol; j++) {
        node = ll2[nRow2-1][j];
        for(m=0; m<M; m++) node[m] = 0.0;

        pt1 = ll1[nRow1-1][2*j];
        pt2 = ll1[nRow1-1][2*j+1];
        up_ll(pt1,M,alpha,node);
        up_ll(pt2,M,alpha,node);
      }
    }

    if(hflag&&wflag) {
      node = ll2[nRow2-1][nCol2-1];
      for(m=0; m<M; m++) node[m] = 0.0;

      pt1 = ll1[nRow1-1][nCol1-1];
      up_ll(pt1,M,alpha,node);
    }
    if(vlevel>4) printf("        Done.\n") ;
}


static void up_ll(
	LIKELIHOOD  *pt1, /* array of log likelihood values, pt1[class] */
	int     M,
	double  **alpha,	/* alpha[class][class] */
	LIKELIHOOD  *pt2 /* array of log likelihood values, pt2[class] */
)
{
    static int mc, mf;
    static double sum,max,cprob[256];
    double logcheck ;
    
    max = pt1[0];
    for(mf=1; mf<M; mf++) 
        if(pt1[mf]>max) max= pt1[mf];

    for(mf=0; mf<M; mf++) {
        cprob[mf] = exp_lookup( pt1[mf] - max );
    }
  
    for(mc=0; mc<M; mc++) { /* classes at coarse resolution */
        sum = 0 ;
        for(mf=0; mf<M; mf++) {  /* classes at fine resolution */
          sum +=  alpha[mf][mc]*cprob[mf] ;
        }  
        logcheck = sum ;
        if(logcheck <= 0) {
          	fprintf(stderr,"Error in up_ll(pt1,M,alpha,pt2), %lf\n", logcheck) ;
        	for(mf=0; mf<M; mf++) {  /* classes at fine resolution */
          	fprintf(stderr,"%d %d %lf %lf \n",mf,mc,max,pt1[mf]) ;
        	}  
          	exit(1) ;
        } else {    
	        pt2[mc] += log( logcheck ) + max;
	}
    }
}
 
