/**********************************************************************/
/*  This version uses "double" to represent the feature vector, */
/*  instead of using "unsigned char".  Hopefully, this prevents*/
/*  clipping and improves the quality*/
/*					*/
/*  March 29, 1998*/
/**********************************************************************/

#include "seg.h"
#include <stdio.h>

extern int vlevel ; 
/* inserted following line -GF*/
void error(char *name);
/**********************************************************************/
/* wavelet transformation */
/*	x11  x12  -> (x11+x12-x21-x22)/4  (x11-x12+x21-x22)/4*/
/*	x21  x22  -> (x11-x12-x21+x22)/4  (x11+x12+x21+x22)/4*/
/***********************************************************************/
static void WaveLetTrans(
	double **img,
	int wd,  
	int ht, 
	double ***OutImg,
	int Rt,
	int *wdPt,
	int *htPt
)
{
  int i,j ;
  int wd1,ht1 ;
  int wd_odd, ht_odd ;
  int Dimen ;	/* dimension of the output multispectrum image(s) */
  double t1,t2,t3,t4 ;
  
  Dimen = 4 ;	/* dimension of the output multispectrum image(s) */
  if(vlevel >1) printf("\twavelet transformation ... \n") ; 
  
  	
  /* calculate the wd and ht of the output image(s) */
  if(Rt != 2) {
      fprintf(stderr, "2x2 subsampling is needed, not Rt=%d\n", Rt) ;
      exit(-10) ;
  }
  wd_odd = wd % Rt ;
  ht_odd = ht % Rt ;
  wd1 = (int) wd / Rt ;
  ht1 = (int) ht / Rt ;
  if(wd_odd>0) wd1 ++ ;
  if(ht_odd>0) ht1 ++ ;
  for(i=0; i<Dimen; i++)
    OutImg[i]=(double **) get_img(wd1,ht1,sizeof(double));
    	
   	
  /* waveLet transfermation of the main body */
  for(i=0; i< ht-1; i+=Rt)
  for(j=0; j< wd-1; j+=Rt) {
    t1= img[i][j]+img[i][j+1]-img[i+1][j]-img[i+1][j+1] ;
    t2= img[i][j]-img[i][j+1]+img[i+1][j]-img[i+1][j+1] ;
    t3= img[i][j]-img[i][j+1]-img[i+1][j]+img[i+1][j+1] ;
    t4= img[i][j]+img[i][j+1]+img[i+1][j]+img[i+1][j+1] ;
    OutImg[0][i/Rt][j/Rt] = t1/4.0 ;
    OutImg[1][i/Rt][j/Rt] = t2/4.0 ;
    OutImg[2][i/Rt][j/Rt] = t3/4.0 ;
    OutImg[3][i/Rt][j/Rt] = t4/4.0 ;
  }
    		
  	
  /* subsampling right margin */
  if(wd_odd>0) {
    for(i=0; i< ht-1; i+=Rt) {  	
     	t1= img[i][wd-1]+img[i][wd-2]-img[i+1][wd-1]-img[i+1][wd-2];
	t2= img[i][wd-1]-img[i][wd-2]+img[i+1][wd-1]-img[i+1][wd-2];
	t3= img[i][wd-1]-img[i][wd-2]-img[i+1][wd-1]+img[i+1][wd-2];
	t4= img[i][wd-1]+img[i][wd-2]+img[i+1][wd-1]+img[i+1][wd-2];
	OutImg[0][i/Rt][wd1-1] = t1/4.0 ;
	OutImg[1][i/Rt][wd1-1] = t2/4.0 ;
	OutImg[2][i/Rt][wd1-1] = t3/4.0 ;
 	OutImg[3][i/Rt][wd1-1] = t4/4.0 ;
    }
  }
  	
  /* subsampling bottom */
  if(ht_odd>0) {
    for(i=0; i< wd-1; i+=Rt) {  	
	t1= img[ht-1][i]+img[ht-2][i]-img[ht-1][i+1]-img[ht-2][i+1];
	t2= img[ht-1][i]-img[ht-2][i]+img[ht-1][i+1]-img[ht-2][i+1];
	t3= img[ht-1][i]-img[ht-2][i]-img[ht-1][i+1]+img[ht-2][i+1];
	t4= img[ht-1][i]+img[ht-2][i]+img[ht-1][i+1]+img[ht-2][i+1];
	OutImg[0][ht1-1][i/Rt] = t1/4.0 ;
	OutImg[1][ht1-1][i/Rt] = t2/4.0 ;
	OutImg[2][ht1-1][i/Rt] = t3/4.0 ;
 	OutImg[3][ht1-1][i/Rt] = t4/4.0 ;
    }
  }
    
  /* calculate the right bottom corner */
  if((ht_odd>0)&&(wd_odd>0)) {
	t1= img[ht-1][wd-1]+img[ht-2][wd-1]-img[ht-1][wd-2]-img[ht-2][wd-2];
	t2= img[ht-1][wd-1]-img[ht-2][wd-1]+img[ht-1][wd-2]-img[ht-2][wd-2];
	t3= img[ht-1][wd-1]-img[ht-2][wd-1]-img[ht-1][wd-2]+img[ht-2][wd-2];
	t4= img[ht-1][wd-1]+img[ht-2][wd-1]+img[ht-1][wd-2]+img[ht-2][wd-2];
	OutImg[0][ht1-1][wd1-1] = t1/4.0 ;
	OutImg[1][ht1-1][wd1-1] = t2/4.0 ;
	OutImg[2][ht1-1][wd1-1] = t3/4.0 ;
 	OutImg[3][ht1-1][wd1-1] = t4/4.0 ;
 }
 (*wdPt) = wd1 ;  (*htPt) = ht1 ;
}


/**********************************************************************/
/* wavelet transformation */
/*	x11  x12  -> (x11+x12-x21-x22)/4  (x11-x12+x21-x22)/4*/
/*	x21  x22  -> (x11-x12-x21+x22)/4  */
/***********************************************************************/
static void WaveLetNoMean(
	double **img,
	int wd,
	int ht,
	double ***OutImg,
	int Rt,
	int *wdPt,
	int *htPt
)
{
  int i,j ;
  int wd1,ht1 ;
  int wd_odd, ht_odd ;
  int Dimen ;	/* dimension of the output multispectrum image(s) */
  double t1,t2,t3 ;
  
  Dimen = 3 ;	/* dimension of the output multispectrum image(s) */
  if(vlevel >1) printf("\twavelet transformation (No Mean)... \n") ; 
  
  	
  /* calculate the wd and ht of the output image(s) */
  if(Rt != 2) {
      fprintf(stderr, "2x2 subsampling is needed, not Rt=%d\n", Rt) ;
      exit(-10) ;
  }
  wd_odd = wd % Rt ;
  ht_odd = ht % Rt ;
  wd1 = (int) wd / Rt ;
  ht1 = (int) ht / Rt ;
  if(wd_odd>0) wd1 ++ ;
  if(ht_odd>0) ht1 ++ ;
  for(i=0; i<Dimen; i++)
    OutImg[i]=(double **)get_img(wd1,ht1,sizeof(double));
    	
   	
  /* waveLet transfermation of the main body */
  for(i=0; i< ht-1; i+=Rt)
  for(j=0; j< wd-1; j+=Rt) {
    t1= img[i][j]+img[i][j+1]-img[i+1][j]-img[i+1][j+1] ;
    t2= img[i][j]-img[i][j+1]+img[i+1][j]-img[i+1][j+1] ;
    t3= img[i][j]-img[i][j+1]-img[i+1][j]+img[i+1][j+1] ;
    OutImg[0][i/Rt][j/Rt] = t1/4.0 ;
    OutImg[1][i/Rt][j/Rt] = t2/4.0 ;
    OutImg[2][i/Rt][j/Rt] = t3/4.0 ;
  }
    		
  	
  /* subsampling right margin */
  if(wd_odd>0) {
    for(i=0; i< ht-1; i+=Rt) {  	
     	t1= img[i][wd-1]+img[i][wd-2]-img[i+1][wd-1]-img[i+1][wd-2] ;
	t2= img[i][wd-1]-img[i][wd-2]+img[i+1][wd-1]-img[i+1][wd-2] ;
	t3= img[i][wd-1]-img[i][wd-2]-img[i+1][wd-1]+img[i+1][wd-2] ;
	OutImg[0][i/Rt][wd1-1] = t1/4.0 ;
	OutImg[1][i/Rt][wd1-1] = t2/4.0 ;
	OutImg[2][i/Rt][wd1-1] = t3/4.0 ;
    }
  }
  	
  /* subsampling bottom */
  if(ht_odd>0) {
    for(i=0; i< wd-1; i+=Rt) {  	
	t1= img[ht-1][i]+img[ht-2][i]-img[ht-1][i+1]-img[ht-2][i+1] ;
	t2= img[ht-1][i]-img[ht-2][i]+img[ht-1][i+1]-img[ht-2][i+1] ;
	t3= img[ht-1][i]-img[ht-2][i]-img[ht-1][i+1]+img[ht-2][i+1] ;
	OutImg[0][ht1-1][i/Rt] = t1/4.0 ;
	OutImg[1][ht1-1][i/Rt] = t2/4.0 ;
	OutImg[2][ht1-1][i/Rt] = t3/4.0 ;
    }
  }
    
  /* calculate the right bottom corner */
  if((ht_odd>0)&&(wd_odd>0)) {
	t1= img[ht-1][wd-1]+img[ht-2][wd-1]-img[ht-1][wd-2]-img[ht-2][wd-2] ;
	t2= img[ht-1][wd-1]-img[ht-2][wd-1]+img[ht-1][wd-2]-img[ht-2][wd-2] ;
	t3= img[ht-1][wd-1]-img[ht-2][wd-1]-img[ht-1][wd-2]+img[ht-2][wd-2] ;
	OutImg[0][ht1-1][wd1-1] = t1/4.0 ;
	OutImg[1][ht1-1][wd1-1] = t2/4.0 ;
	OutImg[2][ht1-1][wd1-1] = t3/4.0 ;
 }

 (*wdPt) = wd1 ;  (*htPt) = ht1 ;
}



/***********************************************************/
/* subsample a image by averaging according to the give*/
/*   sRate.*/
/************************************************************/
void sub_sample(double **InImg, int wd, int ht, int sRate,
    double ***OutImg, int *wdPt, int *htPt)
{
    int wd1,ht1 ;
    int wd_odd, ht_odd ;  /* if wd, ht can be divided by sRate */
    int i,j,i1,j1 ;
    double avg ;
    
    if(vlevel>1) printf("\n\tsubsampling ...... \n") ;
    
    /* calculate the wd and ht of the subsampled image */
    wd_odd = wd % sRate ;
    ht_odd = ht % sRate ;
    wd1 = (int) wd / sRate ;
    ht1 = (int) ht / sRate ;
    if(wd_odd>0) wd1 ++ ;
    if(ht_odd>0) ht1 ++ ;
    
    if( (*OutImg) != NULL ) free_img( (void **) (*OutImg) ) ;/* leave this line GF 3/25/02 */
    *OutImg = (double **) get_img(wd1, ht1, sizeof(double)) ;
    
    /* subsampling main body */
    for(i=0; i<ht-1; i+=sRate)
    for(j=0; j<wd-1; j+=sRate) {
    	avg = 0.0 ;
	for(i1=0; i1<sRate; i1++)
	for(j1=0; j1<sRate; j1++) {
		avg += InImg[i+i1][j+j1] ;
	}
	(*OutImg)[i/sRate][j/sRate] = avg /sRate/sRate ;
    }
    
    /* subsampling right margin */
    if(wd_odd>0) {
	    for(i=0; i<ht-1; i+=sRate) {
	        avg = 0 ;
	    	for(j=0; j<sRate; j++)
 	   	for(i1=0; i1<sRate; i1++) {
 	   		avg += InImg[i+i1][wd-1-j] ;
 	   	}
	    	(*OutImg)[i/sRate][wd1-1] = avg /sRate/sRate ;
	    }
    }
		    	
    /* subsampling bottom */
    if(ht_odd>0) {
	    for(i=0; i<wd-1; i+=sRate) {
	        avg = 0 ;
	    	for(j=0; j<sRate; j++)
 	   	for(i1=0; i1<sRate; i1++) {
 	   		avg += InImg[ht-1-j][i+i1] ;
 	   	}
	    	(*OutImg)[ht1-1][i/sRate] = avg /sRate/sRate ;
	    }
    }
    
    /* calculate the right bottom corner */
    if((ht_odd>0)&&(wd_odd>0)) {
    	avg = 0 ;
    	for(i=0; i<sRate; i++)
    	for(j=0; j<sRate; j++) {
    		avg += InImg[ht-1-i][wd-1-j] ;
    	}
        (*OutImg)[ht1-1][wd1-1] = avg /sRate/sRate ;
    }
    
    *wdPt = wd1 ;
    *htPt = ht1 ;
   
} 


void PrintFeatureCh(struct Par_GMM *GMM_par) 
{
	printf("\n\tFeature Name : %s \n", GMM_par->feature_name) ;
	printf("\tDimension :	%d \n", GMM_par->dimen) ;
	printf("\tRate :		%d \n", GMM_par->Rt) ;
	printf("\tMaxLevel : %d \n\n", GMM_par->MaxLevels) ;
	

}
	
	
/**********************************************************************/
/*  initial feature characters */
/***********************************************************************/
void FeatureInitial(
  eisl_img *img_in, 
  struct Par_GMM *GMM_par,
  int *ncol, int *nrow
)
{
  int Found, nRow, nCol ;
  char *featureName ;
  
  Found = 0 ;
  featureName = GMM_par->feature_name ;
  nRow = img_in->height ;
  nCol = img_in->width ;
  
  
  if(vlevel>1) printf("\tFeature initializing ...\n") ;
  
  if(strcmp(featureName , "WaveletNoMean") == 0) {
	Found = 1 ;
	GMM_par->GenerateFunc = &WaveLetNoMean ;
	GMM_par->Rt = 2 ;
	GMM_par->dimen = 3 ;

        /* the dimension after one 2x2 decimation */
  	nRow = (nRow + GMM_par->Rt -1) / GMM_par->Rt ; 
  	nCol = (nCol + GMM_par->Rt -1) / GMM_par->Rt ; 
  
  } 
 
  if(strcmp(featureName , "WaveletWithMean") == 0) {
	Found = 1 ;
	GMM_par->GenerateFunc = &WaveLetTrans ;
	GMM_par->dimen = 4 ;
	GMM_par->Rt = 2 ;
        
        /* the dimension after one 2x2 decimation */
  	nRow = (nRow + GMM_par->Rt -1) / GMM_par->Rt ; 
  	nCol = (nCol + GMM_par->Rt -1) / GMM_par->Rt ; 
  
  } 

  if(Found == 0) {
    fprintf(stderr, "undefined feature name, %s !\n", featureName);
    error("FeatureInitial()");
  }

  
  (*nrow) = nRow ;
  (*ncol) = nCol ;

}





/**********************************************************************/
/* find the eight neighbors of a pixel */
/***********************************************************************/
void FindNeighbors(int **neighbors, int i0, int j0, int ht, int wd, int Dimen)
{
   int i,j,k, i1,j1 ;
   
   if((ht<2)||(wd<2)) {
   	fprintf(stderr, "The image is too small for FindNeighbors().\n") ;
   	exit(-40) ;
   }
   for(i=0; i<Dimen; i++) {
   	neighbors[i][0] = NULL ;
   	neighbors[i][1] = NULL ;
   }

   k=0 ;
   for(i=-1; i<=1; i++)
   for(j=-1; j<=1; j++) {
   	if((i!=0)||(j!=0)) { 	/* not including itself as a neighbor */
   		i1 = i0 + i ; 	j1 = j0 + j ;
   		if(i1<0) i1 = -i1 ;
   		if(j1<0) j1 = -j1 ;
   		if(i1>=ht) i1 = ht-2 ;
   		if(j1>=wd) j1 = wd-2 ;
   		neighbors[k][0]	= i1 ;
   		neighbors[k][1] = j1 ;
   		k++ ;
   	}
   }
   if(k!=Dimen) {
   	fprintf(stderr,"Dimension conflicts in FindNeighbors(), %d, %d\n",k,Dimen) ;
   	exit(-40) ;
   }
}




/* sample routine */
/* */
/*void main()  */
/*{ */
/*  double ***Imgs; */   /* feature vectors  */
/*  double **Img0, **Img1 ; */   /* org and decimated image */
/*  double ***ImgPt1, ***ImgPt2, ***ImgPt3 ; */   
/*  int D, dimen, nRow,nCol, i,j, i1,j1,k1 ; */
/*  int wd, ht, wd1, ht1 ; */
/*  int RateSS ;    */   /* subsampling rate  */
/*  double ***ll; */ /*log likehood at each pixel of the multispectral images */


/*  D = 2 ; */
/*  RateSS = 2 ; */
/*  dimen = 3 ; */

/*  wd = 14 ; */
/*  ht = 14 ; */
/*  Img0 = (double **) get_img(wd,ht, sizeof(double)) ; */
/*  for(i=0; i<ht; i++) */
/*  for(j=0; j<wd; j++) Img0[i][j] = i+j ; */
  
/*  ImgPt2 = &Img0 ;  */  /* initial for the case i==0  */
/*  wd1 = wd ; */
/*  ht1 = ht ; */
/*  ImgPt1 = &Img1 ; */
 
/*  for(i=0; i<=D; i++) { */
    
 
/*    if(i!=0) sub_sample((*ImgPt1),wd,ht,RateSS,ImgPt2,&wd1,&ht1) ; */
    
/*    printf("\nOrg or Dec Img\n") ; */
/*    for(i1=0; i1<ht1; i1++) { */
/*        for(j1=0; j1<wd1; j1++) printf("  %lf ", (*ImgPt2)[i1][j1]) ; */
/*        printf("\n") ; */
/*    } */
    
/*    Imgs = (double ***) get_spc(dimen,sizeof(double **)); */
/*    WaveLetNoMean((*ImgPt2),wd1,ht1, Imgs, 2, &wd, &ht) ;     */

/*    printf("\nWaveLet w/o mean\n") ; */
/*    for(i1=0; i1<ht; i1++) { */
/*        for(j1=0; j1<wd; j1++) { */
/*            printf("   (") ; */
/*            for(k1=0; k1<3; k1++)  printf("%lf ", Imgs[k1][i1][j1]) ; */
/*            printf("   )") ; */
/*        } */
/*       printf("\n") ; */
/*    } */
    


/*    for(j=0; j< dimen; j++)  free_img((void **) Imgs[j]) ; */
/*    if(Imgs != NULL)  free( (void *) Imgs) ; */

/*    free_img( (void **) (*ImgPt1) ) ;*/ /* switch org img and decimated img */ 
/*    ImgPt3 = ImgPt1 ; */
/*    ImgPt1 = ImgPt2 ; */
/*    ImgPt2 = ImgPt3 ; */
/*    wd = wd1;  ht = ht1; */
/*  } */
/*} */
  

