#include "seg_global.h"
#include "seg.h"


/**********************************************************/
/*  read images and calculate log probability X given Y*/
/***********************************************************/
void FeatureProb(
  struct BigPyramid* Pyramid,
  eisl_img* imgPt,     /* input images */
  struct Par_GMM *GMM_par 
)  
{
  double ***Imgs;   /* feature vectors */
  double **Img0=NULL, **Img1=NULL ;   /* org and decimated images */
  double ***ImgPt1, ***ImgPt2, ***ImgPt3 ;   
  int D, dimen, nClasses, nRow,nCol, i,j ;
  int wd, ht, wd1, ht1 ;
  int RateSS ;      /* subsampling rate */
  double ***ll;      /* log likehood at each pixel of the multispectral images*/


  D = GMM_par->MaxLevels ;
  RateSS = GMM_par->Rt ;
  dimen = GMM_par->dimen ;
  nClasses = GMM_par->nclasses ;

  wd = imgPt->width ;
  ht = imgPt->height ;
  Img0 = (double **) get_img(wd,ht, sizeof(double)) ;
  for(i=0; i<ht; i++)
  for(j=0; j<wd; j++) Img0[i][j] = (double) (imgPt->mono)[i][j] ;

  
  ImgPt2 = &Img0 ;    /* initial for the case i==0 */
  wd1 = wd ;
  ht1 = ht ;
  ImgPt1 = &Img1 ;
 
  for(i=0; i<=D; i++) {
    
 
    if(i!=0) sub_sample((*ImgPt1),wd,ht,RateSS,ImgPt2,&wd1,&ht1) ;
    


    if(GMM_par->existSig[i]) {
            
        Imgs = (double ***) get_spc(dimen,sizeof(double **));
  
        (*(GMM_par->GenerateFunc))((*ImgPt2),wd1,ht1, 
               Imgs, GMM_par->Rt, &wd, &ht) ;    

  
        nRow = (Pyramid->nRow)[i] ;
        nCol = (Pyramid->nCol)[i] ;
        if((nRow != ht) || (nCol != wd)) {
            fprintf(stderr, "\n Inconsistant dimension for features !!\n") ;
            exit(-1) ;
        }
        
        
        /* calculate log likelihood for each pixel and each class */
        /* allocate LProb[level][i][j][class] */
        (Pyramid->LProb)[i] = (double ***) multialloc(sizeof(double), 3, 
            nRow, nCol, nClasses) ;
        ll = (Pyramid->LProb)[i] ;
    
        /* calculate log likelihood of each pixel,  containing in model.c */
        if(vlevel >=3) printf("\t%d  calculating loglikelihood ...", i) ;


        extract(Imgs, nRow, nCol, ll, &((GMM_par->_Sig)[i])) ; 
           

        for(j=0; j< dimen; j++)  {
            if( Imgs[j] != NULL) free_img((void **) Imgs[j]) ;
        }
        if(Imgs != NULL)  free( (void *) Imgs) ;

    }
    
    /* switch org img and decimated img */
    ImgPt3 = ImgPt1 ;
    ImgPt1 = ImgPt2 ;
    ImgPt2 = ImgPt3 ;
    wd = wd1;  ht = ht1;
  }


  if(vlevel >=3) printf("\tFinish FeatureProb(). \n") ;
}
  
  
  

/**********************************************************/
/*  Output log probability X given Y*/
/***********************************************************/
void FeatProbOutput(struct BigPyramid* Pyramid, struct Par_GMM *GMM_par, 
    char *fn) 
{
  FILE *fd ;
  int D, nClasses, nRow,nCol, i,i1,j1,k1 ;
  double ***ll;  /* log likehood at each pixel of the multispectral images*/

  
  if(vlevel >1)  printf("\tWriting Feature LogLikehood into %s ...\n", fn) ;
  fd = fopen (fn, "w");  if (fd == NULL) FileOpenErr(fn);
  
  D = Pyramid->level ;        /* not GMM_par->MaxLevels,  see FeatProbRead */
  nClasses = GMM_par->nclasses ;
  fprintf(fd, "total_level=%d\n", D) ;
  
  for(i=0; i<(GMM_par->MaxLevels); i++) {
    
    fprintf(fd, "level=%d\n", i) ;
    fprintf(fd, "existSig=%d\n", GMM_par->existSig[i]);  /* existSig[i] = 1 */

    if(GMM_par->existSig[i]) {
            
        nRow = (Pyramid->nRow)[i] ;
        nCol = (Pyramid->nCol)[i] ;
        
        fprintf(fd, "nRow=%d\n", nRow) ;
        fprintf(fd, "nCol=%d\n", nCol) ;
        
        /* calculate log likelihood for each pixel and each class */
        /* allocate LProb[level][i][j][class] */
        ll = (Pyramid->LProb)[i] ;
        for(i1=0; i1<nRow; i1++)
        for(j1=0; j1<nCol; j1++) {
            for(k1=0; k1<nClasses; k1++) fprintf(fd, "%lf ", ll[i1][j1][k1]) ;
            fprintf(fd, "\n") ;
        }
    }
  }
}
  
  
  

/**********************************************************/
/*  Output log probability X given Y*/
/***********************************************************/
void FeatProbRead(struct BigPyramid* Pyramid, struct Par_GMM *GMM_par, 
    char *fn) 
{
  FILE *fd ;
  int D,nClasses, nRow,nCol, i,j, i1,j1,k1 ;
  double ***ll;      /* log likehood at each pixel of the multispectral images*/


  
  if(vlevel >1)  printf("\tReading Feature LogLikehood into %s\n", fn) ;
  fd = fopen (fn, "r");  if (fd == NULL) FileOpenErr(fn);
  
  fscanf(fd, "total_level=%d\n", &D) ;
  if(D < GMM_par->MaxLevels) { 
      fprintf(stderr, "D %d < GMM_par->MaxLevel %d\n",D, GMM_par->MaxLevels);  
      exit(-1); 
  }
   
  nClasses = GMM_par->nclasses ;
  Pyramid->level = D ;
  Pyramid->nRow = (int*) mget_spc(D,sizeof(int)) ;
  Pyramid->nCol = (int*) mget_spc(D,sizeof(int)) ;
  Pyramid->LProb = (double****) mget_spc(D,sizeof(double***)) ;
  Pyramid->LLikeH = (double****) mget_spc(D,sizeof(double***)) ;
  Pyramid->_Segment = (unsigned char***) mget_spc(D,sizeof(unsigned char**)) ;

  for(i=0; i<(GMM_par->MaxLevels); i++) {
    
    fscanf(fd, "level=%d\n", &j) ;
    printf("%d,  level=%d\n", i, j) ;
    if(i!=j) { fprintf(stderr, "Wrong format 1\n");  exit(-1);  }
    
    fscanf(fd, "existSig=%d\n", &j) ;    /* existSig[i] = 1 */
    printf("existSig=%d\n", j) ;
    GMM_par->existSig[i] = j ;

    if(GMM_par->existSig[i]) {
            
        fscanf(fd, "nRow=%d\n", &nRow) ;
        fscanf(fd, "nCol=%d\n", &nCol) ;
        printf("nRow=%d\n", nRow) ;
        printf("nCol=%d\n", nCol) ;

        (Pyramid->nRow)[i] = nRow ;
        (Pyramid->nCol)[i] = nCol ;
        
        /* calculate log likelihood for each pixel and each class */
        /* allocate LProb[level][i][j][class] */
        (Pyramid->LProb)[i] = (double ***) multialloc(sizeof(double),
            3,nRow,nCol,nClasses) ;
        ll = (Pyramid->LProb)[i] ;
        for(i1=0; i1<nRow; i1++)
        for(j1=0; j1<nCol; j1++) {
            for(k1=0; k1<nClasses; k1++) fscanf(fd, "%lf ", &(ll[i1][j1][k1]));
        }
    }
  }
}
  
  
  
