#include <sys/stat.h>
#include <sys/times.h>
#include "seg_global.h"
#include "seg.h"

/* */
/*    "model.c" has been modified.  Now a look up table for exp() */
/*     is available.   The table is on-line constructed and stores  */
/*     values of exp() in the range of [0,-20] using 20000 bins.  The */
/*     bins are equally spaced. */
    
/*     08/11/97 */
    
    
/*     modification: */
/*     (1) If nsubclasses = 0,  ll[i][j][m] = -20 ; */
/*     (2) If logcheck < 0.0000000001,  ll[i][j][m] = -15+maxlike ; */
    
/*     02/11/98 */
/* */

/* static struct tms time1, time2 ; */


/***************************************************************************/
/*  Returns TRUE, if there are more than one class containing subclasses.*/
/*  For multiscale model, when the sample images are not enough,  there */
/*  may be empty classes */
/***************************************************************************/
int extract_init(struct SigSet *S)
{
   int m; 
   int i;
   int b1,b2;
   int nbands;
   int non_singular_class ;
   double *lambda;
   double logcheck ;
   struct ClassSig *C;
   struct SubSig *SubS;

   nbands = S->nbands;
   /* allocate scratch memory */
   lambda = (double *)G_malloc(nbands*sizeof(double));

   /* invert matrix and compute constant for each subclass */

   /* for each class */
   non_singular_class = 0 ;
   for(m=0; m<S->nclasses; m++ ) {
     C = &(S->ClassSig[m]);

     if(vlevel>50) printf("\tclass %d  nSubCl %d\n", m, C->nsubclasses) ;
     
     if(C->nsubclasses > 0) non_singular_class ++ ;

     /* for each subclass */
     for(i=0; i<C->nsubclasses; i++) {
        SubS = &(C->SubSig[i]);

        /* Test for symetric  matrix */
        for(b1=0; b1<nbands; b1++)
        for(b2=0; b2<nbands; b2++) {
          if(SubS->R[b1][b2]!=SubS->R[b2][b1]) {
            printf("\nWarning: nonsymetric covariance for class %d ",m+1);
            printf("Subclass %d\n",i+1);
          }
          SubS->Rinv[b1][b2] = SubS->R[b1][b2];
        }

        /* Test for positive definite matrix */
        eigenNoEigenVector(SubS->Rinv,lambda,nbands);
        for(b1=0; b1<nbands; b1++) {
          if(lambda[b1]<=0.0) {
            fprintf(stderr,"Warning: nonpositive eigenvalues for class %d",m+1);
            fprintf(stderr,"Subclass %d\n",i+1);
            fprintf(stderr,"This Class is ignored \n") ;
            return(FALSE) ;
          }
        }

        /* Precomputes the cnst */
        SubS->cnst = (-nbands/2.0)*log(2*PI);
        for(b1=0; b1<nbands; b1++) {
          logcheck =  lambda[b1] ;
          if(logcheck <= 0) {
          	printf("Error in extract_init(S), logcheck = %lf \n", logcheck) ;
            fprintf(stderr,"This Class is ignored \n") ;
            return(FALSE) ;
          }
          SubS->cnst += - 0.5*log(logcheck);
        }

        /* Precomputes the inverse of tex->R */
        invert(SubS->Rinv,nbands);
        
        /* chang SubS.pi into the form exp(a), in order to speed up */
        logcheck = SubS->pi ;
        if(logcheck <= 0) {
          printf("Error in extract_init(S),  %lf \n", logcheck) ;
          fprintf(stderr,"This Class is ignored \n") ;
          return(FALSE) ;
        }
        SubS->pi = log(logcheck);
        
      }
    }
    free((char *)lambda);
    
    
    if( non_singular_class < 2) return(FALSE) ;
    
    return(TRUE) ;

}


void extract(
	double ***img,    /* multispectral image, img[band][i][j] */
	int nrow, int ncol,
	double ***ll,            /* log likelihood, ll[i][j][class] */
	struct SigSet *S        /* class signatures */
)
{
    int  i,j;             /* column and row indexes */
    int  m;               /* class index */
    int  k;               /* subclass index */
    int  b1,b2;           /* spectral index */
    int  nbands;          /* number of spectral bands */
    int  max_nsubclasses; /* maximum number of subclasses */
    double *subll, subtmp;        /* log likelihood of subclasses */
    double *diff; 	
    double maxlike;
    double subsum;
    double logcheck ;
    struct ClassSig *C, *CSig;
    struct SubSig *SubS;
    int nsubclasses;
    double **Rinv ;


    nbands = S->nbands;
    CSig = S->ClassSig ;

    /* determine the maximum number of subclasses */
    max_nsubclasses = 0;
    for(m=0; m<S->nclasses; m++ ) {
      if((CSig[m]).nsubclasses>max_nsubclasses)
        max_nsubclasses = (CSig[m]).nsubclasses;
    }
    
    
    /* allocate memory */
    diff = (double *)G_malloc(nbands*sizeof(double));
    subll = (double *)G_malloc(max_nsubclasses*sizeof(double));


    /* Compute log likelihood at each pixel and for every class. */

    /* for each pixel */
    for(i=0; i<nrow; i++)
    for(j=0; j<ncol; j++) {

        /* for each class */
        for(m=0, C=CSig; m<S->nclasses; m++, C++) { 
	    /*  C = &(CSig[m]);  */
	  
          nsubclasses = C->nsubclasses;

          /* compute log likelihood for each subclass */
          for(k=0, SubS=C->SubSig; k<nsubclasses; k++, SubS++) { 
            /* SubS = &((C->SubSig)[k]); */
            
            Rinv = SubS->Rinv ;
            subtmp = SubS->cnst ;  /* for efficiency */
                        
            
            for(b1=0; b1<nbands; b1++) {
              diff[b1] = img[b1][i][j] - (SubS->means)[b1];
              subtmp -= .5*diff[b1]*diff[b1]*Rinv[b1][b1]; 
            }
            for(b1=0; b1<nbands; b1++) 
            for(b2=b1+1; b2<nbands; b2++) {
              subtmp -= diff[b1]*diff[b2]*Rinv[b1][b2]; 
	        }

	    
	        /* pi is changed into the form of exp(a) */              
            subll[k] = subtmp + SubS->pi;  
         }

         /* shortcut for one subclass */
         if(nsubclasses<0) {
            fprintf(stderr, "ERROR: invalid nsubclasses %d\n", nsubclasses) ;
            exit(-10) ;
            
         } else if(nsubclasses == 0) {  /* empty class */
            ll[i][j][m] = log(MIN_LOG) ;
            
         } else if(nsubclasses==1) {
            ll[i][j][m] = subll[0];
         }
         /* compute mixture likelihood */
         else {
            /* find the most likely subclass */
            for(k=0; k<nsubclasses; k++)
            {
              if(k==0) maxlike = subll[k];
              if(subll[k]>maxlike) maxlike = subll[k];
            }
 
            /* Sum weighted subclass likelihoods */
            subsum = 0;
   
            for(k=0,SubS=C->SubSig; k<nsubclasses; k++,SubS++) {

 	        	 /********* exact calculation of exp() 
            	   subsum += exp( subll[k]-maxlike ); 
	        	 **********/

   	        	 /* calculate exp() using look up table */
            	 subsum += exp_lookup( subll[k]-maxlike ); 
	        }
	    
            logcheck = subsum ;
            
            /* modified on Feb 11, 1998 */
            /*  */
           /*  if(logcheck <= 0) { */
       	    /*	printf("Error in extract(img,region,ll,S), %lf\n", logcheck); */
         /* 	    exit(1) ; */
            /* } */
            /* ll[i][j][m] = log(logcheck) + maxlike; */
            /* */
            
            if(logcheck <= MIN_LOG) {
                ll[i][j][m] = log(MIN_LOG) + maxlike;
            } else {
                ll[i][j][m] = log(logcheck) + maxlike;
            }
            
         }
       }
    }

    free((char *)diff);
    free((char *)subll);

}
