#include "seg.h"
#include <sys/time.h>
#include <sys/times.h>
/*#include <sys/getopt.h>*/
#include <stdio.h>
#include <unistd.h>

void error(char *name) ;
void RandInitial();

/* Global variables */
int  vlevel;            /* level of verbose output */


main(int argc, char *argv[])
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd;
  int  nrow,ncol;             /* image nrow and ncol */
  struct BigPyramid _BigPyramid	;
  struct ModelPar Par ;       /* model parameters */
  eisl_img img_in;            /* input image, eith tiff or pxm,  eisl_img.h */
  char *infn;       	      /* input image file name */
  char *parfn;                /* input parameter set file names */
  char *otfn;                 /* output segmentation file name */
  int  DebugFlag=0 ;          /* default no debug */	
  int  iflag,pflag,oflag,MLflag,SMAPflag,QTrflag;
  int  X0,Y0 ;                /* indices of the pixel need to be written out */
  int  X0flag, Y0flag ;
  int  status;
  
  
  vlevel=1;
  iflag=pflag=oflag=MLflag=SMAPflag=X0flag=Y0flag=QTrflag=0;
  
  while( (c=getopt(argc,argv,"h,o:,i:,p:,x:,y:,v:,M:,S:,Q:"))!=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'p':		/* file name for parameters of data model */
        pflag=1;
        parfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(parfn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      case 'v':
        if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
        if(vlevel<0) error(argv[0]);
        break;
      case 'x':
        if(0==sscanf(optarg,"%d",&X0)) error(argv[0]);
        if(X0<0) error(argv[0]);
        X0flag = 1 ;
        break;
      case 'y':
        if(0==sscanf(optarg,"%d",&Y0)) error(argv[0]);
        if(Y0<0) error(argv[0]);
        Y0flag = 1 ;
        break;
      case 'M':
        MLflag = 1;
        break;
      case 'S':
        SMAPflag = 1;
        break;
      case 'Q':
        QTrflag = 1;
        break;
      default:
        error(argv[0]);
        break;
    }
  }

  /* Check for inconsistant inputs */   
  if( !(oflag && pflag && iflag) ) {
    fprintf(stderr, "input, output, parameter file are required, %s\n",argv[0]);
    error(argv[0]);
  } 
    
  
  /* set default */
  if((!MLflag)&&(!SMAPflag)&&(!QTrflag)) SMAPflag = 1;
  
  /* read parameters */
  ReadQuadTreePar( &(Par.QTr_par), parfn );   /* ../../segment/src/seg_io.c */
  ReadGMM_Par( &(Par.GMM_par), parfn );
  if( ! (QTrflag || MLflag) ) ReadCPT_Par( &(Par.CPT), parfn ); 
  if(vlevel>10) WriteQuadTreePar( stdout, &(Par.QTr_par) ); 
  if(vlevel>100) WriteGMM_Par( &(Par.GMM_par), NULL );
  
  
  
  /* check debug mode */
  if(X0flag && Y0flag) {
  	DebugFlag = 1;
  	printf("\tDebug mode, X=%d, Y=%d\n\n",X0,Y0) ;
  } else {
  	DebugFlag = 0 ;
  	X0= Y0 = -1 ;
  }
  
    
  
  /* read input image */
  if(vlevel >0) printf("\treading input image %s\n", infn) ;
  if( ! read_img (infn, &img_in) ) {
  	  fprintf(stderr,"ERROR: Cannot read %s, exit main() !!\n", infn);
  	  exit (-10);
  }
  if( img_in.img_type != 'g' ) {
  	  fprintf(stderr,"ERROR: input must be graylevel image, exit main() !!\n");
  	  exit (-10);
  }
      

  /* Initialize the data model */
  /* Feature Initialization		feature.c */
  FeatureInitial(&img_in, &(Par.GMM_par), &ncol, &nrow) ;  
  if(vlevel>0) PrintFeatureCh(&(Par.GMM_par)) ;	

  /* construct the pyramid */
  ConstructPyramid(nrow,ncol,&_BigPyramid, &Par) ;  	/* u_pyramid.c */
  RandInitial() ;  /* Initalization */

  FeatureProb(&_BigPyramid, &img_in, &(Par.GMM_par)) ; /* featureProb.c */
  free_img((void **) img_in.mono) ;

  /* segmentation begins */
  if(MLflag) {
  	ML_Test(&_BigPyramid, &Par, otfn) ;  /* ML.c */

  } else {
	upPropagate(&_BigPyramid, &Par) ;  /* decimate.c */

  	if(QTrflag) {
  	    MAP_QuadTree(&_BigPyramid, &Par, otfn, X0, Y0); /* MAP.c */
  	    
        } else {  /* default SMAP */
            MAP_Pyramid(&_BigPyramid, &Par, otfn, X0, Y0) ; /* MAP.c */
        }
	

  }

  fflush(NULL) ;

}


void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i image -- pgm format image\n");
  fprintf(stderr,"     -p params -- File containing GMM parameters.\n");
  fprintf(stderr,"     -o output -- Segmentation image in pgm format\n");
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");
  fprintf(stderr,"     		    > 15 output segment of all scales.\n");
  fprintf(stderr,"     -<algorithm> -- SMAP, Maxlike, QuadTree; Default SMAP\n");
  fprintf(stderr,"     -x,y -- indices of the pixel whose info needs to be\n");
  fprintf(stderr,"             written out for debug. They must be all >0\n") ;
  
  exit(1);
}

