#include "seg.h"
#include "seg_global.h"


void ConstructPyramid (int nrow, int ncol, struct BigPyramid *bigPyramid, 
    struct ModelPar *Par)
{
   int D ;
   int i, row, col ;
   
   if(vlevel >1)
       printf("\tConstructing Pyramid ...\n") ;
       
   D = levels(nrow,ncol) ;     /* number of levels in the pyramid, 0 to D-1 */
   if(vlevel>10) printf("\t\tlevels D= %d, nrow %d, ncol%d \n", D, nrow, ncol) ;
   if(D > Par->QTr_par.levels) D=Par->QTr_par.levels ;
       
   
   bigPyramid->level = D ;
   bigPyramid->nRow = (int*) mget_spc(D,sizeof(int)) ;
   bigPyramid->nCol = (int*) mget_spc(D,sizeof(int)) ;

   bigPyramid->LProb = (double****) mget_spc(D,sizeof(double***)) ;

   bigPyramid->LLikeH = (double****) mget_spc(D,sizeof(double***)) ;

   bigPyramid->_Segment=(unsigned char***) mget_spc(D,sizeof(unsigned char**));
   
   row = nrow ;  
   col = ncol ;  
   
   for(i=0; i<D; i++) {
         (bigPyramid->nRow)[i] = row ;
         (bigPyramid->nCol)[i] = col ;
         if(vlevel>10) printf("\t\trow = %d,  col = %d \n", row, col) ;
         row = row/2 + row%2 ;  /* decimate by 2 ONLY */
         col = col/2 + col%2 ; 

      (bigPyramid->LProb)[i] = NULL ;
      (bigPyramid->LLikeH)[i] = NULL ;
      (bigPyramid->_Segment)[i] = NULL ;
   }
}
    


void FreePyramid (struct BigPyramid *bigPyramid)
{
   int D, i ;
   double ***tmp1, ***tmp2 ;
   unsigned char **seg ;
   
   if(vlevel >1)
       printf("\tFree Pyramid ...\n") ;
       
   D = bigPyramid->level ;     /* number of levels in the pyramid, 0 to D-1 */


   free((void*) (bigPyramid->nRow)) ;
   free((void*) (bigPyramid->nCol)) ;

   for(i=0; i<D; i++) {
       /***********************************************
         tmp = (bigPyramid->LProb)[i] ;
      if(tmp!=NULL) multifree((void *)tmp, 3) ;

         tmp = (bigPyramid->LLikeH)[i] ;
      if(tmp!=NULL) multifree((void *)tmp, 3) ;
	Note: BUG!!! - When i=0, the two pointers above are the same.
            But they are freed separately so that the same memory space
            is freed twice.  GF 3/25/02
 ********************************************************/
         tmp1 = (bigPyramid->LProb)[i] ;
         if(tmp1!=NULL) multifree((void *)tmp1, 3) ;

         tmp2 = (bigPyramid->LLikeH)[i] ;
         if(tmp2!=NULL && tmp2!=tmp1) multifree((void *)tmp2, 3) ;
	 	 
	 seg = (bigPyramid->_Segment)[i] ;
         if(seg!=NULL) free_img((void **) seg) ;
   }

   free((void*) (bigPyramid->_Segment)) ;
   free((void*) (bigPyramid->LLikeH)) ;
   free((void*) (bigPyramid->LProb)) ;
   
}
    

int levels(int nrow, int ncol)
{
    int D;

    D=0;
    while((nrow>2)&&(ncol>2)) {
      D++;
      nrow = nrow/2 + nrow%2;
      ncol = ncol/2 + ncol%2;
    }

    return(D);
}



/*****************************************************************************/
/* generate look up table fo exp(),  input range [0, -20]*/ 
/* When EXPnBins=20000,  exp_lookup() = exp() * (1 +/- 0.001)*/
/*****************************************************************************/

double exp_lookup( double x ) 
{
    static int Initialized=0 ;
    static double *MYEXP, EXPratio ;
    int i, Index ;
    double step ;

    if( ! Initialized )  {   /* initialization */

        Initialized = 1 ;
        MYEXP = (double *)get_spc(EXPnBins + 1, sizeof(double)) ;
        EXPratio = EXPnBins / (EXP_MAX - EXP_MIN) ;
        step = 1.0 / exp( (EXP_MAX - EXP_MIN) / EXPnBins ) ;
        MYEXP[0] = 1.0 ;
        for(i=1; i<=EXPnBins; i++) MYEXP[i]=MYEXP[i-1] * step ;
    } 
    
    
    /* look up */
    if(x > EXP_MAX) {
        fprintf(stderr, "WARNING:  input of exp_lookup %lf > MAX_input %lf\n",
            x, EXP_MAX) ;
        fprintf(stderr, "WARNING:  Set input to MAX_input. \n") ;
        fprintf(stderr, "WARNING:  Large distortion may occurn\n") ;
        x = EXP_MAX ;
    }
    
    if(x < EXP_MIN) x = EXP_MIN ;
    
    Index = (int) ( (EXP_MAX - x) * EXPratio + .5) ;
   	if(Index>EXPnBins) Index=EXPnBins ;

   	return( MYEXP[Index] );
   
}



/*****************************************************************************/
/* generate look up table fo log_e(),  input range [MIN_LOG, 1]*/ 
/* When EXPnBins=50000,  log_e(0.001) - log_e(0.00098) = 0.0202*/
/*****************************************************************************/

double logE_lookup( double x ) 
{
    static int Initialized=0 ;
    static double *MYLOGE, LOGEratio ;
    int i, Index ;
    double step, tmp ;
static int count=0 ;

    if( ! Initialized )  {   /* initialization */

        Initialized = 1 ;
        MYLOGE = (double *)get_spc(LOGEnBins + 1, sizeof(double)) ;
        LOGEratio = LOGEnBins / (LOG_MAX - MIN_LOG) ;
        step = (LOG_MAX - MIN_LOG) / LOGEnBins   ;

        for(i=0, tmp=LOG_MAX ; i<=LOGEnBins; i++, tmp-=step) {
            MYLOGE[i] = log(tmp) ;
        }
    }
    
    /* look up */
    if(x > LOG_MAX) {
        fprintf(stderr, "WARNING:  input of logE_lookup %lf > MAX_input %lf\n",
            x, LOG_MAX) ;
        fprintf(stderr, "WARNING:  Set input to MAX_input. \n") ;
        fprintf(stderr, "WARNING:  Large distortion may occurn\n") ;
        x = EXP_MAX ;
    }
    
    if(x < MIN_LOG) x = MIN_LOG ;
    
    Index = (int) ( (LOG_MAX - x) * LOGEratio + .5) ;
   	if(Index>LOGEnBins) Index=LOGEnBins ;

if(count<10) {
printf("logE_lookup(%lf)=%lf\n", x, MYLOGE[Index]) ;
count++ ;
}
   	
   	return( MYLOGE[Index] );
}
