#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "Tiff_img.h"

#if defined(VMS)
#define unlink delete
#endif

#define    strneq(a,b,n)    (strncmp(a,b,n) == 0)

#define    TRUE    1
#define    FALSE    0

static    int ignore = FALSE;        /* if true, ignore read errors */
static    int outtiled = -1;

static    uint16 defpredictor = (uint16) -1;
static    uint16 predictor ;
static    uint16 defcompression = (uint16) -1;
static    uint32 g3opts ;
static    uint32 defg3opts = (uint32) -1;
static    int quality = 75;        /* JPEG quality */
static    int jpegcolormode = JPEGCOLORMODE_RGB;

static    void usage(void);


/******************************************************************************/
void *tiff_get_spc(int num, size_t size)
{
        void *pt;

        if( (pt=_TIFFmalloc(num*size)) == NULL ) {
                fprintf(stderr, "==> malloc() error\n");
                fprintf(stderr, "need to allocate %d x %d\n", num, (int) size);
                return(NULL) ;
        }
        return(pt);
}


/******************************************************************************/
void **tiff_get_img(int wd,int ht,size_t size)
{
        int i;
        void  **ppt;
        char   *pt;

        ppt = (void **)tiff_get_spc(ht,sizeof(void *));
        pt = (char *)tiff_get_spc(wd*ht,size);
    
        if(ppt != NULL) for(i=0; i<ht; i++) ppt[i] = pt + i*wd*size;

        return(ppt);
}


/******************************************************************************/
void tiff_free_img(void **pt)
{
        free( (void *)pt[0]);
        free( (void *)pt);
}


/******************************************************************************/
static int
checkcmap(int n, uint16* r, uint16* g, uint16* b)
{
    while (n-- > 0) {
        if (*r++ >= 256 || *g++ >= 256 || *b++ >= 256) return (16);
    }
    /* fprintf(stderr, "Warning, assuming 8-bit colormap.\n"); */
    return (8);
}


/******************************************************************************/
static int  
Initial_Read_TIFF(TIFF* in, TIFF_img *tiff_img) 
{
    uint32  w,l;
    int    cmap, ii ;
    uint16 *rmap, *gmap, *bmap ;
    uint16 config;
    uint16 compress;
    uint16 horizSubSampling;    /* YCbCr horizontal subsampling */
    uint16 vertSubSampling;        /* YCbCr vertical subsampling */
    uint16 bitspersample;
    uint16 samplesperpixel;
    uint16 photometric;
    uint16 fillorder;
    
    TIFFGetField(in, TIFFTAG_BITSPERSAMPLE, &bitspersample);
    tiff_img->bitspersample = bitspersample ;
    if ((bitspersample != 8) && (bitspersample != 1)) {
        fprintf(stderr, "Sorry, only handle 8-bit or 1-bit samples.\n");
        fprintf(stderr, "No %d-bit samples.\n", bitspersample);
        return (FALSE);
    }
    if (TIFFIsTiled(in)) {
        fprintf(stderr, "Sorry, only handle strips, not tiled.\n");
        return (FALSE);
    }
    TIFFGetField(in, TIFFTAG_PLANARCONFIG, &config);
    tiff_img->planarConfig = config ;
    if (config == PLANARCONFIG_SEPARATE) {
        fprintf(stderr,
           "Sorry, only handle single image plane, not separate.\n");
        return (FALSE);
    }

    TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &w);
    tiff_img->width = w ;
    
    TIFFGetField(in, TIFFTAG_IMAGELENGTH, &l);
    tiff_img->height = l ;

    TIFFGetField(in, TIFFTAG_FILLORDER, &fillorder);
    tiff_img->fillorder = fillorder ;
    
    TIFFGetField(in, TIFFTAG_SAMPLESPERPIXEL, &samplesperpixel);
    tiff_img->samplesperpixel = samplesperpixel ;
    switch ( samplesperpixel ) {
      case 1:
            tiff_img->mono = (uint8 **) tiff_get_img(w,l,sizeof(uint8)) ;
            tiff_img->luminance = tiff_img->mono ;
            break;
      case 3:
          tiff_img->color = (uint8 ***) tiff_get_spc(samplesperpixel, 
              sizeof(uint8 **));
          for(ii=0; ii<samplesperpixel; ii++) {
              tiff_img->color[ii] = (uint8 **) tiff_get_img(w, l, sizeof(uint8)) ;
          }
          tiff_img->chrominance = (int8 ***) 
              tiff_get_spc(samplesperpixel-1, sizeof(uint8 **));
          for(ii=0; ii<samplesperpixel-1; ii++) {
              tiff_img->chrominance[ii] = (int8 **) tiff_img->color[ii+1] ;
          }
          tiff_img->luminance = tiff_img->mono = (uint8 **) tiff_img->color[0] ;
          break ;
      default:
        fprintf(stderr, "Only support samplesperpixel is 1 or 3\n") ;
        return(FALSE) ;
        break ;
    }
    
    photometric = -1 ;
    TIFFGetField(in, TIFFTAG_PHOTOMETRIC, &photometric);
    tiff_img->photometric = photometric ;
    switch ( photometric ) {
      case PHOTOMETRIC_MINISWHITE:
          fprintf(stdout, "\t0 is white !\n") ;
          tiff_img->TIFF_type = 'g';
          if( bitspersample == 1) tiff_img->TIFF_type = 'b';
          if(samplesperpixel != 1) {
          fprintf(stderr, "Wrong samplesperpixel %d for monochrome image\n",
              samplesperpixel);
          return(FALSE) ;
        }
          break ;
      case PHOTOMETRIC_MINISBLACK:
          fprintf(stdout, "\t0 is black !\n") ;
          tiff_img->TIFF_type = 'g';
          if( bitspersample == 1) tiff_img->TIFF_type = 'b';
          if(samplesperpixel != 1) {
          fprintf(stderr, "Wrong samplesperpixel %d for monochrome image\n",
              samplesperpixel);
          return(FALSE) ;
        }
          break ;
      case PHOTOMETRIC_PALETTE:
          tiff_img->TIFF_type = 'p' ;
          if(samplesperpixel !=1 ) {
          fprintf(stderr,"Wrong samplesperpixel %d for RGB image\n",
              samplesperpixel);
          return(FALSE) ;
        }
        if (!TIFFGetField(in, TIFFTAG_COLORMAP, &rmap, &gmap, &bmap)) {
            fprintf(stderr, "No colormap (not a valid palette image).\n");
            return (FALSE);
        }
        tiff_img->rmap = (uint16 *) tiff_get_spc(1<<bitspersample, sizeof(uint16)) ;
        tiff_img->gmap = (uint16 *) tiff_get_spc(1<<bitspersample, sizeof(uint16)) ;
        tiff_img->bmap = (uint16 *) tiff_get_spc(1<<bitspersample, sizeof(uint16)) ;
        cmap = checkcmap(1<<bitspersample, rmap, gmap, bmap);
        for (ii = (1<<bitspersample)-1; ii >= 0; ii--) {
                
            if (cmap == 16) {
            /*
             * Convert 16-bit colormap to 8-bit.
             */
#define    CVT(x)        (((x) * 255) / ((1L<<16)-1))
                tiff_img->rmap[ii] = CVT(rmap[ii]);
                tiff_img->gmap[ii] = CVT(gmap[ii]);
                tiff_img->bmap[ii] = CVT(bmap[ii]);
#undef CVT
            } else {
                tiff_img->rmap[ii] = rmap[ii];
                tiff_img->gmap[ii] = gmap[ii];
                tiff_img->bmap[ii] = bmap[ii];
            }
            
        }
          break ;
      case PHOTOMETRIC_RGB:
          tiff_img->TIFF_type = 'c' ;
          if(samplesperpixel != 3) {
          fprintf(stderr,"Wrong samplesperpixel %d for RGB image\n",
              samplesperpixel);
          return (FALSE);
        }
          break ;
      case PHOTOMETRIC_YCBCR:
          /******************************************************************/
          /* IMPORTANT:  */
          /* (1) If the image is JPEG compressed, it will be upsample and  */
          /*     change into RGB.  Unless user modifies the following code */
          /*             if(compress ....) {
          /*            .....
          /*        } 
          /* (2) Currently, it can not read a YCbYr image if the subsampling
          /*     rates are not 1
          /******************************************************************/
          tiff_img->TIFF_type = 'y' ;
          TIFFGetField(in, TIFFTAG_COMPRESSION, &compress);
          TIFFGetField(in, TIFFTAG_PLANARCONFIG, &config);
          TIFFGetField(in, TIFFTAG_YCBCRSUBSAMPLING,
                &horizSubSampling, &vertSubSampling);
          tiff_img->horizSubSampling = horizSubSampling ;
          tiff_img->vertSubSampling = vertSubSampling ;
          if (compress != COMPRESSION_JPEG &&
              (horizSubSampling != 1  && vertSubSampling !=1) ) {
              fprintf(stderr, "Does not support YCbCr image with subsmapling rates");
              fprintf(stderr, " other than 1, unless it is JPEG compressed \n") ;
              return(FALSE) ;
          }
          if (compress == COMPRESSION_JPEG && config == PLANARCONFIG_CONTIG) {
                /* can rely on libjpeg to convert to RGB */
                  /* XXX should restore current state on exit */
            TIFFSetField(in, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB);
            tiff_img->TIFF_type = 'c' ;
          }
          if(samplesperpixel != 3) {
            fprintf(stderr,"Wrong samplesperpixel %d for YCbCr image\n",
              samplesperpixel);
            return (FALSE);
          }
          break ;
      case PHOTOMETRIC_CIELAB:
          if(samplesperpixel == 1) {
                tiff_img->TIFF_type = 'l' ;
          } else if(samplesperpixel == 3) {
                tiff_img->TIFF_type = 'L' ;
          } else {
          fprintf(stderr,"Wrong samplesperpixel %d for CIE La*b* image\n",
              samplesperpixel);
          return (FALSE);
        }    
          break ;
      default:
        fprintf(stderr, "The photometric %d is not supported.\n", 
            photometric);
        return (FALSE);
    }
    
    return(TRUE) ;
}


/******************************************************************************/
int  
Read_TIFF_Img_Header(TIFF* in, TIFF_img *tiff_img) 
{
    uint32  w,l;
    int    cmap, ii ;
    uint16 *rmap, *gmap, *bmap ;
    uint16 config;
    uint16 compress;
    uint16 horizSubSampling;    /* YCbCr horizontal subsampling */
    uint16 vertSubSampling;        /* YCbCr vertical subsampling */
    uint16 bitspersample;
    uint16 samplesperpixel;
    uint16 photometric;
    uint16 fillorder;
    
    TIFFGetField(in, TIFFTAG_BITSPERSAMPLE, &(tiff_img->bitspersample));
    TIFFGetField(in, TIFFTAG_PLANARCONFIG, &(tiff_img->planarConfig));

    TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &(tiff_img->width));
    TIFFGetField(in, TIFFTAG_IMAGELENGTH, &(tiff_img->height));

    TIFFGetField(in, TIFFTAG_FILLORDER, &(tiff_img->fillorder));
    TIFFGetField(in, TIFFTAG_SAMPLESPERPIXEL, &(tiff_img->samplesperpixel));
    
    photometric = -1 ;
    TIFFGetField(in, TIFFTAG_PHOTOMETRIC, &photometric);
    tiff_img->photometric = photometric ;
    switch ( photometric ) {
      case PHOTOMETRIC_MINISWHITE:
          fprintf(stdout, "\t0 is white !\n") ;
          tiff_img->TIFF_type = 'g';
          if( tiff_img->bitspersample == 1) tiff_img->TIFF_type = 'b';
          if( tiff_img->samplesperpixel != 1) {
              fprintf(stderr, "Wrong samplesperpixel %d for monochrome image\n",
              tiff_img->samplesperpixel);
              return(FALSE) ;
          }
          break ;
      case PHOTOMETRIC_MINISBLACK:
          fprintf(stdout, "\t0 is black !\n") ;
          tiff_img->TIFF_type = 'g';
          if( tiff_img->bitspersample == 1) tiff_img->TIFF_type = 'b';
          if( tiff_img->samplesperpixel != 1 ) {
          fprintf(stderr, "Wrong samplesperpixel %d for monochrome image\n",
              tiff_img->samplesperpixel);
          return(FALSE) ;
        }
          break ;
      case PHOTOMETRIC_PALETTE:
          tiff_img->TIFF_type = 'p' ;
          if(tiff_img->samplesperpixel !=1 ) {
            fprintf(stderr,"Wrong samplesperpixel %d for RGB image\n",
              tiff_img->samplesperpixel);
            return(FALSE) ;
          }
          break ;
      case PHOTOMETRIC_RGB:
          tiff_img->TIFF_type = 'c' ;
          if(tiff_img->samplesperpixel != 3) {
          fprintf(stderr,"Wrong samplesperpixel %d for RGB image\n",
              tiff_img->samplesperpixel);
          return (FALSE);
        }
          break ;
      case PHOTOMETRIC_YCBCR:
          /******************************************************************/
          /* IMPORTANT:  */
          /* (1) If the image is JPEG compressed, it will be upsample and  */
          /*     change into RGB.  Unless user modifies the following code */
          /*             if(compress ....) {
          /*            .....
          /*        } 
          /* (2) Currently, it can not read a YCbYr image if the subsampling
          /*     rates are not 1
          /******************************************************************/
          tiff_img->TIFF_type = 'y' ;
          TIFFGetField(in, TIFFTAG_COMPRESSION, &compress);
          TIFFGetField(in, TIFFTAG_PLANARCONFIG, &config);
          TIFFGetField(in, TIFFTAG_YCBCRSUBSAMPLING,
                &horizSubSampling, &vertSubSampling);
          tiff_img->horizSubSampling = horizSubSampling ;
          tiff_img->vertSubSampling = vertSubSampling ;
          if (compress != COMPRESSION_JPEG &&
              (horizSubSampling != 1  && vertSubSampling !=1) ) {
              fprintf(stderr, "Does not support YCbCr image with subsmapling rates");
              fprintf(stderr, " other than 1, unless it is JPEG compressed \n") ;
              return(FALSE) ;
          }
          if (compress == COMPRESSION_JPEG && config == PLANARCONFIG_CONTIG) {
                /* can rely on libjpeg to convert to RGB */
                  /* XXX should restore current state on exit */
            TIFFSetField(in, TIFFTAG_JPEGCOLORMODE, JPEGCOLORMODE_RGB);
            tiff_img->TIFF_type = 'c' ;
          }
          if(tiff_img->samplesperpixel != 3) {
            fprintf(stderr,"Wrong samplesperpixel %d for YCbCr image\n",
              tiff_img->samplesperpixel);
            return (FALSE);
          }
          break ;
      case PHOTOMETRIC_CIELAB:
          if(tiff_img->samplesperpixel == 1) {
                tiff_img->TIFF_type = 'l' ;
          } else if(tiff_img->samplesperpixel == 3) {
                tiff_img->TIFF_type = 'L' ;
          } else {
          fprintf(stderr,"Wrong samplesperpixel %d for CIE La*b* image\n",
              tiff_img->samplesperpixel);
          return (FALSE);
        }    
          break ;
      default:
        fprintf(stderr, "The photometric %d is not supported.\n", 
            photometric);
        return (FALSE);
    }
    
    return(TRUE) ;
}


/******************************************************************************/
static int 
TIFF_Read_Contig_Strips(TIFF* in, TIFF_img *tiff_img) 
{
    int i ;
    uint32 w, l;
    uint32 rowsperstrip_in;
    tsize_t stripsize_in ;
    tdata_t buf_in ;
    uint8  **mono, ***color ;
    uint8  *InBuf ;
    register uint8 ch, mask ;
    
    mono = tiff_img->mono ;
    color = tiff_img->color ;
    TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(in, TIFFTAG_IMAGELENGTH, &l);
    TIFFGetField(in, TIFFTAG_ROWSPERSTRIP, &rowsperstrip_in);
    stripsize_in  = TIFFStripSize(in);
    buf_in = _TIFFmalloc(stripsize_in);
    
    /* printf("stripsize:  in %d\n", stripsize_in) ;  */
            
    if (buf_in) {
        tstrip_t s, ns = TIFFNumberOfStrips(in);
        uint32 row_in = 0 ;
        uint32 index_row = 0, index_col = 0, j=0 ;
        
        for (s = 0; s < ns; s++) {
            tsize_t cc_in = (row_in + rowsperstrip_in > l) ?
                TIFFVStripSize(in, l - row_in) : stripsize_in;
                
            
            if (TIFFReadEncodedStrip(in, s, buf_in, cc_in) < 0 && !ignore) {
                fprintf(stderr, "error reading strip %d\n", s) ;
                return(FALSE) ;
                break;
            }
                
            InBuf = (uint8*) buf_in ;
            
            switch ( tiff_img->samplesperpixel ) {
              case 1:
                  for(i=1; i<=cc_in; i++) {
                      if( tiff_img->bitspersample == 8) {
                      mono[index_row][index_col] = (uint8) (*InBuf++) ;
                          index_col ++ ;
                          if( index_col >= w ) {
                            index_col = 0 ;
                            index_row ++ ;
                          }
                          
                        } else if( tiff_img->bitspersample == 1) {    
                           /* bilevel image */
                           /* default fillorder is FILLORDER_MSB2LSB */
                           if( tiff_img->fillorder != FILLORDER_LSB2MSB ) {
                               ch = (uint8) (*InBuf++) ;
                               mask = 128 ;
                               for(j=0; j<8; j++) {
                                   mono[index_row][index_col] = 
                                       (ch & mask) >> (7-j) ;
                                   mask >>= 1 ;
                                   index_col ++ ;
                                   if( index_col >=w ) {
                                       index_col = 0 ;
                                       index_row ++ ;
                                       break ;
                                   }
                               }
                           } else {
                               ch = (uint8) (*InBuf++) ;
                               mask = 1 ;
                               for(j=0; j<8; j++) {
                                   mono[index_row][index_col] = 
                                       (ch & mask) >> j ;
                                   mask <<= 1 ;
                                   index_col ++ ;
                                   if( index_col >=w ) {
                                       index_col = 0 ;
                                       index_row ++ ;
                                       break ;
                                   }
                               }
                           }
                        } else return(FALSE) ;
                }
                  break ;
              case 3:
                  for(i=3; i<=cc_in; i+=3) {    /* samplesperpixel==3 */
                  color[0][index_row][index_col] = (uint8) (*InBuf++) ;
                  color[1][index_row][index_col] = (uint8) (*InBuf++) ;
                  color[2][index_row][index_col] = (uint8) (*InBuf++) ;
                      index_col ++ ;
                      if( index_col >= w ) {
                        index_col = 0 ;
                        index_row ++ ;
                      }
                  }
                  break ;
              default:
                  return(FALSE) ;
                }
                
               
                if((index_row > l) || (index_row==l && index_col > 0)){
                    fprintf(stderr, "Read rows %d > image height %d", index_row, l) ;
                    return(FALSE) ;
                }
            row_in += rowsperstrip_in;
          }
        }

/*
TIFFGetField(in, TIFFTAG_PHOTOMETRIC, &photometric);
printf("photometric after decoding %d\n", photometric) ;
photometric = -1 ;
TIFFGetField(in, TIFFTAG_JPEGCOLORMODE, &photometric);
printf("JPEG color mode after decoding %d\n", photometric) ;
*/
        _TIFFfree(buf_in);
        return (TRUE);

}


/******************************************************************************/
void
PrintTIFF_ImgInfo(TIFF_img * tiff_img) 
{
    switch ( tiff_img->TIFF_type ) {
      case 'b':
          fprintf(stdout, "\t\tBilevel image, ") ;
          break ;
      case 'g':
          fprintf(stdout, "\t\tMonochrome image, ") ;
          break ;
      case 'c':
          fprintf(stdout, "\t\tRGB image, ") ;
          break ;
      case 'y':
          fprintf(stdout, "\t\tYCbCr image, ") ;
          break ;
      case 'l':
          fprintf(stdout, "\t\tCIE L*a*b* monochrome image ") ;
          break ;
      case 'L':
          fprintf(stdout, "\t\tCIE L*a*b* full color image ") ;
          break ;
      case 'p':
          fprintf(stdout, "\t\tPalette-color image ") ;
          break ;
      default:
          fprintf(stderr, "\t\tWrong TIFF_img type %c \n", tiff_img->TIFF_type);
          exit(-1);
    }
    
    fprintf(stdout, "height %d, width %d\n", tiff_img->height, tiff_img->width);
    fprintf(stdout, "bitspersample %d,  samplesperpixel %d\n", 
        tiff_img->bitspersample, tiff_img->samplesperpixel);
}


/******************************************************************************/
void Free_TIFF ( TIFF_img *img )
{
  int i;

  switch( img->TIFF_type ) {
      case 'b' :
      case 'g' :    /* grayscale */
      case 'l' :
            if(img->mono != NULL)
                tiff_free_img ( ( void ** ) ( img->mono ) );
            break ;
      case 'p' :     /* palette-color */
            if(img->mono != NULL) 
                tiff_free_img ( ( void ** ) ( img->mono ) );
            if(img->rmap != NULL) 
                free ( ( void * ) ( img->rmap ) );
            if(img->gmap != NULL) 
                free ( ( void * ) ( img->gmap ) );
            if(img->bmap != NULL) 
                free ( ( void * ) ( img->bmap ) );
        break ;
      case 'c' :    /* full color */
      case 'y' :
      case 'L' :
          if(img->color != NULL) {
              for ( i = 0; i < 3; i++ ) {
                  if(img->color[i]!= NULL) 
                  tiff_free_img ( ( void ** ) ( img->color[i] ) );
              }
              free( (void *) img->color ) ;
          }
          break ;
      default :
          fprintf(stderr, "ERROR: unkown TIFF_type %d\n", img->TIFF_type) ;
          exit(-1) ;
  }
}



/******************************************************************************/
TIFF_img* 
New_TIFF_Img(char* name) 
{
    TIFF_img *tiffimg ;
    
    tiffimg = (TIFF_img *) tiff_get_spc(1, sizeof(TIFF_img)) ;
    
    if(tiffimg==NULL) {
        fprintf(stderr, "Can not generate new TIFF %s\n", name);
        exit(-1) ;
    }
    
    /* optinal parameters,  default values are set               */
    tiffimg->bitspersample =8;
    tiffimg->photometric = PHOTOMETRIC_MINISBLACK;    /* default for    */ 
                                                      /* binary img */
    tiffimg->planarConfig = PLANARCONFIG_CONTIG ;
    tiffimg->samplesperpixel = 0;                     /* no default   */
    tiffimg->IsTile = FALSE;
    tiffimg->rowsperstrip = -1;
    tiffimg->tilewidth = 0;
    tiffimg->tildheight = 0;
    tiffimg->XResolution = 0;        /* number of pixels per */
                                     /* ResolutionUnit */
    tiffimg->YResolution = 0;        /* number of pixels per   */
                                     /* ResolutionUnit         */
    tiffimg->ResolutionUnit = RESUNIT_NONE;    /* absolute unit of meas. */
    tiffimg->fillorder = 0;                    /*               */
    
    return(tiffimg) ;
}







/******************************************************************************/
int 
Allocate_New_TIFF(TIFF_img *tiff_img)
/* This routine is used to allocate momery for a new TIFF. */
/* User must supply tiff_img->width, tiff_img->height,   */
/* tiff_img->TIFF_type, tiff_img->bitspersample */
{
    uint32 w,l;
    int i,k ;
    uint16 samplesperpixel;
    
    
    if( (tiff_img->bitspersample != 8) && (tiff_img->bitspersample != 1) ){
        fprintf(stderr,"Sorry, only support 8-bit or 1-bit sample.\n");
        return(FALSE) ;
    }
    
    if( (tiff_img->bitspersample == 1) && (tiff_img->TIFF_type != 'b') ) {
        fprintf(stderr,"Sorry, 1-bit sample must be bilevel image.\n");
        return(FALSE) ;
    }
    
    w = tiff_img->width ;
    l = tiff_img->height ;
    if( (w<=0) || (l<=0)) {
        fprintf(stderr, "Wrong dimension w=%d, l=%d\n", w,l) ;
        return(FALSE) ;
    }
    
    switch ( tiff_img->TIFF_type ) {
      case 'p':
          k = (1<<(tiff_img->bitspersample)) ;    /* 2^bitspersample */
          tiff_img->rmap = (uint16 *) _TIFFmalloc(k*sizeof(uint16)) ;
          tiff_img->gmap = (uint16 *) _TIFFmalloc(k*sizeof(uint16)) ;
          tiff_img->bmap = (uint16 *) _TIFFmalloc(k*sizeof(uint16)) ;
          if( (tiff_img->rmap == NULL) || 
              (tiff_img->gmap == NULL) ||
              (tiff_img->bmap == NULL) ) {
            fprintf(stderr, "Can not allocate momery!  Allocate_New_TIFF 3\n");
            exit(-1) ;
          }
          /* Has to be initialied.  ow. may cause confusion between */
          /* 8-bit and 16-bit color map */
          for(i=0; i<k; i++) {    
              tiff_img->rmap[i] = 0 ;
              tiff_img->gmap[i] = 0 ;
              tiff_img->bmap[i] = 0 ;
          }
          /* NO break */
      case 'b':
      case 'g':
      case 'l':
          tiff_img->mono = (uint8 **) tiff_get_img(w, l, sizeof(uint8)) ;
          tiff_img->luminance = tiff_img->mono ;
          tiff_img->samplesperpixel = 1 ;
          break ;
      case 'c':
      case 'y':
      case 'L':
          tiff_img->samplesperpixel = samplesperpixel = 3 ;
          tiff_img->color = 
              (uint8 ***) tiff_get_spc(samplesperpixel, sizeof(uint8 **)) ;
          for(i=0; i<samplesperpixel; i++) {
              tiff_img->color[i] = (uint8 **) tiff_get_img(w, l, sizeof(uint8)) ;
          }
          tiff_img->chrominance = (int8 ***) 
          tiff_get_spc(samplesperpixel-1, sizeof(uint8 **));
          for(i=0; i<samplesperpixel-1; i++) {
              tiff_img->chrominance[i] = (int8 **) tiff_img->color[i+1] ;
          }
          tiff_img->luminance = tiff_img->mono = (uint8 **) tiff_img->color[0] ;
          break ;
      default:
          fprintf(stderr, "Wrong TIFF_img type %c \n", tiff_img->TIFF_type);
          return(FALSE);
    }
    
    return(TRUE) ;
}




/******************************************************************************/
static void
processG3Options(char* cp)
{
    if (cp = strchr(cp, ':')) {
        if (defg3opts == (uint32) -1)
            defg3opts = 0;
        do {
            cp++;
            if (strneq(cp, "1d", 2))
                defg3opts &= ~GROUP3OPT_2DENCODING;
            else if (strneq(cp, "2d", 2))
                defg3opts |= GROUP3OPT_2DENCODING;
            else if (strneq(cp, "fill", 4))
                defg3opts |= GROUP3OPT_FILLBITS;
            else
                usage();
        } while (cp = strchr(cp, ':'));
    }
}


/******************************************************************************/
static int
processCompressOptions(char* opt)
{
    if (strneq(opt, "none", 4))
        defcompression = COMPRESSION_NONE;
    else if (strneq(opt, "packbits", 8))
        defcompression = COMPRESSION_PACKBITS;
    else if (strneq(opt, "jpeg", 4)) {
        char* cp = strchr(opt, ':');
        if (cp && isdigit(cp[1]))
            quality = atoi(cp+1);
        if (cp && strchr(cp, 'r'))
            jpegcolormode = JPEGCOLORMODE_RAW;
        defcompression = COMPRESSION_JPEG;
    } else if (strneq(opt, "g3", 2)) {
        processG3Options(opt);
        defcompression = COMPRESSION_CCITTFAX3;
    } else if (strneq(opt, "g4", 2))
        defcompression = COMPRESSION_CCITTFAX4;
    else if (strneq(opt, "lzw", 3)) {
        char* cp = strchr(opt, ':');
        if (cp)
            defpredictor = atoi(cp+1);
        defcompression = COMPRESSION_LZW;
    } else if (strneq(opt, "zip", 3)) {
        char* cp = strchr(opt, ':');
        if (cp)
            defpredictor = atoi(cp+1);
        defcompression = COMPRESSION_DEFLATE;
    } else {
        return (0);
    }
    return (1);
}


/******************************************************************************/
#define    CopyField(tag, v) \
    if (TIFFGetField(in, tag, &v)) TIFFSetField(out, tag, v)
#define    CopyField2(tag, v1, v2) \
    if (TIFFGetField(in, tag, &v1, &v2)) TIFFSetField(out, tag, v1, v2)
#define    CopyField3(tag, v1, v2, v3) \
    if (TIFFGetField(in, tag, &v1, &v2, &v3)) \
        TIFFSetField(out, tag, v1, v2, v3)
#define    CopyField4(tag, v1, v2, v3, v4) \
    if (TIFFGetField(in, tag, &v1, &v2, &v3, &v4)) \
        TIFFSetField(out, tag, v1, v2, v3, v4)

static void
cpTag(TIFF* in, TIFF* out, uint16 tag, uint16 count, TIFFDataType type)
{
    switch (type) {
    case TIFF_SHORT:
        if (count == 1) {
            uint16 shortv;
            CopyField(tag, shortv);
        } else if (count == 2) {
            uint16 shortv1, shortv2;
            CopyField2(tag, shortv1, shortv2);
        } else if (count == 4) {
            uint16 *tr, *tg, *tb, *ta;
            CopyField4(tag, tr, tg, tb, ta);
        } else if (count == (uint16) -1) {
            uint16 shortv1;
            uint16* shortav;
            CopyField2(tag, shortv1, shortav);
        }
        break;
    case TIFF_LONG:
        { uint32 longv;
          CopyField(tag, longv);
        }
        break;
    case TIFF_RATIONAL:
        if (count == 1) {
            float floatv;
            CopyField(tag, floatv);
        } else if (count == (uint16) -1) {
            float* floatav;
            CopyField(tag, floatav);
        }
        break;
    case TIFF_ASCII:
        { char* stringv;
          CopyField(tag, stringv);
        }
        break;
    case TIFF_DOUBLE:
        if (count == 1) {
            double doublev;
            CopyField(tag, doublev);
        } else if (count == (uint16) -1) {
            double* doubleav;
            CopyField(tag, doubleav);
        }
        break;
    }
}

static struct cpTag {
    uint16    tag;
    uint16    count;
    TIFFDataType type;
} tags[] = {
    { TIFFTAG_SUBFILETYPE,        1, TIFF_LONG },
    { TIFFTAG_THRESHHOLDING,    1, TIFF_SHORT },
    { TIFFTAG_DOCUMENTNAME,        1, TIFF_ASCII },
    { TIFFTAG_IMAGEDESCRIPTION,    1, TIFF_ASCII },
    { TIFFTAG_MAKE,            1, TIFF_ASCII },
    { TIFFTAG_MODEL,        1, TIFF_ASCII },
    { TIFFTAG_ORIENTATION,        1, TIFF_SHORT },
    { TIFFTAG_MINSAMPLEVALUE,    1, TIFF_SHORT },
    { TIFFTAG_MAXSAMPLEVALUE,    1, TIFF_SHORT },
    { TIFFTAG_XRESOLUTION,        1, TIFF_RATIONAL },
    { TIFFTAG_YRESOLUTION,        1, TIFF_RATIONAL },
    { TIFFTAG_PAGENAME,        1, TIFF_ASCII },
    { TIFFTAG_XPOSITION,        1, TIFF_RATIONAL },
    { TIFFTAG_YPOSITION,        1, TIFF_RATIONAL },
    { TIFFTAG_RESOLUTIONUNIT,    1, TIFF_SHORT },
    { TIFFTAG_PAGENUMBER,        2, TIFF_SHORT },
    { TIFFTAG_SOFTWARE,        1, TIFF_ASCII },
    { TIFFTAG_DATETIME,        1, TIFF_ASCII },
    { TIFFTAG_ARTIST,        1, TIFF_ASCII },
    { TIFFTAG_HOSTCOMPUTER,        1, TIFF_ASCII },
    { TIFFTAG_WHITEPOINT,        1, TIFF_RATIONAL },
    { TIFFTAG_PRIMARYCHROMATICITIES,(uint16) -1,TIFF_RATIONAL },
    { TIFFTAG_HALFTONEHINTS,    2, TIFF_SHORT },
    { TIFFTAG_INKSET,        1, TIFF_SHORT },
    { TIFFTAG_INKNAMES,        1, TIFF_ASCII },
    { TIFFTAG_DOTRANGE,        2, TIFF_SHORT },
    { TIFFTAG_TARGETPRINTER,    1, TIFF_ASCII },
    { TIFFTAG_SAMPLEFORMAT,        1, TIFF_SHORT },
    { TIFFTAG_YCBCRCOEFFICIENTS,    (uint16) -1,TIFF_RATIONAL },
    { TIFFTAG_YCBCRSUBSAMPLING,    2, TIFF_SHORT },
    { TIFFTAG_YCBCRPOSITIONING,    1, TIFF_SHORT },
    { TIFFTAG_REFERENCEBLACKWHITE,    (uint16) -1,TIFF_RATIONAL },
    { TIFFTAG_EXTRASAMPLES,        (uint16) -1, TIFF_SHORT },
    { TIFFTAG_SMINSAMPLEVALUE,    1, TIFF_DOUBLE },
    { TIFFTAG_SMAXSAMPLEVALUE,    1, TIFF_DOUBLE },
};
#define    NTAGS    (sizeof (tags) / sizeof (tags[0]))

#define    CopyTag(tag, count, type)    cpTag(in, out, tag, count, type)

typedef int (*copyFunc)
    (TIFF* in, TIFF* out, uint32 l, uint32 w, uint16 samplesperpixel);
static    copyFunc pickCopyFunc(TIFF*, TIFF*, uint16, uint16);


/******************************************************************************/
static void
CheckAndCorrectColormap(int n, uint16* r, uint16* g, uint16* b)
{
    int i;

    for (i = 0; i < n; i++) {
        if (r[i] >= 256 || g[i] >= 256 || b[i] >= 256) {
            fprintf(stdout, "\t16-bit colormap \n") ;
            fprintf(stdout, "%d r%d,  g%d,  b%d\n",i, r[i], g[i], b[i]) ;
            return;
        }
    }
    
    /* fprintf(stderr, "Warning: Scaling 8-bit colormap to 16-bit\n"); */
#define    CVT(x)        (((x) * ((1L<<16)-1)) / 255)
    for (i = 0; i < n; i++) {
        r[i] = CVT(r[i]);
        g[i] = CVT(g[i]);
        b[i] = CVT(b[i]);
    }
#undef CVT
}


/******************************************************************************/
int
Initial_TIFF_Tags(TIFF *out, TIFF_img *tiff_img, TIFF *in)
/* If TIFF *in != NULL,  copy tags from in to output TIFF img */
/* If TIFF *in == NULL,  set tages which are not specified by */
/* tiff_img to default values */
{
    uint16 compression ;
    uint16 defconfig = (uint16) -1;
    uint32 rowsperstrip;
    uint32 deftilewidth = (uint32) -1;
    uint32 deftilelength = (uint32) -1;
    uint32 diroff = 0;
    struct cpTag* p;
    uint16 horizSubSampling = 1 ;     /* TIFF default values are 2 */
    uint16 vertSubSampling = 1 ;    /* Currently only can read when they are 1 */

    TIFFSetField(out, TIFFTAG_IMAGEWIDTH, tiff_img->width);
    TIFFSetField(out, TIFFTAG_IMAGELENGTH, tiff_img->height);
    
    if( (tiff_img->bitspersample != 8) && (tiff_img->bitspersample != 1) ){
        fprintf(stderr,"Sorry, only support 8-bit or 1-bit sample.\n");
        return(FALSE) ;
    }
    
    if( (tiff_img->bitspersample == 1) && (tiff_img->TIFF_type != 'b') ) {
        fprintf(stderr,"Sorry, 1-bit sample must be bilevel image.\n");
        return(FALSE) ;
    }
    TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, tiff_img->bitspersample);
    
    if( tiff_img->planarConfig != PLANARCONFIG_CONTIG ) {
        fprintf(stderr, "Sorry, does not support separate color planes.\n") ;
        return(FALSE) ;
    }
        TIFFSetField(out, TIFFTAG_PLANARCONFIG, tiff_img->planarConfig);
    
    if( tiff_img->IsTile ) {
        fprintf(stderr, "Sorry, image must be written out in strips.\n") ;
        return(FALSE) ;
    }
    
    switch ( tiff_img->TIFF_type ) {
      case 'b':
      case 'g':    /* monochrome image */
        TIFFSetField(out, TIFFTAG_PHOTOMETRIC, tiff_img->photometric);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 1);
        tiff_img->samplesperpixel = 1 ;
          break;
      case 'p':       /* color map indexed */
          TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_PALETTE);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 1);
        tiff_img->samplesperpixel = 1 ;
        CheckAndCorrectColormap( 1<<(tiff_img->bitspersample), 
                tiff_img->rmap, tiff_img->gmap, tiff_img->bmap);
            TIFFSetField(out, TIFFTAG_COLORMAP, tiff_img->rmap, 
                tiff_img->gmap, tiff_img->bmap);
            break;
          case 'c':    /* RGB full color image */
          TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 3);
        tiff_img->samplesperpixel = 3 ;
          break;
      case 'y':    /* YCbCr full color image */
          TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_YCBCR);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 3);
        TIFFSetField(out, TIFFTAG_YCBCRSUBSAMPLING,
            1, 1);
        tiff_img->samplesperpixel = 3 ;
        /* Now only support when subsampling rates are 1
            horizSubSampling = tiff_img->horizSubSampling;
            vertSubSampling = tiff_img->vertSubSampling;
            */
          break;
      case 'l':    /* CIE L*a*b* monochrone image */
          TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_CIELAB);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 1);
        tiff_img->samplesperpixel = 1 ;
          break;
      case 'L':    /* CIE L*a*b* full color image */
          TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_CIELAB);
        TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 3);
        tiff_img->samplesperpixel = 3 ;
          break;
      default:
          fprintf(stderr, "Unknown image type %c. Pleaser refer to") ; 
          fprintf(stderr, "tiffImg.h for valid image types\n", 
              tiff_img->TIFF_type);
          return(FALSE);
    }
    
    

    if (!processCompressOptions(tiff_img->compress_type)) {
        if( in==NULL ) {
            fprintf(stderr, "Unknown compression type %s.\n", 
                tiff_img->compress_type);
            fprintf(stderr, "Please refer to tiffImg.h for valid types\n");
            fprintf(stdout, "Set compression to No-Compression ! \n") ;
            defcompression = COMPRESSION_NONE ;
        } 
    }
    compression = defcompression;
    predictor = defpredictor;
    g3opts = defg3opts;
    if (compression != (uint16)-1)
        TIFFSetField(out, TIFFTAG_COMPRESSION, compression);
    else if ( in != NULL )
        CopyField(TIFFTAG_COMPRESSION, compression);
    else TIFFSetField(out, TIFFTAG_COMPRESSION, COMPRESSION_NONE);

    if (compression == COMPRESSION_JPEG && jpegcolormode == JPEGCOLORMODE_RGB)
        TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_YCBCR);
    switch (compression) {
      case COMPRESSION_JPEG:
        TIFFSetField(out, TIFFTAG_JPEGQUALITY, quality);
        TIFFSetField(out, TIFFTAG_JPEGCOLORMODE, jpegcolormode);
        break;
      case COMPRESSION_LZW:
      case COMPRESSION_DEFLATE:
        if (predictor != (uint16)-1)
            TIFFSetField(out, TIFFTAG_PREDICTOR, predictor);
        else
            CopyField(TIFFTAG_PREDICTOR, predictor);
        break;
      case COMPRESSION_CCITTFAX3:
      case COMPRESSION_CCITTFAX4:
        if (compression == COMPRESSION_CCITTFAX3) {
            if (g3opts != (uint32) -1)
                TIFFSetField(out, TIFFTAG_GROUP3OPTIONS,
                    g3opts);
            else
                CopyField(TIFFTAG_GROUP3OPTIONS, g3opts);
        } else
            CopyTag(TIFFTAG_GROUP4OPTIONS, 1, TIFF_LONG);
        CopyTag(TIFFTAG_BADFAXLINES, 1, TIFF_LONG);
        CopyTag(TIFFTAG_CLEANFAXDATA, 1, TIFF_LONG);
        CopyTag(TIFFTAG_CONSECUTIVEBADFAXLINES, 1, TIFF_LONG);
        CopyTag(TIFFTAG_FAXRECVPARAMS, 1, TIFF_LONG);
        CopyTag(TIFFTAG_FAXRECVTIME, 1, TIFF_LONG);
        CopyTag(TIFFTAG_FAXSUBADDRESS, 1, TIFF_ASCII);
        break;
      default:
          break;
    }


    if ( (tiff_img->fillorder == 0) && (in != NULL) )
        CopyTag(TIFFTAG_FILLORDER, 1, TIFF_SHORT);
    else if( (tiff_img->fillorder == 0) && (in == NULL) )    /* default */
        TIFFSetField(out, TIFFTAG_FILLORDER, FILLORDER_MSB2LSB);
/*  else TIFFSetField(out, TIFFTAG_FILLORDER, fillorder);   */


    if ( in != NULL) {
        for (p = tags; p < &tags[NTAGS]; p++)
            CopyTag(p->tag, p->count, p->type);
    }
    
    rowsperstrip = tiff_img->rowsperstrip ;
    if ( (rowsperstrip == (uint32) -1) && (in != NULL) )
        TIFFGetField(in, TIFFTAG_ROWSPERSTRIP, &rowsperstrip);
    rowsperstrip = TIFFDefaultStripSize(out, rowsperstrip);
    TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, rowsperstrip);

    return (TRUE);
}


/******************************************************************************/
static int
Write_TIFF_Contig_Strips(TIFF* out, TIFF_img *tiff_img)
{
    int i ;
    uint32 w, l;
    uint32 rowsperstrip_out;
    tsize_t stripsize_out ;
    tdata_t buf_out = NULL ;
    uint8  **mono, ***color ;
    uint8  *OutBuf ;
    register uint8 ch, mask ;
    uint16 samplesperpixel;

    mono = tiff_img->mono ;
    color = tiff_img->color ;
    TIFFGetField(out, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(out, TIFFTAG_IMAGELENGTH, &l);
    TIFFGetField(out, TIFFTAG_ROWSPERSTRIP, &rowsperstrip_out);
    stripsize_out  = TIFFStripSize(out);
    buf_out = _TIFFmalloc(stripsize_out);
    
    printf("\t\tOutput image width %d, height %d \n", w,l) ;
                
    if (buf_out) {
        tstrip_t s, ns = TIFFNumberOfStrips(out);
        uint32 row_out = 0 ;
        uint32 index_row = 0, index_col = 0, j ;
        
        for (s = 0; s < ns; s++) {
            tsize_t cc_out = (row_out + rowsperstrip_out > l) ?
                TIFFVStripSize(out, l - row_out) : stripsize_out;
                
            OutBuf = (uint8*) buf_out ;
            
            switch ( tiff_img->samplesperpixel ) {
                case 1:
                    if( tiff_img->bitspersample == 8) {
                      for(i=1; i<=cc_out; i++) {    
                        (*OutBuf++) = mono[index_row][index_col] ;
                            index_col ++ ;
                            if( index_col >= w ) {
                            index_col = 0 ;
                            index_row ++ ;
                            }
                        }
                        } else if( tiff_img->bitspersample == 1) {
                            /*bilevel image */
                        /* default fillorder is FILLORDER_MSB2LSB */
                        if( tiff_img->fillorder != FILLORDER_LSB2MSB ) {
                          for(i=1; i<=cc_out; i++) {    
                                ch = 0 ;
                               mask = 128 ;
                               for(j=0; j<8; j++) {
                                   if( mono[index_row][index_col] >0) {
                                       ch = ch | mask ;
                                   }
                                   mask >>= 1 ;
                                   index_col ++ ;
                                   if( index_col >=w ) {
                                       index_col = 0 ;
                                       index_row ++ ;
                                       break ;
                                   }
                               }
                               (*OutBuf++) = ch ; 
                            } 
                        } else {
                          for(i=1; i<=cc_out; i++) {    
                                ch = 0 ;
                               mask = 1 ;
                               for(j=0; j<8; j++) {
                                   if( mono[index_row][index_col] >0) {
                                       ch = ch | mask ;
                                   }
                                   mask <<= 1 ;
                                   index_col ++ ;
                                   if( index_col >=w ) {
                                       index_col = 0 ;
                                       index_row ++ ;
                                       break ;
                                   }
                               }
                               (*OutBuf++) = ch ;    
                            }
                       } 
                        
                        } else return(FALSE) ;    
                  break ;
                case 3:
                  for(i=3; i<=cc_out; i+=3) {    /* samplesperpixel==3 */
                   (*OutBuf++) = color[0][index_row][index_col];
                   (*OutBuf++) = color[1][index_row][index_col];
                   (*OutBuf++) = color[2][index_row][index_col];
                       index_col ++ ;
                       if( index_col >= w ) {
                        index_col = 0 ;
                        index_row ++ ;
                       }
                  }
                  break ;
                default:
                    fprintf(stderr, "Samplesperpixel %d is not supported\n", 
                        tiff_img->samplesperpixel ) ;
                  goto bad ;
                }
                if((index_row > l) || (index_row==l && index_col > 0)){
                    fprintf(stderr, "Output rows %d > image height %d", 
                        index_row, l) ;
                    goto bad ;
                }
               
            if (TIFFWriteEncodedStrip(out, s, buf_out, cc_out) < 0) {
                fprintf(stderr, "TIFFWrite Fals\n") ;
                goto bad ;
            }
            row_out += rowsperstrip_out;
        
/*  fprintf(stdout, "w %d, h%d written out, row_out %d\n", 
        index_col, index_row, row_out) ; 
*/
        }
        
        if(buf_out != NULL) _TIFFfree(buf_out);
        return (TRUE);
    }
        
bad:    
    _TIFFfree(buf_out);
    return(FALSE) ;

}


static char* stuff[] = {
"where compression options are:",
"",
" -c lzw[:opts]    compress output with Lempel-Ziv & Welch encoding",
" -c zip[:opts]    compress output with deflate encoding",
" -c jpeg[:opts]compress output with JPEG encoding",
" -c packbits    compress output with packbits encoding",
" -c g3[:opts]    compress output with CCITT Group 3 encoding",
" -c g4        compress output with CCITT Group 4 encoding",
" -c none    use no compression algorithm on output",
"",
"Group 3 options:",
" 1d        use default CCITT Group 3 1D-encoding",
" 2d        use optional CCITT Group 3 2D-encoding",
" fill        byte-align EOL codes",
"For example, -c g3:2d:fill to get G3-2D-encoded data with byte-aligned EOLs",
"",
"JPEG options:",
" #        set compression quality level (0-100, default 75)",
" r        output color image as RGB rather than YCbCr",
"For example, -c jpeg:r:50 to get JPEG-encoded RGB data with 50% comp. quality",
"",
"LZW and deflate options:",
" #        set predictor value",
"For example, -c lzw:2 to get LZW-encoded data with horizontal differencing",
NULL
};


/******************************************************************************/
static void
usage(void)
{
    char buf[BUFSIZ];
    int i;

    setbuf(stderr, buf);
    for (i = 0; stuff[i] != NULL; i++)
        fprintf(stderr, "%s\n", stuff[i]);
    exit(-1);
}



/******************************************************************************/
TIFF_img *
Read_TIFF_Img(char *fname, int show)
{
    uint32 diroff = 0;
    TIFF* in;
    TIFF_img *tiff_img;

    in = TIFFOpen(fname, "r");
    if (in == NULL) {
        fprintf(stderr, "Read_TIFF error: can not open %s.",fname) ;
        exit(-3);
    }
    
    /* diroff is the offset of a directory when there are */
    /* more than one image in the input tiff file         */
    if (diroff != 0 && !TIFFSetSubDirectory(in, diroff)) {
        TIFFError(TIFFFileName(in),
               "Error, setting subdirectory at %#x", diroff);
        (void) TIFFClose(in);
        exit(-1);
    }
    
    tiff_img = New_TIFF_Img("") ;    /* allocate momery for struct TIFF_img  */
                                     /* and set default values for it    */
        
    if(! Initial_Read_TIFF(in, tiff_img) ) return((TIFF_img*) NULL) ;
    if(! TIFF_Read_Contig_Strips(in, tiff_img) ) return((TIFF_img*) NULL) ;
    if(show) PrintTIFF_ImgInfo( tiff_img ) ;
    (void) TIFFClose(in);

    return(tiff_img) ;
}



/******************************************************************************/
int
Write_TIFF_Img(char *fname, TIFF_img *tiff_img, char *org)
/* If org != NULL,  copy tags from org to output TIFF img */
/* If org == NULL,  set tages which are not specified by  */
/* tiff_img to default values */
{
    uint32 defrowsperstrip = (uint32) -1;
    uint32 diroff = 0;
    TIFF *out, *in=NULL ;

    if( org != NULL ) in = TIFFOpen(org, "r") ;
    /* need not to check if in is NULL */
    
    out = TIFFOpen(fname, "w");
    if (out == NULL) {
        fprintf(stderr, "Cannot open output file %s\n",fname);
        exit (-2);
    }

    tiff_img->rowsperstrip = defrowsperstrip;
    if (! Initial_TIFF_Tags(out, tiff_img, in) ) goto bad ;
    if (! Write_TIFF_Contig_Strips(out, tiff_img) ) goto bad ;
    if (! TIFFWriteDirectory(out) ) goto bad ;
    
    (void) TIFFClose(out);
    if(in != NULL) (void) TIFFClose(in) ;
    return (TRUE);

bad:    (void) TIFFClose(out);
    if(in != NULL) (void) TIFFClose(in) ;
    return (FALSE);
}



