#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "tiffio.h"
#include "tiffImg.h"
#include "color.h"

#define SHOW 1 
#define SILENT 0 

void error(char *name);

int main (int argc, char **argv) 
{
  TIFF_img *input_img, *output_img;
  int i,j, nrow, ncol ;
  float rf,gf,bf, Lf,Af,Bf ;
  unsigned char **l, **R, **G, **B ;
  int8 **a, **b ;
  
    
  if ( argc != 3 ) error( argv[0] );

  input_img = Read_TIFF_Img(argv[1], SHOW) ;
  if ( input_img == NULL ) {
    fprintf ( stderr, "error reading file %s\n", argv[1] );
    exit(1);
  }

  if ( input_img->TIFF_type != 'c' ) {
    fprintf(stderr, "Need a LAB image, not type %c\n", input_img->TIFF_type) ;
    goto byebye ;
  }
  	

  output_img = New_TIFF_Img("") ;
  if ( output_img == NULL ) {
    fprintf ( stderr, "error generating a new TIFF_img, %s\n", argv[1] );
    exit(1);
  }
  
  output_img->width = ncol = input_img->width ;
  output_img->height = nrow = input_img->height ;
  output_img->bitspersample = 8 ;
  output_img->TIFF_type = 'c' ;		/* RGB */
  sprintf(output_img->compress_type,"jpeg:20") ;

  
  sprintf(input_img->compress_type,"jpeg:20") ;
  if ( ! Write_TIFF_Img(argv[2], input_img, argv[1]) ) {
    fprintf ( stderr, "error writting file %s\n", argv[2] );
    exit(1);
  }


byebye:
  /* de-allocate space which was used for the images */
  if( input_img != NULL ) Free_TIFF ( input_img );
  if( output_img != NULL ) Free_TIFF ( output_img );
}


void error(char *name)
{
    printf("usage:  %s  infile.tiff  outfile.tiff\n",name);
    printf("this program reads in a 24-bit color TIFF image,\n");
    printf("pixel replicates it by a factor of 4,\n");
    printf("then writes out the resulting image.\n");
    exit(1);
}

