#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "tiffio.h"
#include "tiffImg.h"


/* find the difference of two TIFF_img */
/* the desired color space of difference image is the type */
/* of the first image */
TIFF_img *
tiffDiff(TIFF_img *tiff_img1, TIFF_img *tiff_img2)
{
    TIFF_img *tiff_diff ;
    char tiff_type1, tiff_type2, backup_type ;
    int i,j,k, w1,h1, w2,h2 ;
    int bitspersample1, bitspersample2 ;
    int samplesperpixel1, samplesperpixel2 ;
    char name[80]="tiffDiff()" ;
    uint8 **um1, **um2, **um3 ;
    uint8 ***uc1, ***uc2, ***uc3 ;
    int8  ***c1, ***c2, ***c3 ;
        
    w1 = tiff_img1->width ;
    h1 = tiff_img1->height ;
    w2 = tiff_img2->width ;
    h2 = tiff_img2->height ;
    if( (w1 != w2) || (h1 != h2) ) {
    	fprintf(stderr, "Dimen conflict w%d!=%d, h%d!=%d in %s\n",
    		w1,w2,h1,h2,name) ;
    	return(NULL) ;
    }
    
    bitspersample1 = tiff_img1->bitspersample ;
    bitspersample2 = tiff_img2->bitspersample ;
    if( bitspersample1 != bitspersample2 ) {
    	fprintf(stderr, "bitspersample conflict %d!=%d in %s\n",
    		bitspersample1, bitspersample2, name) ;
    	return(NULL) ;
    }
    
    samplesperpixel1 = tiff_img1->samplesperpixel ;
    samplesperpixel2 = tiff_img2->samplesperpixel ;
    if( samplesperpixel1 != samplesperpixel2 ) {
    	fprintf(stderr, "samplesperpixel conflict %d!=%d in %s\n",
    		samplesperpixel1, samplesperpixel2, name) ;
    	return(NULL) ;
    }
    
    tiff_type1 = tiff_img1->TIFF_type ;
    tiff_type2 = tiff_img2->TIFF_type ;
    if( tiff_type1 == 'p' || tiff_type2 == 'p' ) {
    	fprintf(stderr, "Currently, palette image is not supported in %s",
    		name) ;
    	return(NULL) ;
    }
    if( tiff_type1 == 'b' || tiff_type2 == 'b' ) {
    	fprintf(stderr, "Currently, binary image is not supported in %s",
    		name) ;
    	return(NULL) ;
    }
    
    if(tiff_type1 != tiff_type2) {
    	/* color space transformation */
    	backup_type = tiff_type2 ;
    	if( ! TIFF_Color_Transform(tiff_img2, tiff_type1) )
    	    return(NULL) ;
    } else backup_type = '!' ;
    
    
    tiff_diff = New_TIFF_Img("tiffDiff()") ;
    tiff_diff->TIFF_type = tiff_type1 ;
    tiff_diff->width = w1 ;
    tiff_diff->height = h1 ;
    tiff_diff->bitspersample = bitspersample1 ;
    if( ! Allocate_New_TIFF(tiff_diff) ) {
    	fprintf(stderr, "Can not allocate tiff_diff\n") ;
    	return(NULL) ;
    }
    
    switch( tiff_type1 ) {
      case 'g':
      case 'l':
        um1 = tiff_img1->mono ;
        um2 = tiff_img2->mono ;
        um3 = tiff_diff->mono ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) {
      	    um3[i][j] = (um1[i][j]-um2[i][j])/2 + 127 ;
      	}
      	break ;
      case 'c':
        uc1 = tiff_img1->color ;
        uc2 = tiff_img2->color ;
        uc3 = tiff_diff->color ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) 
      	for(k=0; k<samplesperpixel1; k++) {
      	    uc3[k][i][j] = (uc1[k][i][j]-uc2[k][i][j])/2 + 127 ;
      	}
      	break ;
      case 'L':
      case 'y':
        um1 = tiff_img1->mono ;
        um2 = tiff_img2->mono ;
        um3 = tiff_diff->mono ;
        c1 =  tiff_img1->chrominance ;
        c2 =  tiff_img2->chrominance ;
        c3 =  tiff_diff->chrominance ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) {
      	    um3[i][j] = ( (um1[i][j]-um2[i][j])/2 + 127 ) ;
      	    for(k=0; k<samplesperpixel1-1; k++) {
      	    	c3[k][i][j] = (c1[k][i][j]-c2[k][i][j]) / 2 ;
      	    }
      	}
      	break ;
      default:
        fprintf(stderr,"%c type of image not supported in %s\n",
        	tiff_type1, name) ;
        return(NULL) ;
    }
      
        
    if( backup_type != '!' ) {
    	if( ! TIFF_Color_Transform(tiff_img2, backup_type) )
    	    return(NULL) ;
    }
    
    return(tiff_diff) ;
}




/* find the difference of two TIFF_img */
/* the desired color space of difference image is the type */
/* of the first image */
TIFF_img *
tiffAbsDiff(TIFF_img *tiff_img1, TIFF_img *tiff_img2)
{
    TIFF_img *tiff_diff ;
    char tiff_type1, tiff_type2, backup_type ;
    int i,j,k, w1,h1, w2,h2 ;
    int bitspersample1, bitspersample2 ;
    int samplesperpixel1, samplesperpixel2 ;
    char name[80]="tiffDiff()" ;
    uint8 **um1, **um2, **um3 ;
    uint8 ***uc1, ***uc2, ***uc3 ;
    int8  ***c1, ***c2, ***c3 ;
        
    w1 = tiff_img1->width ;
    h1 = tiff_img1->height ;
    w2 = tiff_img2->width ;
    h2 = tiff_img2->height ;
    if( (w1 != w2) || (h1 != h2) ) {
    	fprintf(stderr, "Dimen conflict w%d!=%d, h%d!=%d in %s\n",
    		w1,w2,h1,h2,name) ;
    	return(NULL) ;
    }
    
    bitspersample1 = tiff_img1->bitspersample ;
    bitspersample2 = tiff_img2->bitspersample ;
    if( bitspersample1 != bitspersample2 ) {
    	fprintf(stderr, "bitspersample conflict %d!=%d in %s\n",
    		bitspersample1, bitspersample2, name) ;
    	return(NULL) ;
    }
    
    samplesperpixel1 = tiff_img1->samplesperpixel ;
    samplesperpixel2 = tiff_img2->samplesperpixel ;
    if( samplesperpixel1 != samplesperpixel2 ) {
    	fprintf(stderr, "samplesperpixel conflict %d!=%d in %s\n",
    		samplesperpixel1, samplesperpixel2, name) ;
    	return(NULL) ;
    }
    
    tiff_type1 = tiff_img1->TIFF_type ;
    tiff_type2 = tiff_img2->TIFF_type ;
    if( tiff_type1 == 'p' || tiff_type2 == 'p' ) {
    	fprintf(stderr, "Currently, palette image is not supported in %s",
    		name) ;
    	return(NULL) ;
    }
    if( tiff_type1 == 'b' || tiff_type2 == 'b' ) {
    	fprintf(stderr, "Currently, binary image is not supported in %s",
    		name) ;
    	return(NULL) ;
    }
    
    if(tiff_type1 != tiff_type2) {
    	/* color space transformation */
    	backup_type = tiff_type2 ;
    	if( ! TIFF_Color_Transform(tiff_img2, tiff_type1) )
    	    return(NULL) ;
    } else backup_type = '!' ;
    
    
    tiff_diff = New_TIFF_Img("tiffDiff()") ;
    tiff_diff->TIFF_type = tiff_type1 ;
    tiff_diff->width = w1 ;
    tiff_diff->height = h1 ;
    tiff_diff->bitspersample = bitspersample1 ;
    if( ! Allocate_New_TIFF(tiff_diff) ) {
    	fprintf(stderr, "Can not allocate tiff_diff\n") ;
    	return(NULL) ;
    }
    
    switch( tiff_type1 ) {
      case 'g':
      case 'l':
        um1 = tiff_img1->mono ;
        um2 = tiff_img2->mono ;
        um3 = tiff_diff->mono ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) {
      	    um3[i][j] = (uint8) abs((int) (um1[i][j]-um2[i][j])) ;
      	}
      	break ;
      case 'c':
        uc1 = tiff_img1->color ;
        uc2 = tiff_img2->color ;
        uc3 = tiff_diff->color ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) 
      	for(k=0; k<samplesperpixel1; k++) {
      	    uc3[k][i][j] = (uint8) abs((int) (uc1[k][i][j]-uc2[k][i][j])) ;
      	}
      	break ;
      case 'L':
      case 'y':
        um1 = tiff_img1->mono ;
        um2 = tiff_img2->mono ;
        um3 = tiff_diff->mono ;
        c1 =  tiff_img1->chrominance ;
        c2 =  tiff_img2->chrominance ;
        c3 =  tiff_diff->chrominance ;
      	for(i=0; i<h1; i++)
      	for(j=0; j<w1; j++) {
      	    um3[i][j] = ( (um1[i][j]-um2[i][j])/2 + 127 ) ;
      	    for(k=0; k<samplesperpixel1-1; k++) {
      	    	c3[k][i][j] = (uint8) abs((int) (c1[k][i][j]-c2[k][i][j])) ;
      	    }
      	}
      	break ;
      default:
        fprintf(stderr,"%c type of image not supported in %s\n",
        	tiff_type1, name) ;
        return(NULL) ;
    }
      
        
    if( backup_type != '!' ) {
    	if( ! TIFF_Color_Transform(tiff_img2, backup_type) )
    	    return(NULL) ;
    }
    
    return(tiff_diff) ;
}
