/* This version will optimatize the DataWeight and then
   output the samples for constructing decision tree
*/

#include "CPT.h"
/* #include <sys/getopt.h>*/

#include <stdio.h>
#include <unistd.h>

void error(char *name);
void RandInitial();
static int total_number_samples(double *total, char **hs_fn, int nFiles) ;

int vlevel ;


main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd ;
  struct ModelPar Par ;
  
  /* In order to optimize DataWeight,  all images must be processed once  */
  /* After that the level which needs to output samples must be segmented */
  /* again with the optimal DataWeight.  So, some results must be kept    */
  
  char *infn;                 /* input image file name */
  char *otfn;                 /* output segmentation file name */
  char **org_fn ;
  char **hs_fn ;
  int  iflag,oflag,Bflag,wflag,Gflag,Tflag;
  int  outputLevel, nFiles, nClasses, D;
  int  i ;
  int  Precision ;
  int  nNeighb;
  int  max_leaves ;
  int  max_samples ;
  float  groundTruth, rnd_ratio ;
  double total_samples ;    /* use double to prevent overflow */
  char comm[80] ;
  double RndRatio;          /* # of samples / # of pixels */
  double dataWeight ;
  float thr_match ;         /* save samples of CPT, only when enough class */
                            /* labels match the ground truth among parents */  
  
  
  /* default values */
  vlevel=1;
  Precision=50;
  RndRatio=0.20;
  max_leaves=100 ;
  groundTruth=0 ;
  thr_match=0.6 ;
  dataWeight=1 ;
  iflag=oflag=Bflag=wflag=Gflag=Tflag=0;
  
  while((c=getopt(argc,argv,"h,o:,i:,B:,v:,P:,M:,T:"))!=EOF) {
    switch(c) {
      case 'h':        /* help */
        error(argv[0]);
        break;
      case 'i':        /* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':        /* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      case 'v':
        if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
        if(vlevel<0) error(argv[0]);
        break;
      case 'B':
        if(0==sscanf(optarg,"%d",&nNeighb)) error(argv[0]);
        Bflag=1;
        break;
      case 'P':
        if(0==sscanf(optarg,"%lf",&RndRatio)) error(argv[0]);
        if(RndRatio <=0 ) {
            fprintf(stderr, "ERROR: percentage of sample selected <=0\n") ;
            error(argv[0]) ;
        } else RndRatio /= 100.0 ;
        break;
      case 'M':
        if(0==sscanf(optarg,"%d",&max_leaves)) error(argv[0]);
        if(max_leaves <=0 ) {
            fprintf(stderr, "ERROR: maximal number of leaves <=0\n") ;
            error(argv[0]) ;
        } 
        break;
      case 'T':
        if(0==sscanf(optarg,"%f",&thr_match)) error(argv[0]);
        if(thr_match<0 || thr_match>1) {
            fprintf(stderr, "ERROR: T<0 or T>1\n") ;
            error(argv[0]) ;
        } 
        Tflag=1 ;
        break ;
      default:
        error(argv[0]);
        break;
    }
  }

  /* Check for inconsistant inputs */   
  if(!( iflag && Bflag && oflag )) {
    fprintf(stderr, "input, output, number of coarse neighbors, \n") ;
    fprintf(stderr, "are required, %s\n", argv[0]) ;
    error(argv[0]);
  } 

  
  if(nNeighb!=1 && nNeighb!=9 && nNeighb!=25 && nNeighb!=49 && nNeighb!=81) {
      fprintf(stderr,"ERROR: # of neighbors must be 1, 9, 25, 49 or 81 \n");
      goto Bad;
  }
   
  
  ReadQuadTreePar( &(Par.QTr_par), otfn );   /* ../../segment/src/seg_io.c */
  D = Par.QTr_par.levels ;

  ReadGMM_Par( &(Par.GMM_par), otfn );
  if(vlevel>10) WriteQuadTreePar( stdout, &(Par.QTr_par) ); 
  if(vlevel>100) WriteGMM_Par( &(Par.GMM_par), NULL ); 
  RandInitial() ;  /* Initalization */
  
  
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;
  fscanf(fd, "nFiles=%d\n", &nFiles) ;
  fscanf(fd, "nClasses=%d\n", &nClasses) ;
  if((nFiles<1) || (nClasses<2)) {
      fprintf(stderr, "ERROR: invalid par\n") ;
      fprintf(stderr, "\tnFiles %d (<1?)\n", nFiles) ;
      fprintf(stderr, "\tnClasses %d (<2?)\n", nClasses) ;
  }

  if(vlevel>0) printf("\tnF %d, nClasses %d\n", nFiles, nClasses) ;
 
  hs_fn  = (char **) get_mtrx(nFiles, 128, sizeof(char)) ;
  org_fn = (char **) get_mtrx(nFiles, 128, sizeof(char)) ;
  
  for(i=0; i<nFiles; i++) {
      fscanf(fd, "%s", org_fn[i]) ;
      fscanf(fd, "%s", hs_fn[i]) ;
      sprintf(hs_fn[i], "%s.tmp.%d.hs", otfn, i) ;
      if(vlevel > 0) {
          printf("\t%d\t%s\n", i, org_fn[i]) ;
          printf("\t%d\t%s\n", i, hs_fn[i]) ;
      }
  }
  fclose(fd) ;


  if(! total_number_samples(&total_samples,  hs_fn, nFiles) ) goto Bad ;
  else max_samples = (int)  (total_samples * 1.2 * RndRatio ) ;
  printf("\tmax_samples %d, total %lf, Ratio %lf\n", max_samples,  
      total_samples, RndRatio) ;
  
  /* this must be excuted AFTER max_samples is calculated */
  for(i=D-1; i>1; i--) RndRatio *= 4 ;
  printf("\tRand Ratio %lf, D%d\n", RndRatio,D) ;
 
  /* initial CPT parameters */ 
  Par.CPT.nlevels = Par.QTr_par.levels ;  
  Par.CPT.info.nclasses = nClasses ;
  Par.CPT.info.dimen = nNeighb ;
  Par.CPT.info.nTree = 4 ;
  Par.CPT.info.RangeF = 2 ;
  Par.CPT.info.MaxLeaves = max_leaves ;
  Par.CPT.info.MinSamples = 2 ;
  Par.CPT.info.Margin = (int) ( sqrt((double) Par.CPT.info.dimen) ) ;
  if(Par.CPT.info.Margin * Par.CPT.info.Margin != Par.CPT.info.dimen) {
      fprintf(stderr, "ERROR: CPT dimen %d must be the square of an integer\n",
          Par.CPT.info.dimen) ;
      exit(-1) ;
  } else {
      Par.CPT.info.Margin = (Par.CPT.info.Margin-1) /2 ;
  }
    
  InitializeCPT( &(Par.CPT) ) ; 
    
  for(outputLevel=D-1; outputLevel>0; outputLevel--, RndRatio/=4.0) {

      printf("%d RndRation %lf\n", outputLevel, RndRatio ) ;
      AllocateCPT_Data(&(Par.CPT.trees[outputLevel]), max_samples);

      if(RndRatio>1) rnd_ratio = 1.0 ;
      else rnd_ratio = RndRatio ;
      
      if(groundTruth < 1) {  /* use SMAP estimate to generate samples */
	  if( ! extract_samples_CPT( &Par, nFiles, org_fn, hs_fn, outputLevel, 
    	  (1-groundTruth)*rnd_ratio, max_samples, otfn, thr_match ) ) goto Bad ;
      }

      if(groundTruth > 0) {  /* use ground truth segmentatio only */
	  if( ! extract_groundtruth_CPT( &Par, nFiles, hs_fn, outputLevel, 
    	  groundTruth*rnd_ratio, max_samples, otfn ) ) goto Bad ;
      }

      construct_CPT( &((Par.CPT.trees)[outputLevel]) ) ;

      FreeCPT_Data( &((Par.CPT.trees)[outputLevel]) );
  }

  SaveCPT_Par(&(Par.CPT), otfn) ;

  /* delete all intermediate results */
  if( (vlevel%10)!=6 ) {
       sprintf(comm, "rm -f %s.tmp*", otfn) ;
       system(comm) ; /* comment out to see results*/
  }

Good:
  free_mtrx( (void **) org_fn ) ;
  free_mtrx( (void **) hs_fn ) ;
  
  exit(0) ;
  
  
Bad:
  free_mtrx( (void **) org_fn ) ;
  free_mtrx( (void **) hs_fn ) ;
  
  exit(-1) ;
  

}


void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");
  fprintf(stderr,"     -P percentage of samples selected -- from 0.0 to 100,");
  fprintf(stderr,"              double precision, default value is 20.0 \n");
  fprintf(stderr,"     -B number  -- nNeighb, valid neighborhood sizes are:\n");
  fprintf(stderr,"                 -B  1 : 1x1 neighborhood\n");
  fprintf(stderr,"                 -B  9 : 3x3 neighborhood\n");
  fprintf(stderr,"                 -B 25 : 5x5 neighborhood\n");
  fprintf(stderr,"                 -B 49 : 7x7 neighborhood\n");
  fprintf(stderr,"                 -B 81 : 9x9 neighborhood\n");
  fprintf(stderr,"     -M max number of leaves -- >0, default is 100 \n");
  fprintf(stderr,"     -T float -- the portion of class lables match\n"); 
  fprintf(stderr,"                 ground truth among parents\n"); 
  fprintf(stderr,"                 0<= T <=1, defautl 0.6\n"); 
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");
  fprintf(stderr,"                 > 15 output segment of 3 finest scales.\n");
  fprintf(stderr,"                 > 31 output segment of all scales.\n");
  
  exit(1);
}



/****************************************************************************/
/*  compute the random sampling rate */ 
/****************************************************************************/
static int total_number_samples(double *total,  char **hs_fn, int nFiles)
{
  int i ;
  char fname[256] ;
  eisl_img img_in ;

  *total = 0.0 ;
  for(i=0; i<nFiles; i++) {

      /* decimated ground truth image */

      sprintf(fname, "%s.1", hs_fn[i]) ;   /* number of samples is determined
                                              by the number of parents        */
      if(vlevel>10) printf("\treading decimated ground truth %s\n", fname);
	  if( ! read_img_header (fname, &img_in) ) {
          fprintf(stderr,"ERROR: Cannot read header of %s!\n", fname);
          return(FALSE) ;
	  }
      if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
          fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
          fprintf(stderr," a pallete image\n") ;
          return(FALSE) ;
      }

      (*total) += (double) img_in.width * img_in.height ;
  }

  return (TRUE) ;
}
