#include "CPT.h"
#include "CPT_global.h"


/**********************************************************
     MAP segmentation of the image with Pyramid model
     The transition prob is modeled as the joint distribution
     P(son | father, majority of neighbors of father )
***********************************************************/
void sampleMAP_NO_OPT( 
  struct BigPyramid* Pyramid,
  struct ModelPar *Par,
  char *outfn,
  int outputLevel
)
{
  struct CP_Trees *CPT ;
  int D, nClasses, nRow,nCol, k ;
  unsigned char **Segment, ***SegPt ;
  static int Counts = 0 ;  /* The number of times this program is executed */


  if(vlevel >=1) printf("\tbegining sampleMAP_Pyramid() ......\n") ;

  Counts ++ ;	
  CPT = &(Par->CPT) ;
  nClasses = CPT->info.nclasses ;
  D = (Pyramid->level) ;
  SegPt = (Pyramid->_Segment) ;

  
  /* segment the coarsest resolution */
  SegCoarsestResolution(Pyramid, nClasses) ;
  
  /* This part is only appliable to Pyramid model */
  for(k=D-2; k>=outputLevel; k--) {

	/* the weight of data term should decrease exponentially */

    nRow = (Pyramid->nRow)[k] ;
    nCol = (Pyramid->nCol)[k] ;
  	Segment = (unsigned char **) get_img(nCol,nRow,sizeof(unsigned char)) ;  
  	SegPt[k] = Segment ;
	
	SegPyramidModel(Pyramid, Par, k, (FILE*) NULL);
  
  	if(vlevel >10) printf("\t\tsampleMAP_Pyramid() level %d......\n",k) ;
  }
  
  
  if(vlevel >=100) printf("\tfinish sampleMAP_Pyramid() ......\n") ;
  
  /* save the segmentation results,  /segment/src/seg_io.c  */	
  /* for those levels who have not been segmented, no segment will be */
  /* written out.  */
  if(outfn!=NULL) OutputSeg(Pyramid, outfn, "P") ;	
}



/*****************************************************************************/
/*****************************************************************************/
void InitialDataWeight(struct ModelPar *Par, char *fn, double dataWeight) 
{
    FILE *fd;
    char parfn[80];
    double tmp, *Weight ;
    int i, nlevels ;
   
    nlevels = (Par->QTr_par).levels ; 
    Weight=(Par->DataWeight)=(double*) get_spc(nlevels,sizeof(double)) ;
    
    if(dataWeight<0) {
        fprintf(stderr, "Error:  dataWeight %lf < 0 !\n", dataWeight) ;
        exit(-10) ;
    }
        
    /* initialize *.weight */
    sprintf(parfn, "%s.weight", fn) ;
    if(vlevel>1) fprintf(stderr, "\tSet weights, %s\n", 
        parfn) ;
    fd = fopen (parfn, "w");
    if (fd == NULL) FileOpenErr(parfn) ;
    
    for(i=(Par->QTr_par.levels-2); i>=0; i--) {    
            fprintf(fd,"level=%d  ", i) ;
            fprintf(fd,"%lf\n", dataWeight) ;
            Weight[i]= dataWeight ;
            if(vlevel>1) printf("\t\tDataWeight[%d]=%lf\n",i,dataWeight) ;
    }
    	
    fclose(fd) ;
}





