#include "CPT.h"
#include "CPT_global.h"

/* Samples of CPT are accepted only when there are Score_Thr percent */
/* of samples are correctly segmentated at the next coarser scale */


/****************************************************************************/
/*  compute the random sampling rate */ 
/****************************************************************************/
double CalculateRndRatio(int N,  char **hs_fn, int nFiles)
{
  int i ;
  char fname[256] ;
  double sum ; 
  eisl_img img_in ;

  sum = 0.0 ;
  for(i=0; i<nFiles; i++) {

      /* decimated ground truth image */

      sprintf(fname, "%s.0", hs_fn[i]) ;
      if(vlevel>10) printf("\treading decimated ground truth %s\n", fname);
	  if( ! read_img_header (fname, &img_in) ) {
          fprintf(stderr,"ERROR: Cannot read header of %s!\n", fname);
          return(FALSE) ;
	  }
      if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
          fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
          fprintf(stderr," a pallete image\n") ;
          return(FALSE) ;
      }

      sum += img_in.width * img_in.height ;
  }

  if(sum==0.0) {
      fprintf(stderr, "ERROR: total number of pixels=0\n") ;
      exit(-10) ;
  }
  return ( (double) N / sum ) ;
}



/****************************************************************************/
/*  initial CPT */ 
/****************************************************************************/
void InitializeCPT( struct CP_Trees *CPT )
{
  int i ;
  struct Trees *treesPt ;
  struct Tree_Info *InfoPt ;
  
  InfoPt = &(CPT->info) ;
  CPT->trees = (struct Trees *) get_spc(CPT->nlevels, sizeof(struct Trees)) ;
  treesPt = CPT->trees ;
  
  for(i=0; i<(CPT->nlevels); i++) {
      
      (treesPt[i]).infoPt = InfoPt ;
      (treesPt[i]).roots = (struct TreeNode **) get_spc( InfoPt->nTree, 
          sizeof(struct TreeNode*)) ;
      (treesPt[i]).nLeaves = (int *) get_spc( InfoPt->nTree, sizeof(int)) ;
      ((treesPt[i]).data).nSamples = 0 ;
  }
}



/****************************************************************************/
/****************************************************************************/
void AllocateCPT_Data( struct Trees *treePt, int N )
{
  struct Tree_Data *Data ;
  struct Tree_Info *InfoPt ;
  
  InfoPt = treePt->infoPt ;
  Data = &(treePt->data) ;
  
  Data->Y = (int **) get_mtrx(InfoPt->nTree, N, sizeof(int)) ;
  Data->X = (int **) get_mtrx(N, InfoPt->dimen, sizeof(int)) ;
  Data->nSamples = 0 ;

}



/****************************************************************************/
/****************************************************************************/
void FreeCPT_Data( struct Trees *treePt )
{
  struct Tree_Data *Data ;
  
  Data = &(treePt->data) ;
  free_mtrx( (void **) Data->Y ) ;
  free_mtrx( (void **) Data->X ) ;

}



/****************************************************************************/
/****************************************************************************/
static void SaveSamplesData(
  unsigned char **segL, 
  unsigned char **segH,   /* segment from Low or High resolution */
  int nRowL, int nColL, int nRowH, int nColH,
  struct Trees *treePt,
  unsigned char **seg_hs_L,
  float thr_match, 
  double RndRatio, 
  int N                   /* max number of samples */
)
{
  FILE *fd;
  char fn[128] ;
  static int count=0 ;

  int  i,j,k,i1,j1,i2,j2;  /* indexes for band and column */
  double Rand;	      /* the percentage of samples will be selected */
  static int  *Neighbors, First=1;  
  int Margin, nClasses, nNeighb, m, index ;
  int **DataX, **DataY, *nSamples ;
  int score,  score_min ;

 
  Margin = (treePt->infoPt)->Margin ;  
  nClasses = (treePt->infoPt)->nclasses ;
  nNeighb = (treePt->infoPt)->dimen ;
  DataX = (treePt->data).X ;
  DataY = (treePt->data).Y ;
  nSamples = &((treePt->data).nSamples) ;
  score_min = (int) (nNeighb * thr_match) ;

if(count==0) printf("score_min %d\n", score_min) ;
count++ ;
  
  if(First==1) {
  	Neighbors = (int*) get_spc(nNeighb, sizeof(int)) ;
  	First = 0 ;
  }

  printf("\trL%d, cL%d, rH%d, cH%d\n", nRowL, nColL, nRowH, nColH) ;
  
  if((nRowH<nRowL) || (nColH<nColL)) {
  	printf("Wrong dimension,  SaveSamples\n") ;
  	exit(-1) ;
  }

  for(i=0; i<nRowL; i++) 
  for(j=0; j<nColL; j++) {	/* parents */
  
  	Rand = random2() ; 
	if( ( Rand<=RndRatio ) && ( N> (*nSamples) ) ) {

  	    index =0 ;
  	    score =0 ;
  	    for(i1=-Margin; i1<=Margin; i1++)
  	    for(j1=-Margin; j1<=Margin; j1++) {

  		    i2=i+i1; j2=j+j1 ;

  		    /* if a pixel is outside an image, map it back to */
  		    /* the nearest pixel inside the image */

  		    if(i2<0) i2=0 ;   /* do not use default class */
  		    if(i2>=nRowL) i2=nRowL-1 ;
  		    if(j2<0) j2=0 ;
  		    if(j2>=nColL) j2=nColL-1 ;


  		    if((i2>=0)&&(i2<nRowL)&&(j2>=0)&&(j2<nColL)) {
  			m =  (int) segL[i2][j2] ;
 			CheckClassNumber("SaveSampleData 1", m, 0,nClasses);
 			Neighbors[index]=m ;

 			m = (int) seg_hs_L[i2][j2] ;
  			CheckClassNumber("SaveSampleData 1", m, 0,nClasses);
			if(m==Neighbors[index]) score ++ ;    

  		    } else { 

  			fprintf(stderr, "ERROR: check SaveSamples()\n")  ;
  			exit(-10) ;
  		    }

  		    index++ ;
  	    }
  		
            
            if( score >= score_min )  {
  		index = 0 ;
  		for(i1=0; i1<2; i1++)
  		for(j1=0; j1<2; j1++) {

  		    i2=2*i+i1;  j2=2*j+j1 ;

  		    /* if a pixel is outside the image, */
  		    /* map it back to the nearest */
  		    /* pixel inside the image */
  		    if(i2<0) i2=0 ;
  		    if(i2>=nRowH) i2=nRowH-1 ;
  		    if(j2<0) j2=0 ;
  		    if(j2>=nColH) j2=nColH-1 ;

   		    if((i2>=0)&&(i2<nRowH)&&(j2>=0)&&(j2<nColH)) {

   			m = segH[i2][j2] ;
 			CheckClassNumber("SaveSample 2", m, 0,nClasses) ;
  			DataY[index][(*nSamples)] = m ;


   		    } else  {
  			fprintf(stderr, "ERROR: Invalid indices in SaveSamples()\n")  ;
  			exit(-10) ;
  		    }

  		    index ++ ;
  		}

  		for(k=0; k<nNeighb; k++) {
  	   	    DataX[(*nSamples)][k] = Neighbors[k] ;

  		}
   		(*nSamples) ++ ;
            }
  	}
   }
   
}



/****************************************************************************/
/****************************************************************************/
static void SaveSamples(
  unsigned char **segL, 
  unsigned char **segH,   /* segment from Low or High resolution */
  int nRowL, int nColL, int nRowH, int nColH,
  struct Trees *treePt,
  FILE *fd_out,
  double RndRatio
)
{
  int  i,j,k,i1,j1,i2,j2;  /* indexes for band and column */
  double Rand;	      /* the percentage of samples will be selected */
  static int flag1Row, Max1Row, Class1Row ;  /* detect rows containing single class */
  static int *Current1Row, *Neighbors, First=1;  
  int Margin, nNeighb, nClasses, m, index ;
  
  Margin = (treePt->infoPt)->Margin ;  
  nNeighb = (treePt->infoPt)->dimen ;  
  nClasses = (treePt->infoPt)->nclasses ;
  
  if(First==1) {
  	Neighbors = (int*) get_spc(nNeighb, sizeof(int)) ;
  	Current1Row = (int*) get_spc(nClasses, sizeof(int)) ;
  	Max1Row = 100000 ;
  	for(i=0; i<nClasses; i++) Current1Row[i]=0 ;
  	First = 0 ;
  }

  printf("\trL%d, cL%d, rH%d, cH%d\n", nRowL, nColL, nRowH, nColH) ;
  
  if((nRowH<nRowL) || (nColH<nColL)) {
  	printf("Wrong dimension,  SaveSamples\n") ;
  	exit(-1) ;
  }

  for(i=0; i<nRowL; i++) 
  for(j=0; j<nColL; j++) {	/* parents */
  
  	flag1Row=1 ;
  	index =0 ;
  	for(i1=-Margin; i1<=Margin; i1++)
  	for(j1=-Margin; j1<=Margin; j1++) {
  		
  		i2=i+i1; j2=j+j1 ;
  		
  		/* if a pixel is outside the image, map it back to the nearest */
  		/* pixel inside the image */
  		if(i2<0) i2=0 ;
  		if(i2>=nRowL) i2=nRowL-1 ;
  		if(j2<0) j2=0 ;
  		if(j2>=nColL) j2=nColL-1 ;
  		
  		if((i2>=0)&&(i2<nRowL)&&(j2>=0)&&(j2<nColL)) {
  			m =  (int) segL[i2][j2] ;
 			CheckClassNumber("SaveSample 1", m, 0,nClasses) ;
 			Neighbors[index]=m ;
  		
  		} else { 
  		    fprintf(stderr, "ERROR: check SaveSamples()\n")  ;
  		    exit(-10) ;
  		}
  		
  		if( Neighbors[index] != Neighbors[0] )  flag1Row =0 ;
  		
  		index++ ;
  	}
  		
  	if(flag1Row) {
  		Class1Row = Neighbors[0] ;	
  		
  		if(Class1Row<0 || Class1Row>= nClasses) {
  		    fprintf(stderr, "ERROR: Invalid Class1Row in SaveSamples()\n")  ;
  		    exit(-10) ;
  		}
   		
  		Current1Row[Class1Row] ++ ;
  		
  	} else Class1Row = 0 ;
  		
  	Rand = random2() ; 
	if( ((!flag1Row)||( (flag1Row)&&(Current1Row[Class1Row]<Max1Row) )) && 
	     (Rand<=RndRatio) ) {
  
  		
  		for(i1=0; i1<2; i1++)
  		for(j1=0; j1<2; j1++) {
  		
  			i2=2*i+i1;  j2=2*j+j1 ;
  			
  			/* if a pixel is outside the image, map it back to the nearest */
  			/* pixel inside the image */
  			if(i2<0) i2=0 ;
  			if(i2>=nRowH) i2=nRowH-1 ;
  			if(j2<0) j2=0 ;
  			if(j2>=nColH) j2=nColH-1 ;
  		
   			if((i2>=0)&&(i2<nRowH)&&(j2>=0)&&(j2<nColH)) {
   			
   				m = segH[i2][j2] ;
 				CheckClassNumber("SaveSample 2", m, 0,nClasses) ;
  		    		
  		    	fprintf(fd_out, " %2d  ", m) ;
  		    		
   		    } else  {
  		        fprintf(stderr, "ERROR: Invalid indices in SaveSamples()\n")  ;
  		        exit(-10) ;
  		    }
  		}
  			
  	   	for(k=0; k<nNeighb; k++) fprintf(fd_out, "%2d  ", Neighbors[k]) ;
   		fprintf(fd_out, "\n") ;

  	}
   }
   
   for(i=0; i<nClasses; i++) {
   	printf("Class %d,  Current1Row %d\n",i,Current1Row[i]) ;
   }
   printf("\n") ;
      
}



/****************************************************************************/
/*  extract sample set for CPT construction*/ 
/****************************************************************************/
int extract_samples_CPT( struct ModelPar *ParPt, int nFiles, char **org_fn,
    char **hs_fn, int Level, double RndRatio, int N /* max_n_samples */, 
    char *otfn, float thr_match )
{
  struct BigPyramid _BigPyramid ;
  eisl_img img_in;            /* input image */
  unsigned char **seg_hs_L;   /* ground-truth segmentation at low resolution */
  unsigned char **seg_hand;   /* ground-truth segmentation for one image */
  
  int ncol, nrow, *nRow, *nCol ;
  int i ;
  char filename[80], filename2[80] ;
  
  if(vlevel>1) printf("\tneed %f percent match before a sample is taken\n",
    100* thr_match ) ;
  
  for(i=0; i<nFiles; i++) {
  
      if(vlevel >0) printf("\t\treading input image %s\n", org_fn[i]) ;
      if( ! read_img (org_fn[i], &img_in) ) {
          fprintf(stderr,"ERROR: Cannot read image of %s!\n", filename);
          goto Bad ;
      }
      if( img_in.img_type != 'g' ) {
          fprintf(stderr,"ERROR: orginal image must be gray level image,") ;
          fprintf(stderr," %s, extract_samples_CPT\n", org_fn[i]) ;
          goto Bad ;
      }
  

      /* Initialize the data model */
      /* Feature Initialization        feature.c */

      FeatureInitial(&img_in, &(ParPt->GMM_par), &ncol, &nrow) ;  

      /* construct the pyramid */
      ConstructPyramid(nrow,ncol,&_BigPyramid, ParPt) ;      /* u_pyramid.c */
  

      /* subsampling rate = 2,  in featureProb.c */ 
      FeatureProb(&_BigPyramid, &img_in, &(ParPt->GMM_par)); /* featureProb.c */

      /* free the orginal image */
      free_img((void **) img_in.mono) ;

      nRow=_BigPyramid.nRow ;
      nCol=_BigPyramid.nCol ;


      /* the level where samples need to be outputed */
      if(Level>0) {
          sprintf(filename, "%s.%d", hs_fn[i], Level-1) ;
          if(vlevel >1) printf("\treading input image %s\n", filename);
		  if( ! read_img (filename, &img_in) ) {
        	  fprintf(stderr,"ERROR: Cannot read image of %s!\n", filename);
        	  goto Bad ;
	  }
    	  if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
        	  fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
        	  fprintf(stderr," a pallete image\n") ;
        	  goto Bad ;
    	  }
          if((nRow[Level-1] != img_in.height) || 
             (nCol[Level-1] != img_in.width)) {
              fprintf(stderr,"ERROR: Dimension conflicts %s !!\n",filename);
              fprintf(stderr,"ERROR: %d!=%ld, ",nRow[Level-1],img_in.height);
              fprintf(stderr,"ERROR: %d!=%ld !\n",nCol[Level-1],img_in.width);
              goto Bad;
          }
          seg_hand=img_in.mono ;    /* to output samples */
      }

  
      /* read in ground truth segmentation at the parent resolution */
      sprintf(filename, "%s.%d", hs_fn[i], Level) ;
      if(vlevel >1) printf("\treading input image %s\n", filename);
      if( ! read_img (filename, &img_in) ) {
              fprintf(stderr,"ERROR: Cannot read image of %s!\n", filename);
              goto Bad ;
      }
      if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
              fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
              fprintf(stderr," a pallete image\n") ;
              goto Bad ;
      }
      if((nRow[Level] != img_in.height) || 
         (nCol[Level] != img_in.width)) {
          fprintf(stderr,"ERROR: Dimension conflicts %s !!\n",filename);
          fprintf(stderr,"ERROR: %d!=%ld, ",nRow[Level],img_in.height);
          fprintf(stderr,"ERROR: %d!=%ld !\n",nCol[Level],img_in.width);
          goto Bad;
      }
      seg_hs_L=img_in.mono ;    /* to evaluate the SMAP estimate */

  
      /* segmentation begins */
      upPropagate(&_BigPyramid, ParPt) ;  /* decimate.c */
   

      if( (vlevel%10)!=6 ) {
         sampleMAP_NO_OPT(&_BigPyramid, ParPt, (char*) NULL, Level);  
      } else {
         sprintf(filename2, "%s.tmp.cpt_map.n%d", otfn, i) ;
         sampleMAP_NO_OPT(&_BigPyramid, ParPt, filename2, Level);  
      }

      if( RndRatio >1.0) {
          SaveSamplesData(_BigPyramid._Segment[Level], seg_hand, 
            nRow[Level],nCol[Level], nRow[Level-1],
            nCol[Level-1], &((ParPt->CPT.trees)[Level]), 
            seg_hs_L, thr_match, 1.0, N ); 
      } else {
          SaveSamplesData(_BigPyramid._Segment[Level], seg_hand, 
            nRow[Level],nCol[Level], nRow[Level-1],
            nCol[Level-1], &((ParPt->CPT.trees)[Level]), 
            seg_hs_L, thr_match,  RndRatio, N);
      }
      
      if(seg_hand != NULL) free_img((void **) seg_hand) ;
      if(seg_hs_L != NULL) free_img((void **) seg_hs_L) ;
      FreePyramid(&_BigPyramid) ; /* u_pyramid.c */
      
  }
  
  return(TRUE) ;
  
Bad:
  fprintf(stderr, "ERROR: in extract_samples_CPT()\n") ;
  return(FALSE) ;
}



/****************************************************************************/
/*  extract sample only from ground truth images for CPT construction*/ 
/****************************************************************************/
int extract_groundtruth_CPT( struct ModelPar *ParPt, int nFiles,
    char **hs_fn, int Level, double RndRatio, int N /* max_n_samples */, 
    char *otfn )
{
  eisl_img img_in;                /* input image */
  unsigned char **seg_hs_L;       /* Low resolution ground-truth segmentation */
  unsigned char **seg_hs_H;       /* High resolution ground-truth segmentation */
  int ncol_L, nrow_L, ncol_H, nrow_H;
  int i ;
  char filename[80], filename2[80] ;
  
  for(i=0; i<nFiles; i++) {
  
      /* Level, the level where samples need to be outputed */
      if(Level>0) {     /* coarse resolution ground truth segmentation */
          sprintf(filename, "%s.%d", hs_fn[i], Level-1) ;
          if(vlevel >1) printf("\treading input image %s\n", filename);
	  if( ! read_img (filename, &img_in) ) {
              fprintf(stderr,"ERROR: Cannot read header of %s!\n", filename);
              goto Bad ;
	  }
    	  if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
              fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
              fprintf(stderr," a pallete image\n") ;
              goto Bad ;
    	  }
    	  ncol_H = img_in.width ;
    	  nrow_H = img_in.height ;
          seg_hs_H=img_in.mono ;    /* to output samples */
      }

      /* Level, the level where samples need to be outputed */
      if(Level>0) {     /* fine resolution ground truth segmentation */
          sprintf(filename, "%s.%d", hs_fn[i], Level) ;
          if(vlevel >1) printf("\treading input image %s\n", filename);
	  if( ! read_img (filename, &img_in) ) {
              fprintf(stderr,"ERROR: Cannot read header of %s!\n", filename);
              goto Bad ;
	  }
    	  if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
              fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
              fprintf(stderr," a pallete image\n") ;
              goto Bad ;
    	  }
    	  ncol_L = img_in.width ;
    	  nrow_L = img_in.height ;
    	  seg_hs_L = img_in.mono ;
    	  
          if( (ncol_H/2 + ncol_H%2) != ncol_L || 
              (nrow_H/2 + nrow_H%2) != nrow_L ) {
              fprintf(stderr,"ERROR: Dimension conflicts %s !!\n",filename);
              fprintf(stderr,"ERROR: %d!=%d, ",ncol_H, ncol_L);
              fprintf(stderr,"ERROR: %d!=%d !\n",nrow_H, nrow_L);
              goto Bad;
          }
      }

  
      if( RndRatio >1.0) {
          SaveSamplesData(seg_hs_L, seg_hs_H, nrow_L, ncol_L, nrow_H, ncol_H,
          &((ParPt->CPT.trees)[Level]), seg_hs_L, 0.0, 1.0, N ); 
      } else {
          SaveSamplesData(seg_hs_L, seg_hs_H, nrow_L, ncol_L, nrow_H, ncol_H,
          &((ParPt->CPT.trees)[Level]), seg_hs_L, 0.0, RndRatio, N);
      }
      
      if(seg_hs_L != NULL) free_img((void **) seg_hs_L) ;
      if(seg_hs_H != NULL) free_img((void **) seg_hs_H) ;
      
  }
  
  return(TRUE) ;
  
Bad:
  fprintf(stderr, "ERROR: in extract_groundtruth_CPT()\n") ;
  return(FALSE) ;
}
