/*****************************************************************************/
/* We want to obtain 8x8 block segmentation of input images */
/* In order to use SMAP-2 package,  we will first decimate the orginal */
/* input image 4x4.  Then combined with the 8x8 groundtruth segmentation */
/* we can directly apply SMAP-2 package. */
/*****************************************************************************/
 

#include "../../utilities/src/util.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <time.h>


void error(char *name);
static decimate_nxn( char *org_fn, char *out_fn, int dec_rate ) ;


int main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd;
  char *infn;       	      /* input image file */
  char *otfn;                 /* output segmentation file name */
  char filename[256], filename2[256] ;
  int  nFiles ;		          /* number of input files */
  int  oflag,iflag;
  int  i, nClasses ;


  iflag=oflag=0;
  
  while( (c=getopt(argc,argv,"h,i:,o:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      default:
        fprintf(stderr, "Invalid char %c\n",c) ;
        error(argv[0]);
        break;
    }
  }


  /* Check for inconsistant inputs */   
  if(!(oflag&&iflag)) {
    fprintf(stderr, "Input,output are required!\n",argv[0]);
    error(argv[0]);
  }


  /* read input image */
  printf("\treading input file %s \n", infn) ;
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;
  fscanf(fd, "nFiles=%d\n", &nFiles) ;
  fscanf(fd, "nClasses=%d\n", &nClasses) ;
 

  /* decimate original images */
  for(i=0; i<nFiles; i++) {
      fscanf(fd, "%s", filename) ;   /* name of org images */
      sprintf(filename2, "%s.org%d", otfn, i) ;
      decimate_nxn(filename, filename2, 4) ;
      fscanf(fd, "%s", filename) ;   /* name of ground truth images */
  }


Good:
  exit(0) ;  /* use this to ingore un-de-allocated memory */

Bad:
  exit(-1) ;  
}




void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");

  exit(1);
}


/*****************************************************************************/
/*  Decimate org_fn image by dec_rate x dec_rate */
/*****************************************************************************/

static decimate_nxn( char *org_fn, char *out_fn, int dec_rate )
{
  char filename[256] ;
  int  i,j, ii,jj, k,m,n ;
  unsigned int  sum, N, M ;
  unsigned long nRowIn, nColIn, nRowOut, nColOut ;
  eisl_img img_out ;
  eisl_img img_in ;                     /* input image, tiff or pxm format,*/
                                        /* ../../utilities/src/eisl_img.h*/
  unsigned char **mono, ***color ;

  if( dec_rate <=0 ) {
      fprintf(stderr, "ERROR: invalide decimation rate %d\n", dec_rate) ;
      goto Bad ;
  }  
  
  /* orginal image */
  printf("\treading original image %s .\n", org_fn) ;
  if( ! read_img (org_fn, &img_in) ) {
      fprintf(stderr,"ERROR: Cannot read image %s!\n", org_fn);
      goto Bad ;
  }
  sprintf(filename, "%s.decImg", out_fn) ;
  
  
  /* allocate image_out */
  nRowIn = (unsigned long) img_in.height ;
  nColIn = (unsigned long) img_in.width ;
  nRowOut = (nRowIn + dec_rate -1) / dec_rate ;
  nColOut = (nColIn + dec_rate -1) / dec_rate ;
  
  printf("\tnRowIn %d, nColIn %d, nRowOut %d, nColOut %d, %c\n",
      nRowIn, nColIn, nRowOut, nColOut, img_in.img_type ) ;

  if(! allocate_img(&img_out, nRowOut, nColOut, img_in.img_type) ) {
      fprintf(stderr, "ERROR: cannot allocate output image %s\n",
          filename) ;
      goto Bad ;
  }
  
  
  /* decimate */
  
  N = dec_rate * dec_rate ;
  switch( img_in.img_type ) {
    case 'g':
    case 'l':
    
      mono = img_in.mono ;
      for(i=0, m=0; m<nRowOut; i+=dec_rate, m++)
      for(j=0, n=0; n<nColOut; j+=dec_rate, n++) {

    	  sum = 0;  /* average pixels in a window of dec_rate x dec_rate */
          M = 0 ;
    	  for(ii=i+dec_rate-1; ii>=i; ii--)
    	  for(jj=j+dec_rate-1; jj>=j; jj--) {
              if(ii<nRowIn && jj<nColIn) {
                  sum += mono[ii][jj] ;
                  M ++ ;
              }
          }
          if( M>0 ) { 
              (img_out.mono)[m][n] = (unsigned char) 
                  ( (double) sum / (double) M + .5 ) ;
          } else {
              fprintf(stderr, "ERROR:  can this happen ? M<=0\n") ;
              fprintf(stderr, "ERROR:  ii %d, jj %d, i %d, j %d, m %d, n %d\n",
                  ii,jj,i,j,m,n) ;
              exit(-1) ;
          }
      }

      break ;
   
    case 'c':
    case 'y':
    case 'L':
    
      color = img_in.color ;
      for(i=0, m=0; m<nRowOut; i+=dec_rate, m++)
      for(j=0, n=0; n<nColOut; j+=dec_rate, n++) {
          
          for(k=0; k<3; k++) {
    	      sum = 0;       /* average in a window of dec_rate x dec_rate */
              M = 0 ;
    	      for(ii=i+dec_rate-1; ii>=i; ii--)
    	      for(jj=j+dec_rate-1; jj>=j; jj--) {
                  if(ii<nRowIn && jj<nColIn) {
    	              sum += color[k][ii][jj] ;
                      M ++ ;
                  }
              }
              if( M>0 ) {
                  (img_out.color)[k][m][n] = (unsigned char) 
            	      ( (double) sum / (double) M + .5 ) ;
              } else {
                  fprintf(stderr, "ERROR:  can this happen? M<=0\n") ;
                  fprintf(stderr, "ERROR:  ii %d, jj %d, i %d, j %d, m %d, n %d\n",
                      ii,jj,i,j,m,n) ;
                  exit(-1) ;
              }
          }
      }
      break ;
      
    default:
      fprintf(stderr, "ERROR: invalid img_type %c\n", img_in.img_type) ;
      goto Bad ;
  }
   
  
  /* write out the decimated image */ 
  
  if(! write_img(filename, &img_out, 't', "lzw:2") ) {
      fprintf(stderr, "ERROR: can not write out %s\n", filename) ;
      goto Bad ;
  }

  free_eisl_img( &img_in ) ; 
  free_eisl_img( &img_out ) ; 
  return(TRUE) ;
  
Bad :
  free_eisl_img( &img_in ) ;
  free_eisl_img( &img_out ) ; 
  return(FALSE) ;
  
}



