#include "EM_Dec.h"
extern int vlevel ;

/********************************************************************
	Maximization at a pixel 
	ll={ll[i][j]}[class] is the loglikelihood at pixel i,j
	TranProb =  transition prob at pixel i,j.   
*********************************************************************/
static unsigned char SegMaxProb(double *ll, int nClasses) 
  /* IndMax is a vector to handle ties,  */
  /* use static to avoid frequently allocation */
{
  unsigned char tmp ;
  int j,m ;
  int Tie ;
  double _Max, sum ;
  int *IndMax ;
  
  IndMax = (int *) get_spc(nClasses, sizeof(int)) ;
  
  /* find the max */
  _Max = ll[0] ;
  Tie = 0 ;	/* no tie */
  IndMax[Tie] = 0 ;

  for(j=1; j<nClasses; j++) {
  	if(_Max < ll[j]) {
   		Tie = 0 ;	/* no tie */
 		_Max = ll[j] ;
  		IndMax[Tie] = j ;
  			
  	} else if(_Max== ll[j]) {
  		Tie ++ ;
  		IndMax[Tie] = j ;
  	}
  }
  
  
  /* normalize ll */
  for(j=0; j<nClasses; j++) ll[j] -= _Max ;
  sum = 0.0 ;
  for(j=0; j<nClasses; j++) sum += exp(ll[j]) ;
  if(sum > 0) {
  	sum = log(sum) ;
  	for(j=0; j<nClasses; j++) ll[j] -= sum ;
  } else {
  	ll[j] = log((double) 1.0/ nClasses) ;
  }
  	
  
  if(Tie==0) { 
    
    tmp = (unsigned char) (IndMax[0]) ;	/* no tie */
  } else {
    	
  	m = rand() ;
  	if(vlevel>100) printf("Tie=%d, m=%d, %d\n",Tie,m,m%(Tie+1)) ;
  	m = m % (Tie+1) ;
  	
    tmp = (unsigned char) (IndMax[m]) ;	/* for a tie, randomly pick one */
  }
    
  free( (void *) IndMax) ;
  return(tmp) ;
}



/**********************************************************************/
/*	Calculate the density P(X(n)|X(0)) */
/**********************************************************************/
void Decimate(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par) 
{
  int D, nRow,nCol,M;
  int *nRowPt,*nColPt, nRow1, nCol1;
  int m,i1,j1, i2,j2, ic,jc, kk ;
  unsigned char ***Seg, **SegSon, **SegCurrent ;
  int son, sonR, sonC ;
  double ****LL, ***Lcoarse, ***Lfine, tmp1, tmp2 ;
  double ***alpha ;
  
  D = QTr_par->levels ;
  M = QTr_par->nclasses ;
  alpha = QTr_par->Alpha0 ;
  nRowPt = Pyramid->nRow ;
  nColPt = Pyramid->nCol ;
  Seg = Pyramid->_Segment ;
  LL = Pyramid->LLikeH ;
    
  for(kk=1; kk<D; kk++) {  /* kk starts from 1,  not the finest scale */
   	  nRow = nRowPt[kk] ;
   	  nCol = nColPt[kk] ;
	  LL[kk] = (double ***) multialloc(sizeof(double),3,nRow,nCol,M) ;
  }


  /********  Need this when excutes without EM algorithm  *******/
  RandInitial() ;
  

  /* Calculate P(X(0)|X(1)) */
  SegCurrent = Seg[1] ;
  Lcoarse = LL[1] ;
  nRow1 = nRowPt[1] ;
  nCol1 = nColPt[1] ;
  	
  SegSon = Seg[0] ;
  nRow = nRowPt[0] ;
  nCol = nColPt[0] ;
  
  for(ic=0; ic<nRow1; ic++)
  for(jc=0; jc<nCol1; jc++) {
  	
  	sonR = ic*2 ;
  	sonC = jc*2 ;
  	for(m=0; m<M; m++) Lcoarse[ic][jc][m]=0.0 ;
  		
  	for(i1=0; i1<2; i1++)
   	for(j1=0; j1<2; j1++) {
   	
 		i2 = sonR +i1 ;
  		j2 = sonC +j1 ;
  		
   		if((i2<nRow) && (j2<nCol)) {
 	
  			son = SegSon[i2][j2] ;

  			for(m=0; m<M; m++) {
  				Lcoarse[ic][jc][m] += alpha[1][son][m] ;  
 				/* ll[m]>0 means prob[m]==0 */
  			} 
  		}
  	}
  	
  	SegCurrent[ic][jc] = SegMaxProb(Lcoarse[ic][jc], M) ;
 }
  
  
  /* Calculate P(X(0)|X(kk)) */
  for(kk=2; kk<D; kk++) {
  
  	nRow1 = nRowPt[kk] ;
  	nCol1 = nColPt[kk] ;
  	SegCurrent = Seg[kk] ;
  	Lcoarse = LL[kk] ;
  	
	nRow = nRowPt[kk-1] ;
  	nCol = nColPt[kk-1] ;
  	Lfine = LL[kk-1] ;
  	
  	for(ic=0; ic<nRow1; ic++)
  	for(jc=0; jc<nCol1; jc++) {
  	
  		sonR = ic*2 ;
  		sonC = jc*2 ;
  		for(m=0; m<M; m++) Lcoarse[ic][jc][m]=0.0 ;
  		
  		for(i1=0; i1<2; i1++)
   		for(j1=0; j1<2; j1++) {
   	
 			i2 = sonR +i1 ;
  			j2 = sonC +j1 ;
  		
   			if((i2<nRow) && (j2<nCol)) {
   			
  			    for(m=0; m<M; m++) {
  			    
  			        tmp1 = 0.0 ;
 			        for(son=0; son<M; son++) { /* integrate over son */
 			        	
 			        	/* m->son(i2,j2)->X(0) */
 			        	tmp2 = Lfine[i2][j2][son] + alpha[kk][son][m] ;
 			        	tmp1 += exp(tmp2) ;
			        	
  			        } 
  			        
 			        if(tmp1>MIN_LOG) Lcoarse[ic][jc][m] += log(tmp1) ;  
  				    else  Lcoarse[ic][jc][m] += log(MIN_LOG) ;
  			    }
  			}
  		}
  		
  		SegCurrent[ic][jc] = SegMaxProb(Lcoarse[ic][jc], M) ;
  	} 
  }
   		
}  



