/*****************************************************************************/
/* EM algorithm for estimating transition probabilities with quadtree model */ 
/* Decimate ground truth using ML estimate given quadtree parameters and  */
/* ground truth segmentation. */
/*****************************************************************************/
 

#include "EM_Dec.h"
/*#include <sys/getopt.h>*/
#include <stdio.h>
#include <unistd.h>

int vlevel ;

void error(char *name);
void RandInitial() ;


int main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd;
  struct BigPyramid *_BigPyramid;
  struct Par_QTR QTr_par ;    /* quadtree parameters */
  char *infn;       	      /* input image file */
  char *otfn;                 /* output segmentation file name */
  char **org_fn, **hs_fn ;
  char fname[128] ;
  int  nFiles ;		          /* number of input files */
  int  *nRow, *nCol ;
  int  oflag,iflag;
  int  i, N, nClasses ;


  vlevel=1;  
  N=10 ;
  iflag=oflag=0;
  
  while( (c=getopt(argc,argv,"h,i:,o:,N:,v:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      case 'N':
        if(0==sscanf(optarg,"%d",&N)) error(argv[0]);
        break;
      case 'v':
        if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
        if(vlevel<0) vlevel=0;
        break;
      default:
        fprintf(stderr, "Invalid char %c\n",c) ;
        error(argv[0]);
        break;
    }
  }


  /* Check for inconsistant inputs */   
  if(!(oflag&&iflag)) {
    fprintf(stderr, "Input,output are required!\n",argv[0]);
    error(argv[0]);
  }

  
  if(N<1) { 
      fprintf(stderr, "WARNING: Invalid number of iterations %d < 1\n", N) ;
      fprintf(stderr, "WARNING: use default, 10\n") ;
      N=10 ;
  }

  /* read input image */
  if(vlevel >0) printf("\treading input information from %s\n", infn) ;
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;
  fscanf(fd, "nFiles=%d\n", &nFiles) ;
  fscanf(fd, "nClasses=%d\n", &(QTr_par.nclasses)) ;
  nClasses = QTr_par.nclasses ;
  if((nFiles<1) || (nClasses<2)) {
      fprintf(stderr, "ERROR: invalid par\n") ;
      fprintf(stderr, "\tnFiles %d (<1?)\n", nFiles) ;
      fprintf(stderr, "\tnClasses %d (<2?)\n", QTr_par.nclasses) ;
  }
 
  nRow = (int *) get_spc(nFiles, sizeof(int)) ;
  nCol = (int *) get_spc(nFiles, sizeof(int)) ;
  hs_fn = (char **) get_mtrx(nFiles, 128, sizeof(char)) ;
  org_fn = (char **) get_mtrx(nFiles, 128, sizeof(char)) ;


  /* read training images and test consistancy */
  if( ! read_training_images(fd, QTr_par.nclasses, nFiles, org_fn,
      hs_fn, nRow, nCol) ) {
      goto Bad ;
  }
  fclose(fd) ;

  QTr_par.levels = number_resolutions_allowed( -1, nRow, nCol, nFiles );  
  QTr_par.Alpha0 = (double ***) mget_spc(QTr_par.levels, sizeof(double**)) ;
 
  Initial_QuadTree(&QTr_par) ; /* EM_alpha.c */
  if(vlevel>10) EM_WriteQuadTreePar(stdout, &QTr_par, 0/* no decimation */);
    
    
  /* Initalization */
  RandInitial() ;

  
  /* read and initialize other images */
  _BigPyramid = get_spc(nFiles, sizeof(struct BigPyramid)) ;
  EM_InitialPyramid(_BigPyramid, &QTr_par, hs_fn, nFiles) ;


  /* estimate quadtree parameters using EM algorithm */
  EM(_BigPyramid, &QTr_par, 1,1,N, nFiles, otfn) ;  
  
  
  /* output estimated quadtree parameters */
  sprintf(fname, "%s.qtr", otfn) ;
  if(vlevel >0) printf("\tsave quadtree par to %s\n", fname) ;
  fd = fopen (fname, "w");  if (fd == NULL) FileOpenErr(fname) ;
  EM_WriteQuadTreePar(fd, &QTr_par, DECIMATE_FACTOR ) ;
  fclose(fd) ;
    
    
  /* decimate ground truth segmentation using ML estimate */
  /* given the quadtree parameters */
  for(i=0; i<nFiles; i++)  Decimate(&(_BigPyramid[i]), &QTr_par) ;
  

  /* output decimated ground truth segmentation */
  for(i=0; i<nFiles; i++) {
      sprintf(fname, "%s.tmp.%d.hs", otfn, i) ;
      OutputDecimatedSeg(&(_BigPyramid[i]), fname, "", DECIMATE_FACTOR);
  }



Good:
  free_mtrx( (void**) org_fn) ;
  free_mtrx( (void**) hs_fn) ;
  free( (void*) nRow) ;
  free( (void*) nCol) ;

  exit(0) ;  

Bad:
  free_mtrx( (void**) org_fn) ;
  free_mtrx( (void**) hs_fn) ;
  free( (void*) nRow) ;
  free( (void*) nCol) ;
  
  exit(-1) ;  
}




void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");
  fprintf(stderr,"     -N # of EM iterations -- >0, default value is 10.\n");
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");
  fprintf(stderr,"     		    > 15  output segmetation of all scales\n");

  exit(1);
}


