#include "../../utilities/src/util.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>


#define MIN_LOG  0.0000000001
#define MAX_EM_ITERATIONS 30
#define MIN_CHANGES 0.0005

/*********************************************************************/
/*  DecimateFactor: due to the feature extraction of data model */
/*     there will be a DecimateFactor.  When "WaveLetNoMean" is used */
/*     the DecimateFactor is 1, which means that  */
/*     the number of resolution allowed will be substracted by  */
/*     1,  and the level 1 will be used as "the new finest resolution". */
/*********************************************************************/
#define DECIMATE_FACTOR   1

#define FALSE 0
#define TRUE  1

/******************************************************************************/
/* Quadtree parameters */
/******************************************************************************/

#ifndef QTR_DEF
#define QTR_DEF

struct Par_QTR
{
   int nclasses ;
   int levels ;           /* number of resolutions */
   double ***Alpha0 ;     /* trasition para between pixel and
                             its father [level][fine_class][coarse_class] */
} ;  
#endif  /* QTR_DEF */


/* BigPyramid */

struct BigPyramid
{
   /* basic parameters */
   int level ;     /* number of levels of the BigPyramid */
   int *nRow ;     /* number of rows of each level */
   int *nCol ;     /* number of columns of each level */

   /* segmentation related parameters */
   double ****LProb ;      /* log probability of Xn given Yn
                              _LProb[level][j][k][class] */
   double ****LLikeH ;     /* LLikeH[level][j][k][class], log likelihood*/

   unsigned char ***_Segment ;     /* _Segment[d][i][j], segment at
                                      level d, row i, column j */
} ;


/* EM_u.c */

void free_bigPyramid (struct BigPyramid *bigPyramid) ;
void OutputDecimatedSeg(struct BigPyramid *Pyramid, char *fn, 
    char *String, int DecimateFactor) ;


/* EM_alpha.c */

void Initial_QuadTree( struct Par_QTR *QTr_par ) ;
void ReadAlpha( struct Par_QTR *QTr_par , char *alpha_fn) ;
void free_Par_QTR( struct Par_QTR *QTr_par ) ;
void EM_WriteQuadTreePar(FILE *fd, struct Par_QTR *QTr_par, 
    int DecimateFactor ) ;
void EM_InitialPyramid(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par, 
    char **hs_fn, int nFiles) ;


/* EM_sample.c */

void EM( 
  struct BigPyramid *bigPyramid,
  struct Par_QTR *QTr_par ,
  int nM ,	/* number of iteration to generate a sample */
  int nN ,	/* number of iteration after a sample is generated */
  int N ,	/* number of EM updates */
  int nF, 	/* number of sample images */
  char* otfn 	/* output file name */
) ;



/* Dec_u.c */

void Decimate(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par) ;


/* train_u.c */

int read_training_images( FILE *fd, int nClasses, int nFiles, char **org_fn, 
    char **hs_fn, int *nRow, int *nCol ) ;
int number_resolutions_allowed( int DD, int *nRow, int *nCol, int nFiles ) ;



