#include "EM_Dec.h"
extern int vlevel ;

/***************************************************************************/
/*  randomly initialized a the segmentation of pyramid at all resolutions, */ 
/*  except the finest resolution */
/***************************************************************************/
static void RandInitialSeg(struct BigPyramid *Pyramid, 
    struct Par_QTR *QTr_par)     
{
  int i, i1, j1 ;
  int nRow,nCol,M, D;
  unsigned char ***SegPt ;
  unsigned char **Segment ;
  
  
  /* Initial random number generator  RandInitial() */

  /* generate random labels for the segmentation */
  M = QTr_par->nclasses ;
  D = QTr_par->levels ;
  SegPt = (Pyramid->_Segment) ;
  
  if(vlevel >=30) printf("\tRandom labeling initial segmentation\n") ;
  
  if(M >255) {
      fprintf(stderr,"number of classes should be less than 256.\n") ;
      exit(-10) ;
  }
  
  /* do not need to initialize the finest resolution */
  for(i=D-1; i>=1; i--) { 
  
      nRow = (Pyramid->nRow)[i] ;
      nCol = (Pyramid->nCol)[i] ;
      Segment = (unsigned char **) get_img(nCol,nRow,sizeof(unsigned char)) ;
      SegPt[i] = Segment ;
      
      for(i1=0; i1<nRow; i1++)
      for(j1=0; j1<nCol; j1++)  Segment[i1][j1] = rand() % M ;
       
  }
}



/***************************************************************************/
/***************************************************************************/
static int EM_levels(int nrow, int ncol)
{
    int DD;

    DD=0;
    while((nrow>2)&&(ncol>2)) {
      DD++;
      nrow = nrow/2 + nrow%2;
      ncol = ncol/2 + ncol%2;
    }
    return(DD);
}


  


/***************************************************************************/
/***************************************************************************/
static void EM_ConstructPyramid (int D, int nrow, int ncol, 
  struct BigPyramid *bigPyramid)
{
   int dd ;
   int i, row, col ;
   
   if(vlevel >10) printf("\tConstructing Pyramid ...\n") ;
       
   dd = EM_levels(nrow,ncol);  /* number of levels in the pyramid, 0 to D-1 */
   if(dd<D) {
       printf("number of resolution inputed are too large.\n") ;
       printf("D %d, dd %d\n", D, dd) ;
       fprintf(stderr, "number of resolution inputed are too large.\n") ;
       exit(-1) ;
   }
     
   if(vlevel>10) printf(" levels = %d \n", D) ;
   
   bigPyramid->level = D ;

   bigPyramid->nRow = (int*) mget_spc(D,sizeof(int)) ;
   bigPyramid->nCol = (int*) mget_spc(D,sizeof(int)) ;

   bigPyramid->LProb = (double****) mget_spc(D,sizeof(double***)) ;
   bigPyramid->LLikeH = (double****) mget_spc(D,sizeof(double***)) ;
   bigPyramid->_Segment = (unsigned char***) mget_spc(D,sizeof(int**)) ;
   
   row = nrow ;  
   col = ncol ;  
   
   for(i=0; i<D; i++) {
       (bigPyramid->nRow)[i] = row ;
       (bigPyramid->nCol)[i] = col ;
       if(vlevel>1) printf(" %d  row = %d,  col = %d \n",i, row, col) ;
       row = row/2 + row%2 ;  /* decimate by 2 ONLY */
       col = col/2 + col%2 ; 

       (bigPyramid->LProb)[i] = NULL ;
       (bigPyramid->LLikeH)[i] = NULL ;
       (bigPyramid->_Segment)[i] = NULL ;
  }
}
    


/***************************************************************************/
/***************************************************************************/
void EM_InitialPyramid(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par,
    char **hs_fn, int nFiles)
{
  eisl_img img_in;      /* input image,  tiff or pxm format */
  int i ;
  
  for(i=0; i<nFiles; i++) {	/* start from 1 not 0 */
  	if(vlevel>0) printf("\treading ground truth segmentation %s\n", hs_fn[i]);
	if( ! read_img (hs_fn[i], &img_in) ) {
        fprintf(stderr,"ERROR: Cannot read image %s!\n", hs_fn[i]);
        fprintf(stderr,"ERROR: EM_InitialPyramid\n") ;
        	exit(-1) ;
	}
    if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
        fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
        fprintf(stderr," a pallete image, %s, EM_InitialPyramid\n", hs_fn[i]) ;
        exit(-1) ;
    }
  
  	/* construct the pyramid,     EM_pyramid.c */
    EM_ConstructPyramid(QTr_par->levels, img_in.height, img_in.width,
        &(Pyramid[i])) ;
   	
    RandInitialSeg(&(Pyramid[i]), QTr_par) ;	/* EM_sample.c */
    if( ((Pyramid[i])._Segment[0]) != NULL) {
  	free_img((void **) ((Pyramid[i])._Segment[0])) ;
    }
    (Pyramid[i])._Segment[0] = img_in.mono; 
    /* set X[0] to be the hand segmentation */


    /* added the following to fix a memory leak.  JLH  18 Dec 02 */
    /* need to de-allocate colormap if image was palette type */
    if( img_in.img_type == 'p' ) {
      free( (void *)(img_in.rmap) );
      free( (void *)(img_in.gmap) );
      free( (void *)(img_in.bmap) );
    }
    /* end of addition */

  }

}



/******************************************************************************/
/*  free BigPyramid */
/******************************************************************************/
void free_bigPyramid (struct BigPyramid *bigPyramid)
{
  int ii, DD ;
  
  DD = bigPyramid->level ;
  
  for(ii=0; ii<DD; ii++) {
      if( ((bigPyramid->LProb)[ii]) != NULL ) 
          multifree( (void*) ((bigPyramid->LProb)[ii]), 3) ;
      if( ((bigPyramid->LLikeH)[ii]) != NULL ) 
          multifree( (void*) ((bigPyramid->LLikeH)[ii]), 3) ;
      if( ((bigPyramid->_Segment)[ii]) != NULL ) 
          free_img( (void **) ((bigPyramid->_Segment)[ii]) ) ;
  }

  if( (bigPyramid->nRow) != NULL )  free( (void *) (bigPyramid->nRow) ) ;
  if( (bigPyramid->nCol) != NULL )  free( (void *) (bigPyramid->nCol) ) ;
  if( (bigPyramid->LProb) != NULL )  free( (void *) (bigPyramid->LProb) ) ;
  if( (bigPyramid->LLikeH) != NULL )  free( (void *) (bigPyramid->LLikeH) ) ;
  if( (bigPyramid->_Segment) != NULL )  free( (void *) (bigPyramid->_Segment) );

}



  
/*****************************************************************************/
/*  same as what is in ../../segment/src/seg_io.c */
/*****************************************************************************/
static int EM_Dec_write_segmentation(unsigned char **seg, int height, 
    int width, char *fn)
{
  static eisl_img img ;
  static int FirstFlag=1 ;
  static int KK ;
  int i ;
  
  if(FirstFlag) {   /* generate color map when first using this function */
      img.bitspersample = 8 ;
      KK = (1<<(img.bitspersample)) ;    /* 2^bitspersample */
      img.rmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      img.gmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      img.bmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
      if( (img.rmap == NULL) || (img.gmap == NULL) ||
          (img.bmap == NULL) ) {
        fprintf(stderr, "ERROR: Can not allocate color map, write_segment\n");
        return(FALSE) ;
      }

      /* Has to be initialied.  o.w. may cause confusion between */
      /* 8-bit and 16-bit color map,  randomly initialization */
      for(i=0; i<KK; i++) {  
          img.rmap[i] = rand() % KK ;
          img.gmap[i] = rand() % KK ;
          img.bmap[i] = rand() % KK ;
      }
      
      /* Since the number of classes are usually less than 10 */
      /* manually set the first 8 colors */
      img.rmap[0]=255;  img.gmap[0]=  0;  img.bmap[0]=  0;  
      img.rmap[1]=  0;  img.gmap[1]=255;  img.bmap[1]=  0;
      img.rmap[2]=  0;  img.gmap[2]=  0;  img.bmap[2]=255;  
      img.rmap[3]=255;  img.gmap[3]=255;  img.bmap[3]=255;  
      img.rmap[4]=255;  img.gmap[4]=  0;  img.bmap[4]=255;  
      img.rmap[5]=  0;  img.gmap[5]=255;  img.bmap[5]=255;  
      img.rmap[6]=255;  img.gmap[6]=255;  img.bmap[6]=  0;  
      img.rmap[7]=  0;  img.gmap[7]=  0;  img.bmap[7]=  0;  
  }
  
  
  img.height = height ;
  img.width = width ;
  img.img_type = 'p' ;     /* palette image */
  img.img_format = 't' ;   /* tiff image */
  img.mono = seg ;
  
  if( ! write_img (fn, &img, 't', "none") ) { 
       fprintf(stderr,"ERROR: Cannot write out segment %s.\n",fn);
       return(FALSE) ;
  }

  /* added to fix memory leak.   JLH   18 Dec 02 */
  /* FirstFlag always = 1, so the rmap, gmap, bmap get allocated */
  /* every time this function is called.  */
  free ( (void *)(img.rmap) ), img.rmap = NULL;
  free ( (void *)(img.gmap) ), img.gmap = NULL;
  free ( (void *)(img.bmap) ), img.bmap = NULL;
  /* end of addition */
  
  return(TRUE) ;           

}




/******************************************************************************/
/*  DecimateFactor: due to the feature extraction */
/*     there will be a DecimateFactor.  When DecimateFactor>0,  */
/*     the number of resolution allowed will be substracted by  */
/*     DecimateFactor,  and the level DecimateFactor will be used */
/*     as "the finest resolution", level 0. */
/******************************************************************************/
void OutputDecimatedSeg(struct BigPyramid *Pyramid, char *fn, 
    char *String, int DecimateFactor)
{
  int nRow,nCol, D;
  int i ;
  unsigned char ***SegPt ;
  char fname[80] ;
  
  if(vlevel>0) printf("\tSave decimated ground truth\n") ;

  SegPt = Pyramid->_Segment ;
  D = Pyramid->level ;
  
  /* save the segmentation */    
  for(i=DecimateFactor; i<D; i++) {
    nRow = (Pyramid->nRow)[i] ;
    nCol = (Pyramid->nCol)[i] ;
    
    if((SegPt[i])==NULL) {
        fprintf(stderr, "WARNING:  no segment at level %d \n ", i) ;

    } else {

        sprintf(fname,"%s%s.%d",fn,String,i-DecimateFactor) ;
        if( ! EM_Dec_write_segmentation(SegPt[i], nRow, nCol, fname) ) { 
            fprintf(stderr,"ERROR: Cannot write %s, exit.\n",fname);
            exit (-10);
        }            
        if(vlevel >=3) fprintf(stderr, "\t\t%s is written out\n",fname) ;
    }
  }
}    
       



