#include "EM_Dec.h"

extern int vlevel ;

/**************************************************************************/
/*	Initialization of quadtree par */
/**************************************************************************/
void Initial_QuadTree( struct Par_QTR *QTr_par )
{
    int m1, m2, k ;
    int nClasses ;
    double p, q ;

    nClasses = QTr_par->nclasses ;
    
    if( nClasses > 1) {
    	q = log( 0.7 );
    	p = log( (1-0.7 ) / ( nClasses -1 ) );
    } else {
        q = 0.0 ;   /* log(1) */
        p = log( MIN_LOG ) ;
    }

		/* this is what was originally here.*/
		/* I have modified because it's indexing past where */
		/* (QTr_par->Alpha0)[k] has been allocated.  17 Dec 02   JLH */
		/*
    for(k=1; k<=QTr_par->levels; k++) {
        (QTr_par->Alpha0)[k] = (double**) get_mtrx(nClasses, nClasses,
            sizeof(double)) ;
		*/

		/* beginning of lines I added/changed  18 Dec 02   JLH */
		/* added (QTr_par->Alpha0[0]) = NULL so can de-allocate QTr_par */
		/* without seg faulting.  */
		(QTr_par->Alpha0)[0] = NULL;
    for(k=1; k<QTr_par->levels; k++) {
		/* end of lines I added/changed  18 Dec 02   JLH */
        (QTr_par->Alpha0)[k] = (double**) get_mtrx(nClasses, nClasses,
            sizeof(double)) ;
  		for(m2=0; m2<nClasses; m2++)   /* classes at coarse resolution */
  		for(m1=0; m1<nClasses; m1++) { /* classes at fine resolution */
  			if(m1==m2) (QTr_par->Alpha0)[k][m1][m2]=q ;
			else (QTr_par->Alpha0)[k][m1][m2]=p;
  		}
  	}
}


/**************************************************************************/
/*	Read and initialization of alpha0 */
/**************************************************************************/
void ReadAlpha( struct Par_QTR *QTr_par, char *alpha_fn ) 
{
  FILE *fd ;
  char fn[128] ;
  int m1,m2, i, index;
  int nClasses ;
  double tmp ;
  double ***alpha ;

  sprintf(fn,"%s.qtr",alpha_fn) ;

  if(vlevel>3) printf("\treading %s\n",fn) ;

  fd = fopen(fn,"r") ;
  if(fd==NULL) FileOpenErr(fn) ;
  
  fscanf(fd, "nClasses=%d\n", &(QTr_par->nclasses)) ;
  fscanf(fd, "levels=%d\n", &(QTr_par->levels)) ;
  
  (QTr_par->Alpha0)=(double ***) mget_spc(QTr_par->levels, sizeof(double*)) ;
  alpha = QTr_par->Alpha0 ;
  nClasses = QTr_par->nclasses ;
    
  for(i=1; i<QTr_par->levels; i++) {

    if(feof(fd)) FeofErr("ReadAlpha()") ;
    fscanf(fd,"level=%d\n",&index) ;

  	if(index != i) {
  		fprintf(stderr,"ERROR: Wrong format for quadtree par.\n") ;
  		exit(-10) ;
  	}
  	
  	alpha[i] = (double**) get_img(nClasses,nClasses,sizeof(double)) ;
  	
  	for(m2=0; m2<nClasses; m2++)  /* classes at coarse resolution */
  	for(m1=0; m1<nClasses; m1++) { /* classes at fine resolution */

  		if(feof(fd)) FeofErr("ReadAlpha()") ;
  		fscanf(fd, "%lf", &tmp) ;
 		if(tmp != 0.0)
	  		alpha[i][m1][m2] = tmp ; 
	  	else {
	  		alpha[i][m1][m2] = MIN_LOG ; 
	  	}
	}
  }
  fclose(fd) ;
}




/*************************************************************************/
/*	write out QTr_par */
/*  DecimateFactor: due to the feature extraction */
/*     there will be a DecimateFactor.  When DecimateFactor>0,  */
/*     the number of resolution allowed will be substracted by  */
/*     DecimateFactor,  and the level DecimateFactor will be used */
/*     as "the finest resolution", level 0. */
/*************************************************************************/
void EM_WriteQuadTreePar(FILE *fd, struct Par_QTR *QTr_par, 
    int DecimateFactor ) 
{
  int m1,m2, i, j;
  double ***alpha ;

  if(vlevel>10) printf("\twritting quadtree parameters\n") ;
  
  alpha = QTr_par->Alpha0 ;
  
  fprintf(fd, "nClasses=%d\n", QTr_par->nclasses) ;
  fprintf(fd, "levels=%d\n", QTr_par->levels -DecimateFactor ) ;

	/* question:  should this really start with i=1?  similar problem existed */
	/*   in Initial_QuadTree()     17 Dec 02   JLH*/
  for(i=1; i<(QTr_par->levels -DecimateFactor); i++) {
  	fprintf(fd, "level=%d\n", i) ;
  	for(m2=0; m2<(QTr_par->nclasses); m2++) { /* classes at coarse resolution */
  	  for(m1=0; m1<(QTr_par->nclasses); m1++) { /* classes at fine resolution */
  	      j=i+DecimateFactor ;
  	      fprintf(fd,"   %lf", alpha[j][m1][m2]) ;
  	  }
 	  fprintf(fd,"\n") ;
  	}
  }
}


/******************************************************
	free alpha
******************************************************/
void free_Par_QTR( struct Par_QTR *QTr_par )
{
  int i ;
  double ***alpha ;
  
  alpha = QTr_par->Alpha0 ;
  
  for(i=1; i<QTr_par->levels; i++) {
      if( alpha[i] != NULL ) free_img( (void **) alpha[i] ) ;
  }
  
  free( (void *) alpha) ;
}




