#include "EM_Dec.h"
extern int vlevel ;


/**********************************************************************/
/*    Calculate the density X(K) given X(k!=K), Y */
/**********************************************************************/
static void Density(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par, 
    double *ll, int ii, int jj, int kk, int finestLevel) 
{
  int nRow,nCol,M,D;
  int *nRowPt,*nColPt;
  int i,m,i1,j1, i2,j2 ;
  double MaxLL, tmp, sum ;
  unsigned char ***Seg, **SegSon ;
  unsigned char segFather ;
  int fatherR, fatherC, son, sonR, sonC ;
  double ***alpha ;
  
  
  M = QTr_par->nclasses ;
  D = QTr_par->levels ;
  alpha = QTr_par->Alpha0 ;
  
  nRowPt = Pyramid->nRow ;
  nColPt = Pyramid->nCol ;
  Seg = Pyramid->_Segment ;


  /* initialization */
  for(m=0; m<M; m++) ll[m] = 0. ;

  
  if(kk<D-1) {        /* not the coarsest level */
      nRow = nRowPt[kk+1] ;
      nCol = nColPt[kk+1] ;
      fatherR = ii/2 ;
      fatherC = jj/2 ;
      segFather = Seg[kk+1][fatherR][fatherC] ;
      
    
      for(m=0; m<M; m++) {    /* varying current pixel class */
          tmp = alpha[kk+1][m][segFather] ;  
          ll[m] += tmp ;    /* ll[m]>0 means prob[m]==0 */
      }  
  }
  
  

  if(kk>finestLevel) {        /* not the finest level */
  
      SegSon = Seg[kk-1] ;
      nRow = nRowPt[kk-1] ;
      nCol = nColPt[kk-1] ;
      
      sonR = ii*2 ;
      sonC = jj*2 ;
      
      for(i1=0; i1<2; i1++)
      for(j1=0; j1<2; j1++) {
       
         i2 = sonR +i1 ;
         j2 = sonC +j1 ;
          
         if((i2<nRow) && (j2<nCol)) {
     
              son = SegSon[i2][j2] ;

              for(m=0; m<M; m++) {
                  tmp = alpha[kk][son][m] ;  
                  ll[m] += tmp ;    /* ll[m]>0 means prob[m]==0 */
              } 
         }
     } 
      
  }      
        
 
  
 /* normalize the log likelihood */
  MaxLL = 0.00 ;  
  i = 1 ;
  for(m=0; m<M; m++) {
      if(ll[m] <= 0)    {    /* 999 denote 0 */
          if(i) {
              MaxLL = ll[m] ; 
              i = 0 ;
          }
          if(MaxLL<ll[m]) MaxLL = ll[m] ;
      } 
  }
  
  sum = 0.0 ;
  for(m=0; m<M; m++) {
      if(ll[m] <= 0) {
          ll[m] = exp(ll[m]-MaxLL) ;
          sum += ll[m] ;
      } else  ll[m] = -999 ;
  }
  
  
  if(sum==0) {
      for(m=0; m<M; m++) {
          printf("Warning:  Density, sum=0, %lf\n", ll[m]) ;
          ll[m] = 1/M ;
      }
  } else {
      
      for(m=0; m<M; m++) {
          if(ll[m] >= 0 ) {
              ll[m] = ll[m] / sum ;
          } else  ll[m] = 0.0 ;
      }
  }

}  




/********************************************************************** */
/*    Calculate an example according to density ll  */
/**********************************************************************/
static int GenSampDensity(double *ll, int M)
{ 
  double tmp, sum ;
  int i ;
  
  tmp = random2() ;
  
  sum = 0. ;
  for(i=0; i<M; i++) {
      sum += ll[i] ;
      if((sum >= tmp)&&(ll[i]!=0)) return(i) ;
  }
  
  fprintf(stderr, "Error,  Density ll not sum to 1") ;
  fprintf(stderr, "sum=%lf, rand=%lf, sample.c\n", sum, tmp) ;
  for(i=0; i<M; i++) {
      fprintf(stderr, "ll[%d]=%lf \n", i, ll[i]) ;
  }
  exit(-1) ;
 
}  




/********************************************************************** */
/*    Generate a sample of X given Y, transition prob and  */
/*    Gaussian mixture model,  not used for quadtree par estimation */
/**********************************************************************/

static void GenerateSample(struct BigPyramid *Pyramid, struct Par_QTR *QTr_par,
    int N, char* fn, char* otfn) 
{
  int nRow,nCol,M,D;
  int *nRowPt, *nColPt ;
  int k, i1,j1,k1 ;
  double *ll;  /* log likehood at each pixel, ll[class] */
  unsigned char ***Seg ;
  char name[80] ;
  
  
  if(vlevel>3) printf("      Generate a Sample of X given Y ...... \n") ;

  M = QTr_par->nclasses ;
  D = QTr_par->levels ;
  nRowPt = Pyramid->nRow ;
  nColPt = Pyramid->nCol ;
  Seg = Pyramid->_Segment ;
  
  ll = (double*) mget_spc(M,sizeof(double)) ;
  
   
  for(k=0; k<N; k++) {    /* iterations */
    printf("\t%s    %d :    %d \n ", fn, k, N) ;
        

    /* simulating by adding fine resolutions one by one */
    
     if(k<31) {
      for(k1=2; k1<D; k1++)    {    /* each resolution */

          nRow = nRowPt[k1] ;
          nCol = nColPt[k1] ;
          
          for(i1=0; i1<nRow; i1++)
          for(j1=0; j1<nCol; j1++) {
          
              Density(Pyramid,QTr_par,ll,i1,j1,k1, 2) ;
              Seg[k1][i1][j1] = GenSampDensity(ll,M) ;
          }
      }
      
     } else  if(k<61) {
   
      for(k1=1; k1<D; k1++)    {    /* each resolution */

          nRow = nRowPt[k1] ;
          nCol = nColPt[k1] ;
          
          for(i1=0; i1<nRow; i1++)
          for(j1=0; j1<nCol; j1++) {
          
              Density(Pyramid,QTr_par,ll,i1,j1,k1, 1) ;
              Seg[k1][i1][j1] = GenSampDensity(ll,M) ;
          }
      }
      
     }  else {
   
      for(k1=0; k1<D; k1++)    {    /* each resolution */

          nRow = nRowPt[k1] ;
          nCol = nColPt[k1] ;
          
          for(i1=0; i1<nRow; i1++)
          for(j1=0; j1<nCol; j1++) {
          
              Density(Pyramid,QTr_par,ll,i1,j1,k1, 0) ;
              Seg[k1][i1][j1] = GenSampDensity(ll,M) ;
          }
      }
     }
     
      
     if(((k%20) == 0) && (k!=0)) {
      sprintf(name,"m%d",k) ;
      OutputDecimatedSeg(Pyramid, otfn, name, 0) ;
     }
          
  }
  
  free((void*) ll) ;
  
  if(vlevel>3) printf("      Done.\n") ;
}


/**********************************************************************/
/**********************************************************************/
static void Alpha_Reset(double ***alpha, int M, int DD) 
{
  int i,j,k ;
  
  for(i=0; i<DD; i++)
  for(j=0; j<M; j++)
  for(k=0; k<M; k++)  alpha[i][j][k]=0.0 ;
  
}



/**********************************************************************/
/**********************************************************************/
static void Alpha1_Estimate(double **alpha, unsigned char **Parent,
    int nrF, int ncF, unsigned char **Children, int nrC, int ncC,int M) 
{
  int i,j, i1,j1, i2,j2, i3,j3, mf, mc ;
  
  for(i=0; i<nrF; i++)
  for(j=0; j<ncF; j++) {
  	mf = Parent[i][j] ;
  	
  	if((mf<0) || (mf>=M)) {
  	    fprintf(stderr, " Wrong class number %d, Alpha1_Estimate\n",mf) ;
  	    exit(-1) ;
  	}
  	
  	i2=2*i;  j2=2*j;
  	
  	for(i1=0; i1<2; i1++)
  	for(j1=0; j1<2; j1++) {
  	
  		i3=i2+i1;  j3=j2+1;
  		if((i3>=0)&&(i3<nrC)&&(j3>=0)&&(j3<ncC)) {
  			mc = Children[i3][j3] ;
  			if((mf<0) || (mf>=M)) {
  		    	    fprintf(stderr, " Wrong class number mc%d, Alpha1_Estimate\n",
  				    mc) ;
  				exit(-1) ;
  			}
  		
  			(alpha[mc][mf]) ++ ;
  		}
  	}
  }
}  


/**********************************************************************/
/**********************************************************************/
static void EM_Alpha1_Update(double ***alpha, struct Par_QTR *QTr_par, 
    int M, int DD, char *name, double *avg_diff) 
{
  double sum, tmp, diff ;
  int k, mf,mc ;
  char name_log[256] ;
  FILE *fd, *fd_log ;
  double ***alpha0 ;
  
  alpha0 = QTr_par->Alpha0 ; 
  fd = fopen (name, "w");  if (fd == NULL) FileOpenErr(name) ;
  
  sprintf(name_log, "%s.log", name) ;
  fd_log = fopen (name_log, "w");  
  if (fd_log == NULL) FileOpenErr(name_log) ;
 
  fprintf(fd, "nClasses=%d\n", QTr_par->nclasses) ;
  fprintf(fd, "levles=%d\n", QTr_par->levels ) ;

  fprintf(fd_log, "nClasses=%d\n", QTr_par->nclasses) ;
  fprintf(fd_log, "levles=%d\n", QTr_par->levels ) ;
 
  *avg_diff = 0.0 ; 
  for(k=1; k<DD; k++) {
      diff = 0.0 ;
      fprintf(fd, "level=%d\n", k) ;
      fprintf(fd_log, "level=%d\n", k) ;
     
      for(mf=0; mf<M; mf++) {
          sum = 0.0 ;
  	      for(mc=0; mc<M; mc++)  sum += alpha[k][mc][mf] ;

          for(mc=0; mc<M; mc++)  {

              if(sum !=0) {
                  tmp = alpha[k][mc][mf] / sum ;
  			  } else {
  				  tmp = (double) 1.0 / M ;
  			  }

    	      fprintf(fd,"%lf ", tmp) ;
    	      diff += fabs(tmp - exp(alpha0[k][mc][mf])) ;

  			  if(tmp>0.000001) {
  				  alpha0[k][mc][mf] = log(tmp) ;
  			  } else {
  				  alpha0[k][mc][mf] = -16 ;
  			  }
  			  
    	      fprintf(fd_log,"%lf ", alpha0[k][mc][mf]) ;
  		  }
    	  fprintf(fd,"\n") ;
    	  fprintf(fd_log,"\n") ;
     }
     
     printf("\t%d average diff/parameter %lf, diff %lf \n", k, diff/M/M, diff);
     *avg_diff += diff/M/M ;
  }

  if(DD>1)  *avg_diff /= (DD-1) ;
  
  fclose(fd) ;
  fclose(fd_log) ;
}





/**********************************************************************/
/*	Using EM algorithm to estimate the trasition rate */
/*	Samples are generated from P(X|Y,Z) to estimate E(X|Y,Z) */
/* 	Ground Truth is stored in X(0)  */
/**********************************************************************/

void EM( 
  struct BigPyramid *bigPyramid,
  struct Par_QTR *QTr_par ,
  int nM ,	/* number of iteration to generate a sample */
  int nN ,	/* number of iteration after a sample is generated */
  int N ,	/* number of EM updates */
  int nF, 	/* number of sample images */
  char* otfn 	/* output file name */
)
{
  struct BigPyramid *Pyramid ;
  char name[80] ;
  int nClasses, D ;
  int nRow,nCol,nR1,nC1;
  int *nRowPt, *nColPt ;
  int k, i1,j1,k1, k2, k3, K, n;
  double avg_diff ;
  double d_k2 ;
  double *ll ;  /* log likehood at each pixel, ll[class] */
  double ***Alpha1 ;	/* estimate Alpha1 */
  unsigned char ***Seg, tmp ;
  
  
  
  if(vlevel>3) printf("\tEstimate quadtree parameters using EM algorithm\n") ;
 
  nClasses = QTr_par->nclasses ;
  D = QTr_par->levels ; 
  
  /* constant items,  same for all sample images */
  Pyramid = &(bigPyramid[0]) ;
  ll = (double*) mget_spc(nClasses,sizeof(double)) ;
  Alpha1 = (double***) multialloc(sizeof(double), 3, D, nClasses, nClasses) ;
 

/*  EM update */
for(K=0; K<N; K++) {	/* T number of updates */
  
  if(vlevel>3) printf("\titeration:  %d  out of %d\n", K, N) ;

  Alpha_Reset(Alpha1, nClasses, D) ;
  
  for(n=0; n<nF; n++) {			/* for each image */
    
    Pyramid = &(bigPyramid[n]) ;
    nRowPt = Pyramid->nRow ;
    nColPt = Pyramid->nCol ;
    Seg = Pyramid->_Segment ;
  
    
    for(k=0; k<nM; k++) {	/* generate a sample */	
    
        /* start from 1, k1=0 is now the hand segmentation */
  		for(k1=1; k1<D; k1++)	{	

  			nRow = nRowPt[k1] ;
  			nCol = nColPt[k1] ;

  			for(i1=0; i1<nRow; i1++)
  			for(j1=0; j1<nCol; j1++) {

  				Density(Pyramid,QTr_par,ll,i1,j1,k1, 0) ;
  				tmp = GenSampDensity(ll,nClasses) ;
  				Seg[k1][i1][j1] = (unsigned char) tmp ;
  			}
  		}
    }
    
    
    /* 1st nM iteration is for convergence and  */
    /* the last nN is to estimate the new alpha */ 
    for(k=0; k<nN; k++) {	
 	
        /* start from 1, k1=0 is now the hand segmentation */
        d_k2 = 1 ;
  	for(k1=1; k1<D; k1++)	{	


	   /* In order to generate more samples for coarser resolution */
	   /* We increase the number of samples by a factor of 2 from  */
	   /* fine to coarse resolution */
	   if(k1>1) d_k2 *= 1.5  ;
	   k2 = (int) d_k2 ;

	   for(k3=0; k3<k2; k3++) {  
	   /* generate same number of samples for each level */

  		nRow = nRowPt[k1] ;
  		nCol = nColPt[k1] ;

  		for(i1=0; i1<nRow; i1++)
  		for(j1=0; j1<nCol; j1++) {

  			Density(Pyramid,QTr_par,ll,i1,j1,k1, 0) ;
  			tmp = GenSampDensity(ll,nClasses) ;
  			Seg[k1][i1][j1] = (unsigned char) tmp ;
  		}



  		/* present level is k1-1, parent level is k1 */
  		nR1 = nRowPt[k1-1] ;   
  		nC1 = nColPt[k1-1] ;

 		Alpha1_Estimate(Alpha1[k1],Seg[k1],nRow,nCol,Seg[k1-1],
 		    nR1,nC1,nClasses) ;

  	   }
       }
    }
  
    
    if( ((vlevel%10==6) && K==1) || 
        ((vlevel%10==6) && (K==N-1)) || 
        ((vlevel>200) && (K%3==0)) ) { 
  	 sprintf(name,".tmp.%dsamp%d",K,n) ;
  	 OutputDecimatedSeg(Pyramid, otfn, name, 0) ;
    }

    fflush(NULL) ;
  }
  
  sprintf(name,"%s.tmp.alpha%d.0",otfn,K) ;
  EM_Alpha1_Update(Alpha1, QTr_par, nClasses, D, name, &avg_diff) ;
  
  if(avg_diff < MIN_CHANGES) break ;
 
  fflush(NULL) ;
}  


free( (void *) ll ) ;
multifree( (void *) Alpha1, 3) ;


}


