/* Find and save sample for clutering algorithm */

#include "GMM.h"
/*#include <sys/getopt.h>*/  /* add for getopt*/
#include <stdio.h>
#include <unistd.h>

#define MAX_IMAGE_CLASS 64
#define MAX_GMM_SAMPLES   200000
#define MAX_GMM_CLUSTERS  20

int vlevel ;

void error(char *name);
void RandInitial() ;
static int Percentage_Samples_Selected( int nF, int D, int nclasses, 
    char **hsfn, double **Rate, double Precision ) ;
/* added following line add void part, not sure-GF*/
void GMM_cluster(int init_num_of_subclasses, char *in_fn, char *out_fn, char *option); 

main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fdin, *fd ;
  int  nrow,ncol;             /* image nrow and ncol */
  int  nclasses;              /* number of classes */
  int  D;                     /* number of levels in pyramid */
  struct BigPyramid _BigPyramid;
  struct Par_GMM GMM_par ;
  eisl_img img_in;            /* input image,  tiff or pxm format */
  eisl_img img_hs;            /* input hand_seg image,  tiff or pxm formate */
  char *infn;       	      /* input image file name */
  char **hsfn;       	      /* hand segmentation file name */
  char **imgfn;               /* input img file name */
  char *otfn;                 /* output segmentation file name */
  char filename[80];
  char filename2[80];
  char options[32] ;
  char fn_gmm[80] ;
  char comm[80] ;
  int  nF ;		      /* number of input files */
  int  iflag,oflag;
  int  i,j ;
  int  nRow, nCol ;
  int  m, MaxPerClass ;
  int  init_subclusters ;
  int  **count ;
  double **Rate, Precision ;
  double Noise_Std; /* interval length for UNIF RV, */
                    /* standard deviation for Gaussian RV */

  /* malloc_debug(2); */

  vlevel=1;
  Precision=20.0 ;
  Noise_Std=0.2 ;
  init_subclusters=15 ;
  iflag=oflag=0;
  
  while( (c=getopt(argc,argv,"h,o:,i:,v:,P:,C:,R:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      case 'v':
        if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
        if(vlevel<0) error(argv[0]);
        break;
      case 'P':
        if(0==sscanf(optarg,"%lf",&Precision)) error(argv[0]);
        break;
      case 'C':
        if(0==sscanf(optarg,"%d",&init_subclusters)) error(argv[0]);
        if(init_subclusters < 1) {
            fprintf(stderr, "ERROR: initial number of subclusters") ;
            fprintf(stderr, " must be greater than 0\n") ;
            error(argv[0]) ;
        }
        break;
      case 'R':
        if(0==sscanf(optarg,"%lf",&Noise_Std)) error(argv[0]);
        if(Noise_Std < 0) {
            fprintf(stderr, "ERROR: the standar deviation of an RV") ;
            fprintf(stderr, " must be greater than 0\n") ;
            error(argv[0]) ;
        }
        break;
      default:
        error(argv[0]);
        break;
    }
  }
  
  if(vlevel>1) printf("\tGenerate samples for parameter estimation of GMM\n");

  /* Check for inconsistant inputs */   
  if(!(oflag&&iflag)) {
    fprintf(stderr, "Input, output filenames are required\n") ;
    error(argv[0]);
  }
     

  if(Precision<0) {
      fprintf(stderr, "WARNING: Invalid Precision %f\n", Precision) ;
      fprintf(stderr, "WARNING: Precision set to 20\n") ;
      Precision=20.0 ;
  }
  if(Precision>100) Precision=100.0 ;
  Precision /= 100.0 ;
  /* MaxPerClass = MAX_GMM_SAMPLES * Precision / 100 ; */

  
  /* read input file names */
  if(vlevel >0) printf("\treading input images ...... \n") ;
  fdin = fopen (infn, "r");  if (fdin == NULL) FileOpenErr(infn) ;
  fscanf(fdin, "nFiles=%d\n", &nF) ;
  fscanf(fdin, "nClasses=%d\n", &nclasses) ;
  if((nF<1) || (nclasses<2)) {
      fprintf(stderr, "ERROR: invalid par\n") ;
      fprintf(stderr, "\tnFiles %d (<1?)\n", nF) ;
      fprintf(stderr, "\tnClasses %d (<2?)\n", nclasses) ;
  }

  /* write out input parameters */
  if(vlevel>0) {
       printf("\tPrecision %lf, Init_#_clusters %d, nFiles %d, nClasses %d\n",
           Precision, init_subclusters, nF, nclasses) ;
  } 

  /* Check for too many classes */
  if(nclasses>MAX_IMAGE_CLASS) {
    fprintf(stderr, "number of classes must be <= %d\n", MAX_IMAGE_CLASS);
    exit(1);
  }
  hsfn = (char **) get_mtrx(nF, 256, sizeof(char)) ;
  imgfn = (char **) get_mtrx(nF, 256, sizeof(char)) ;
  
  for(i=0; i<nF; i++) {
  	fscanf(fdin, "%s", imgfn[i]);  printf("\t%s\n", imgfn[i]) ;
  	fscanf(fdin, "%s", hsfn[i]);   
  	sprintf(hsfn[i], "%s.tmp.%d", otfn, i);   /* default hand_seg filename */
  	printf("%s\n", hsfn[i]);
  }
  fclose(fdin) ;

  
  /* read number of levels allowed */ 
  sprintf(filename, "%s.qtr", otfn) ;
  if(vlevel>0) printf("\treading %s\n", filename) ;
  fdin = fopen (filename, "r");  if (fdin == NULL) FileOpenErr(filename) ;
  fscanf(fdin, "nClasses=%d\n", &m) ;
  fscanf(fdin, "levels=%d\n", &D) ;
  if(vlevel>0) printf("\tnClasses=%d, levels=%d\n", m, D) ;
  fclose(fdin) ;
  if(m!=nclasses) {
      fprintf(stderr, "ERROR: inconsistance nclasses %d, %d\n",nclasses,m) ;
      goto Bad ;
  }
   
  GMM_par.nclasses = nclasses ;
  GMM_par.MaxLevels = D ;
  sprintf(GMM_par.feature_name, "WaveletNoMean") ;
  
 
  Rate = (double **) get_mtrx(D, nclasses, sizeof(double)) ;
  count = (int **) get_mtrx(D, nclasses,sizeof(int)) ;
  
  
  
  /* Initalization */
  
  if(! Percentage_Samples_Selected(nF, D, nclasses, hsfn, Rate, Precision)) {
      goto Bad ;
  }
  RandInitial() ;
  
  for(i=0; i<D; i++) 
  for(j=0; j<nclasses; j++)  count[i][j]=0 ;

  /* delete old sample files, because fopen(,"a") is used for */
  /* saving samples */
  sprintf(fn_gmm, "%s.tmp.gmm", otfn) ;
  sprintf(comm, "rm -f %s.tmp.gmm*", otfn)  ;
  system(comm) ;
  

  /* read and initialize other images */
  for(i=0; i<nF; i++) {	/* start from 1 not 0 */

  	/* read the first image and initialization */
  	if(vlevel>1) printf("\tProcess %s\n", imgfn[i]) ;
	if( ! read_img (imgfn[i], &img_in) ) {
        fprintf(stderr,"ERROR: Cannot read %s, GMM!\n", imgfn[i]);
        goto Bad ;
	}
	if( img_in.img_type != 'g' ) {
            fprintf(stderr,"ERROR: original image must be a gray level") ;
            fprintf(stderr," image, %s, GMM\n", filename) ;
            goto Bad ;
	}
  

  	/* Feature Initialization		feature.c */
  	FeatureInitial(&img_in, &GMM_par, &ncol, &nrow) ;  
  	if(vlevel>10) fprintf(stdout, "nrow=%d, ncol=%d\n", nrow, ncol) ;

  	/* construct the pyramid */
  	GMM_ConstructPyramid(nrow,ncol,&_BigPyramid,D) ;    /* u_pyramid.c */

  	if(vlevel>0) PrintFeatureCh(&GMM_par) ;	/* feature.c */
  
  	
        for(j=0; j<D; j++) {
    	    sprintf(filename,"%s.hs.%d",hsfn[i],j) ;
  	    if(vlevel >0) printf("\treading hand segmentation %s .\n", filename) ;
	    if( ! read_img (filename, &img_hs) ) {
        	fprintf(stderr,"ERROR: Cannot read %s!\n", filename);
        	goto Bad ;
	    }
    	    if( img_hs.img_type != 'g' && img_hs.img_type != 'p' ) {
        	fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
        	fprintf(stderr," a pallete image, %s, GMM\n", filename) ;
        	goto Bad ;
    	    }

  	    nRow = _BigPyramid.nRow[j] ;
  	    nCol = _BigPyramid.nCol[j] ;
  	    if((nRow != img_hs.height)||(nCol != img_hs.width)) {
  		   fprintf(stderr,"\n Dimension of ground truth %s", filename) ;
  		   fprintf(stderr," confilcts with image dimesion.\n");
  		   fprintf(stderr," %d!=%ld, %d!=%ld.\n",nRow,img_hs.height,
  		       nCol,img_hs.width);
  		   exit (-10);
  	    }

            /* set X[i] to be the hand segmentation */
  	    _BigPyramid._Segment[j] = img_hs.mono; 
   	}
  

  	/* subsampling rate = 2 */
  	SaveFeature(&_BigPyramid, &img_in, &GMM_par, fn_gmm, Rate, 
  	    count, Noise_Std); 
  	free_img((void **) img_in.mono) ;
  	GMM_free_bigPyramid( &_BigPyramid ) ;
  }


  
  /* saving info_par files */
  for(i=0; i<D; i++) {
  	sprintf(filename, "%s.tmp.gmm.s%d.info",otfn,i) ;
  	fdin = fopen (filename, "w");  if (fdin == NULL) FileOpenErr(filename) ;
  	
  	fprintf(fdin,"%d\n",nclasses) ;
  	fprintf(fdin,"%d\n",GMM_par.dimen) ;
  	
  	for(j=0; j<nclasses; j++) {
  		sprintf(filename, "%s.tmp.gmm.s%d.%d",otfn,i,j) ;
  		fprintf(fdin,"%s  %d\n",filename, count[i][j]) ;
  		if(vlevel>10) printf("%d, %d,   %d \n",i,j, count[i][j]) ;
  	}
  	
  	fclose(fdin) ;
  }
  
  
  
  /* clustering */
  /* j = MAX_GMM_CLUSTERS * Precision / 100; */
  sprintf(options, "full") ;
  for(i=0; i<D; i++) {
      sprintf(filename, "%s.tmp.gmm.s%d.info",otfn,i) ;
      sprintf(filename2, "%s.gmm.s%d.par",otfn,i) ;
      GMM_cluster(init_subclusters, filename, filename2, options) ; 	
  }
  
  
  /* save GMM paramters */
  
  sprintf(filename, "%s.gmm",otfn) ;
  if(vlevel>0) printf("\tsave GMM parameters to %s\n", filename) ;
  fd = fopen (filename, "w");  if (fd == NULL) FileOpenErr(filename) ;
  fprintf(fd, "NAME %s\n",  GMM_par.feature_name);
  fprintf(fd, "nbands=%d\n", GMM_par.dimen);
  fprintf(fd, "nClasses=%d\n", GMM_par.nclasses);
  fprintf(fd, "MaxLevels=%d\n", D);
  fclose(fd) ;	
  
  
  /* remove tempary files */
  if(vlevel%10 != 6) {
      sprintf(comm, "rm -f %s.tmp.gmm*", otfn) ;
      system(comm) ; 
  }
 
Good:
  if(Rate != NULL) free_mtrx( (void **) Rate ) ;
  if(hsfn != NULL) free_mtrx( (void **) hsfn ) ;
  if(imgfn != NULL) free_mtrx( (void **) imgfn ) ;
  if(count != NULL) free_mtrx( (void **) count ) ; 
  
  exit(0) ;

Bad:
  if(Rate != NULL) free_mtrx( (void **) Rate ) ;
  if(hsfn != NULL) free_mtrx( (void **) hsfn ) ;
  if(imgfn != NULL) free_mtrx( (void **) imgfn ) ;
  if(count != NULL) free_mtrx( (void **) count ) ; 
  
  exit(-1) ;
}




void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");
  fprintf(stderr,"     		> 15  output segmetation of all scales\n");
  fprintf(stderr,"     -P percentage of samples selected -- double \n");
  fprintf(stderr,"              precision from 0 to 100, default is 20\n");
  fprintf(stderr,"     -R double -- The interval length of a zero mean \n");  
  fprintf(stderr,"              uniform RV which is added to each feature\n");
  fprintf(stderr,"              to regularize the clustering \n") ;
  fprintf(stderr,"              algorithm.  Default is 0.2 \n");
  fprintf(stderr,"     -C number of initial subclusters -- >0\n") ;
  fprintf(stderr,"              default value is 15  ") ;

  exit(1);
}




/******************************************************************/
/*   estimate Rate, the percentage of samples will be written out */
/*   when given the maximal number of samples allowed */
/******************************************************************/
static int Percentage_Samples_Selected(
  int nF, int D, int nclasses, char **hsfn, double **Rate, double Precision
)
{ 
  eisl_img img_hs;            /* tiff or pxm formate */
  char filename[128] ;
  int  i,j, ii,jj, m ;
  double max_rate ;


  for(i=0; i<D; i++) 
  for(j=0; j<nclasses; j++) Rate[i][j]=0.0 ;
  
  /* estimate Rate, the percentage of samples will be written out */
  for(i=0; i<nF; i++) {
      sprintf(filename,"%s.hs.%d",hsfn[i],0) ;
      if(vlevel >0) printf("\treading hand segmentation %s\n", filename) ;
	  if( ! read_img (filename, &img_hs) ) {
          fprintf(stderr,"ERROR: Cannot read %s!\n", filename);
          goto Bad ;
	  }
      if( img_hs.img_type != 'g' && img_hs.img_type != 'p' ) {
          fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
          fprintf(stderr," a pallete image, %s, GMM\n", filename) ;
          goto Bad ;
      }
  	
      for(ii=0; ii<img_hs.height; ii++)
      for(jj=0; jj<img_hs.width; jj++) {
          m = (int) ((img_hs.mono)[ii][jj]) ;
          CheckClassNumber("main()", m, 0,nclasses) ;
          (Rate[0][m]) ++ ;
      }
  		
      free_eisl_img( &img_hs ) ;
  }  	
  
  max_rate = Rate[0][0] ;
  for(i=1; i<nclasses; i++) {
      if( Rate[0][i] > max_rate ) max_rate = Rate[0][i] ;
  }
  
  for(i=0; i<nclasses; i++) {
      if( Rate[0][i] != 0 ) {
          Rate[0][i] = max_rate / Rate[0][i] * Precision ;
      } else {
          Rate[0][i] = 1.0 ;
          fprintf(stderr, "WARNING: class %d has no samples\n", i) ;
      }   
  }
  
  for(i=1; i<D; i++)
  for(j=0; j<nclasses; j++) {
  	  Rate[i][j] = Rate[i-1][j]*4.0;  /* decimation rate from level to level */
      if( Rate[i][j] > 1.0 )  Rate[i][j] = 1.0 ;
      if(vlevel>10) {
           printf("level %d,  class %d,  rate %lf\n", i,j,Rate[i][j]) ;
      } 
  }
  
  return(TRUE) ;

Bad:
  return(FALSE) ;
}



/******************************************************************/
/*   estimate Rate, the percentage of samples will be written out */
/*   when given the maximal number of samples allowed */
/******************************************************************/
static int Estimate_Percentage_Samples_Select_Old(
  int nF, int D, int nclasses, char **hsfn, int MaxPerClass, double **Rate
)
{ 
  eisl_img img_hs;            /* tiff or pxm formate */
  char filename[128] ;
  int  i,j, ii,jj, m ;


  for(i=0; i<D; i++) 
  for(j=0; j<nclasses; j++) Rate[i][j]=0.0 ;
  
  /* estimate Rate, the percentage of samples will be written out */
  for(i=0; i<nF; i++) 
  for(j=0; j<D; j++) {
      sprintf(filename,"%s.hs.%d",hsfn[i],j) ;
      if(vlevel >0) printf("\treading hand segmentation %s\n", filename) ;
	  if( ! read_img (filename, &img_hs) ) {
          fprintf(stderr,"ERROR: Cannot read %s!\n", filename);
          goto Bad ;
	  }
      if( img_hs.img_type != 'g' && img_hs.img_type != 'p' ) {
          fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
          fprintf(stderr," a pallete image, %s, GMM\n", filename) ;
          goto Bad ;
      }
  	
      for(ii=0; ii<img_hs.height; ii++)
      for(jj=0; jj<img_hs.width; jj++) {
          m = (int) ((img_hs.mono)[ii][jj]) ;
          CheckClassNumber("main()", m, 0,nclasses) ;
          (Rate[j][m]) ++ ;
      }
  		
      free_eisl_img( &img_hs ) ;
  }  	
  
  
  for(i=0; i<D; i++)
  for(j=0; j<nclasses; j++) {
      if(Rate[i][j]>MaxPerClass) {
  	  Rate[i][j] = (double) MaxPerClass/Rate[i][j] ;
      } else {
           Rate[i][j] = 1.0 ;
      }
      if(vlevel>10) {
           printf("level %d,  class %d,  rate %lf\n", i,j,Rate[i][j]) ;
      } 
  }
  
  return(TRUE) ;

Bad:
  return(FALSE) ;
}

