#include "GMM.h"

extern int vlevel ;

/******************************************************************************/
/*  calculate the maximal number of levels allowed */
/******************************************************************************/
static int GMM_levels(int nrow, int ncol)
{
    int D;

    D=0;
    while((nrow>2)&&(ncol>2)) {
      D++;
      nrow = nrow/2 + nrow%2;
      ncol = ncol/2 + ncol%2;
    }
    return(D);
}


    	
void GMM_ConstructPyramid (int nrow, int ncol, 
    struct BigPyramid *bigPyramid, int DD)
{
   int D ;
   int i, row, col ;
   
   if(vlevel >1)
   	printf("   Constructing Pyramid ...\n") ;
   	
   D = GMM_levels(nrow,ncol) ; 	/* number of levels in the pyramid, 0 to D-1 */
   if( D > DD ) {
       D=DD ;
   } else if( D< DD ) {
       fprintf(stderr, "ERROR: number of scales %d < required %d\n", D, DD) ;
       exit(-1) ;
   }
   if(vlevel>30) printf(" levels = %d \n", D) ;
   
   bigPyramid->level = D ;

   bigPyramid->nRow = (int*) mget_spc(D,sizeof(int)) ;
   bigPyramid->nCol = (int*) mget_spc(D,sizeof(int)) ;


   bigPyramid->LProb = (double****) mget_spc(D,sizeof(double***)) ;
   bigPyramid->LLikeH = (double****) mget_spc(D,sizeof(double***)) ;
   bigPyramid->_Segment = (unsigned char***) mget_spc(D,sizeof(int**)) ;
   
   for(i=0; i<D; i++) {
       bigPyramid->LProb[i] = NULL ;
       bigPyramid->LLikeH[i] = NULL ;
       bigPyramid->_Segment[i] = NULL ;
   }
   
   row = nrow ;  
   col = ncol ;  
   
   for(i=0; i<D; i++) {
   	(bigPyramid->nRow)[i] = row ;
   	(bigPyramid->nCol)[i] = col ;
   	if(vlevel>1) printf(" %d  row = %d,  col = %d \n",i, row, col) ;
   	row = row/2 + row%2 ;  /* decimate by 2 ONLY */
   	col = col/2 + col%2 ; 
                 		
	(bigPyramid->LProb)[i] = NULL ;
   }
}
	
	
/******************************************************************************/
/*  free BigPyramid */
/******************************************************************************/
	
void GMM_free_bigPyramid (struct BigPyramid *bigPyramid)
{
  int ii, DD ;
  
  DD = bigPyramid->level ;
  
  for(ii=0; ii<DD; ii++) {
      if( ((bigPyramid->LProb)[ii]) != NULL ) 
          multifree( (void*) ((bigPyramid->LProb)[ii]), 3) ;
      if( ((bigPyramid->LLikeH)[ii]) != NULL ) 
          multifree( (void*) ((bigPyramid->LLikeH)[ii]), 3) ;
      if( ((bigPyramid->_Segment)[ii]) != NULL ) 
          free_img( (void **) ((bigPyramid->_Segment)[ii]) ) ;
  }

  if( (bigPyramid->nRow) != NULL )  free( (void *) (bigPyramid->nRow) ) ;
  if( (bigPyramid->nCol) != NULL )  free( (void *) (bigPyramid->nCol) ) ;
  if( (bigPyramid->LProb) != NULL )  free( (void *) (bigPyramid->LProb) ) ;
  if( (bigPyramid->LLikeH) != NULL )  free( (void *) (bigPyramid->LLikeH) ) ;
  if( (bigPyramid->_Segment) != NULL )  free( (void *) (bigPyramid->_Segment) );

}


