#include "GMM.h"

extern int vlevel ;

/**********************************************************
  read images and calculate log probability X given Y
***********************************************************/
void SaveFeature(
  struct BigPyramid* Pyramid, 
  eisl_img* imgPt, 	/* input images */
  struct Par_GMM *GMM_par,   
  char   *otfn ,
  double **Rate,
  int    **count,
  double Noise_Std
)
{
  FILE **fd ;
  char filename[256] ;
  double ***Img ;
  double **Img0=NULL, **Img1=NULL ;   /* org and decimated images */
  double ***ImgPt1, ***ImgPt2, ***ImgPt3 ;   
  unsigned char ***Seg, m ;
  int D, dimen,nClasses, nRow,nCol, i,j,k, i1,j1,k1 ;
  int RateSS; 	    /* subsampling rate */
  int wd, ht, wd1, ht1;
  double Randtmp;	/* save rate for T,B,I,H */


  if(vlevel>100) PrintFeatureCh(GMM_par) ;
  
  D = GMM_par->MaxLevels ;
  if(D>Pyramid->level) D=Pyramid->level ;
  dimen = GMM_par->dimen ;
  RateSS = GMM_par->Rt ;
  nClasses = GMM_par->nclasses ;
  Seg = Pyramid->_Segment ;
  fd = (FILE **) mget_spc(nClasses, sizeof(FILE *)) ;
  Img = (double ***) get_spc(dimen,sizeof(double **));

  wd = imgPt->width ;
  ht = imgPt->height ;
  Img0 = (double **) get_img(wd,ht, sizeof(double)) ;

  Randtmp = 0 ;
  for(i=0; i<ht; i++)
  for(j=0; j<wd; j++) {
/*      Add no noise to image  */
/*      Randtmp = (random2() - .5) *2 *Noise_Std ;  */
/*      Randtmp = normal() * Noise_Std ;  */
      Img0[i][j] = (double) imgPt->mono[i][j] + Randtmp ;
  }
  
  ImgPt2 = &Img0 ;    /* initial for the case i==0 */
  wd1 = wd ;
  ht1 = ht ;
  ImgPt1 = &Img1 ;


  if(vlevel >10) fprintf(stderr, "\t\t") ;
  for(i=0; i<D; i++) {
  
    for(j=0; j<nClasses; j++) {
      sprintf(filename, "%s.s%d.%d",otfn,i,j) ;
      fd[j] = fopen (filename, "a");  
      if (fd[j] == NULL) FileOpenErr(filename) ;
    }
	
    if(vlevel>10) fprintf(stdout, "	%2d ",i) ;
  
  
    if(i!=0) sub_sample((*ImgPt1),wd,ht,RateSS,ImgPt2,&wd1,&ht1) ;

    (*(GMM_par->GenerateFunc))((*ImgPt2),wd1,ht1, Img, 2, &wd, &ht) ;	
  
    nRow = (Pyramid->nRow)[i] ;
    nCol = (Pyramid->nCol)[i] ;
    if((nRow != ht) || (nCol != wd)) {
    	fprintf(stderr, "\n Inconsistant dimension for features !!\n") ;
    	fprintf(stderr, "ht%d,wd%d,   nRow%d, nCol%d !!\n",ht,wd,nRow,nCol) ;
    	exit(-1) ;
    }
    	    	
    /* Randomly output features */
    for(i1=0; i1<nRow; i1++) 
    for(j1=0; j1<nCol; j1++) {
  	
  	m = Seg[i][i1][j1] ;
  	Randtmp = random2() ;

  	if(( ((int) m)< 0) || (((int) m )>= nClasses))  /* m cant be <0 since it was
 unsigned char before and isnt changed GF 3/25/02 */ 
          /* if( (int) m >= nClasses)*/{
  		fprintf(stderr,"Wrong classes %d, nClasses %d, SaveSamp \n",
  		    m, nClasses) ;
  		exit(-1) ;
  	}
  	if(Rate[i][m]>Randtmp) {  

/* a small amount of random noise is added to each sample */
/* this may prevent classes with small variance to be     */
/* eliminated by clustering algorithm as singular class   */ 

if(1)  {   /* add noise to feature vectors */  		
    for(k=0; k<dimen; k++) {
        Randtmp = (random2() - .5) * Noise_Std ;
        fprintf((fd[m]), "%lf  ",Img[k][i1][j1]+Randtmp);
    }
} else {   /* do not add noise to feature, */
           /* but add noise to the whole image */

    for(k=0; k<dimen; k++) {
        fprintf((fd[m]), "%lf  ",Img[k][i1][j1]);
    }
}

  		fprintf(fd[m], "\n") ;
  		(count[i][m]) ++ ;
  	}
    }
    	    	
    for(j=0; j< dimen; j++)  free_img((void **) Img[j]) ;
    
    /* switch org img and decimated img */
    /* if((*ImgPt1) != NULL) {
        free_img( (void **) (*ImgPt1) ) ;  
    } */ /* this will be freed later in sub_sample() GF 3/25/02 */
    ImgPt3 = ImgPt1 ;
    ImgPt1 = ImgPt2 ;
    ImgPt2 = ImgPt3 ;
    wd = wd1;  ht = ht1;
    
    for(j=0; j< nClasses; j++) fclose( fd[j] ) ;
  }
  
  free((void *) Img) ;

  /* added to fix memory leak 16 Jan 2003  JLH */
  free((void *) fd) ;
  /* end of addition  JLH */
  
  if(vlevel >=10) printf("finish saving samples. \n") ;
}
  
  
  
