#include "CPT.h"
#include "CPT_global.h"
#include <math.h>

static int debug = 0 ;


/* For some unknown reason, log2() doesnt work, so we use log() */

/* entropy
%
%  [entp] = entropy(Y, F, nClasses, Show) takes X the class of a pixel,
%  F -- its feature value,  and returns the entropy, Y and F should be 
%  Nx1 column vectors of Integers, 
%  where N is the number of samples.  If Show==1, then intermediate 
%  results will be printed out.
*/
void entropyEntpDrop(int *YY, struct TreeNode *node, int nClasses,
    int RangeF, int Show)
/*    int Show ;*/  /* show intermediate results */
{
    int n0, n1, n2 ;
    int i,j,k, sum_len ;
    double sum1, sum2, entropyOld ;
    double **Hist, *Prior ;   /* to avoid to many allocations, */
                             /* also return Hist and Prior */
    double *EntropyDrop, *Entropy ;  /* return the different before */
                                    /* and after this split */

    
    n0=node->n0;  n1=node->n1;  n2=node->n2;
    if(vlevel >100) printf("n0=%d, n1=%d, nC=%d   Rf=%d\n",n0,n1,nClasses,
        RangeF) ;
    if(n1<=n0) {
    	fprintf(stderr, "Error: n0=%d, >= n1=%d \n", n0,n1) ;
    	exit(-1) ;
    }
    Hist = node->Hist ;
    Prior = node->Prior ;
    EntropyDrop = &(node->costDrop) ;
    Entropy = &(node->entropy) ;

    /* reset Hist and Prior */
    for (i=0; i< RangeF; i++) {
  		Prior[i] = 0 ;
  		for(j=0; j<nClasses; j++) Hist[i][j] = 0 ;
    }
  

    /* calculate the histogram */
    for (k=n0; k<n1; k++) {
    	i=YY[k] ;
    	if((i>=0) && (k<n2)) {	/* >=0 <-> [n0,n2) */
            (Hist[1][i]) ++ ;
    	    (Prior[1]) ++ ;
        } else {
  	    (Hist[0][i]) ++ ;
            (Prior[0]) ++ ;
     	}
    }
    sum_len = Prior[0] + Prior[1] ;
    if(Show >0)  PrintHistPrior(Hist,Prior, nClasses, RangeF) ;
    

    /* Calculate the entropy before this split 				*/
    /* Before split:   H(U) = n1/N ( sum(h_i(1)/n1 * log(h_i(1)/n1) ) 	*/
    entropyOld = 0 ;
    for (i=0; i<nClasses; i++) {
    	sum1 = 0 ;
    	for (j=0; j<RangeF; j++) sum1 += Hist[j][i] ;
    	sum1 = sum1 / sum_len ;
    	if(sum1 > 0) entropyOld -= sum1 * (double) log((double) sum1) ;
    }

        
    /* normalize histogram into conditional prob to calculate entropy */
    for (j=0; j<RangeF; j++)  {
    	if(Prior[j]>0) {
    		for (i=0; i<nClasses; i++)  Hist[j][i] = Hist[j][i] / Prior[j] ;
    		Prior[j] = Prior[j] / sum_len ;
    	}
    }
    if(Show >0)  PrintHistPrior(Hist,Prior,nClasses,RangeF) ;
    

    
    /* Calculate the entropy after this split */
    /* calculate entropy */
    sum2 = 0.0 ;
    for (j=0; j<RangeF; j++) {
    
    	sum1 = 0 ;
    	for (i=0; i<nClasses; i++)  {
    		if(Hist[j][i] >1) {
    			fprintf(stderr, "Probability cannt be greater than 1, %lf.\n", 
    			    Hist[j][i]) ;
    			exit(-1) ;
    		}
    		if(Hist[j][i] > 0) {
    			sum1 += Hist[j][i] * (double) log((double) Hist[j][i]) ;
    		} 
    	}
    	sum2 -= sum1 * Prior[j] ;
    }
    (*Entropy) = sum2 ;
   
   
    /* Calculate the difference in entropy due to this partition */
    /* Before split:   H(U) = n1/N ( sum(h_i(1)/n1 * log(h_i(1)/n1) ) */
    /* After  split:   H(U) = n2/N ( sum(h_i(2)/n2 * log(h_i(2)/n2) )  */
    /*				+ n3/N ( sum(h_i(3)/n3 * log(h_i(3)/n3) ) */
    /* where n1=n2+n3,  h_i(1)=h_i(2)+h_i(3),  n_j=sum(h_i(j)) */
    
    (*EntropyDrop) = entropyOld - sum2 ; 
    if(Show>0 || debug ) printf("entropy %lf,  entropyDrop %lf\n", sum2, (*EntropyDrop)) ;
    if(vlevel>100) printf("entropy %lf,  entropyDrop %lf\n", (*Entropy), (*EntropyDrop)) ;
}





/* simply calcaulate the entropy of vector Y */
double entropy(int *YY, int n0, int n1, int nClasses, int RangeF)
{
    int i,k ;
    double entropyOld=-1 ;
    double sum1 ;
    int *Hist, sum_len ;   

    
    if(vlevel >100) printf("n0=%d, n1=%d, nC=%d   Rf=%d\n",
        n0,n1,nClasses,RangeF) ;
    if(n1<=n0) {
    	fprintf(stderr, "Error: n0=%d, >= n1=%d \n", n0,n1) ;
    	exit(-1) ;
    }
    Hist = get_spc(nClasses, sizeof(double)) ;

    /* reset Hist and Prior */
    for (i=0; i< nClasses; i++) Hist[i] = 0 ;
  

    /* calculate the histogram */
    for (k=n0; k<n1; k++) {
    	i=YY[k] ;
    	(Hist[i]) ++ ;
    }
    
    sum_len=0 ;
    for (i=0; i<nClasses; i++) sum_len += Hist[i] ;
    

    /* Calculate the entropy before this split 				*/
    /* Before split:   H(U) = n1/N ( sum(h_i(1)/n1 * log(h_i(1)/n1) ) 	*/
    entropyOld = 0.0 ;
    for (i=0; i<nClasses; i++) {
    	sum1 = (double) Hist[i] / (double) sum_len ;
    	if(sum1 > 0) entropyOld -= sum1 * (double) log(sum1) ;
    }

        
    if(vlevel>100) {
    	printf("##Hist: total %d\n", sum_len) ;
    	for(i=0; i<nClasses; i++) printf("%d  ",Hist[i]) ;
    	printf("entropy %lf\n", entropyOld) ;
    }
    
    free( (void *) Hist ) ;
    
    return(entropyOld) ;
}




void PrintHistPrior(double **Hist, double *Prior, int nClasses, int RangeF)
{
  	int i,j ;
  	
   	printf("nClasses %d   RangeF %d\n", nClasses, RangeF) ;
 	
  	printf("Hist:\n") ;
    	for (i=0; i<RangeF; i++) { 
      	    for (j=0; j<nClasses; j++)  printf(" %lf  ", Hist[i][j]) ;
      	    printf("\n") ;
      	}
      	
       	printf("\nPrior:\n") ;
     	
    	for (i=0; i<RangeF; i++) { 
      	    printf(" %lf  ", Prior[i]) ;
      	}
       	printf("\n\n") ;

}
