#include "CPT.h"
#include "CPT_global.h"

extern int vlevel ;

/* Given matrix X - known classifications, Y - desired class.  
   First transfer X into binarized form bX, where
   for any column c_i of X, there are nClasses-1 columns, 
   C(i), C(i)+1,..., C(i)+nClasses-2 in
   bX corresponding to c_i, when x(j,i)=k,  k=0,1,..,nClass-2
   bX(j,C(i)+k) = 1 ;  When x(j,i)=nClasses-1, 
   bX(j,C(i)+m) = 0,  for any m, m=0,1,...,nClasses-2
   
   Binarize Y to bY, where when Y(j) = k, k=0,1,...,nClasses-1
   bY(j,k) =1 ;
   
   Then calculate and return [bX    ]*[bX', 1(Nx1)],  
   			                 [1(1xN)]
   			     
   bX*bY for mean square estimate
   		    bY = A * bX + v ;
   	<=> 	bY = [A,v] * [bX    ]
   			     [1(1xN)]
*/
void FeatBinCor(int *Y, int **X, int nL, int nH, int nNeighb, int nClasses,
   double **corXX, double **corXY, int Show )
{
   int i1,j1,k, i2, ii,jj, Ind_r, Ind_c ;
   
   Ind_r = Ind_c = nNeighb*(nClasses-1) +1 ;
   
   for(i1=0; i1<Ind_r; i1++) {
   	for(j1=0; j1<Ind_c; j1++) corXX[i1][j1]=0.0 ;
   	for(j1=0; j1<nClasses; j1++) corXY[j1][i1]=0.0 ;
   }
   
   
   for(i1=0; i1<nNeighb; i1++)
   for(i2=0; i2<nNeighb; i2++) {
   
   	
	for(k=nL; k<nH; k++) {
	
		/* calculate bX * bX' */
		if((X[k][i1]!=(nClasses-1)) && (X[k][i2]!=(nClasses-1))) {
			ii = (nClasses-1)*i1 + X[k][i1] ;
			jj = (nClasses-1)*i2 + X[k][i2] ;
		
			if(ii>=jj) (corXX[jj][ii]) +=1 ;	/* symmetric */
		
			if(i1==i2) {	
				(corXX[ii][Ind_c-1]) +=1 ;	/* calculate bX * 1(Nx1) */

				jj = Y[k] ;	/* calculate bY * bX' */
				
				if( (jj<0) || (jj>=nClasses) ) {
					fprintf(stderr, "Error, class out of range, %d, FeatBinCor()\n",jj) ;
					exit(-1) ;
				}
				
				(corXY[jj][ii]) +=1 ;
			}
		}
		
		if((i1==0) && (i2==0)) {	/* be sure just to do once */
			jj = Y[k] ;	
			if( (jj<0) || (jj>=nClasses) ) {
				fprintf(stderr, "Error, class out of range, %d, FeatBinCor()\n",jj) ;
				exit(-1) ;
			}
			(corXY[jj][Ind_c-1]) +=1 ;	/* calculate bY * 1(Nx1) */
		}
   	}
   
   }
   
   /* correct the last term of corXX */
   corXX[Ind_r-1][Ind_c-1] = nH-nL ;
   
   /* corXX is symmetric */
   for(i1=0; i1<Ind_r; i1++) 
   for(j1=i1+1; j1<Ind_c; j1++) corXX[j1][i1]=corXX[i1][j1] ;
   
   if(Show) {
   	fprintf(stdout, "autocorrelation of binarized X:\n") ;
   	PrintMatrix(stdout,corXX,0,Ind_r-1,0,Ind_c-1) ;
   	fprintf(stdout, "correlation of binarized X and binarized Y:\n") ;
   	PrintMatrix(stdout,corXY,0,nClasses-1,0,Ind_c-1) ;
   }  
}



/* Given an observation X - known classifications.  
   Transfer X into binarized form bX, where
   for any column c_i of X, there are nClasses-1 columns, 
   C(i), C(i)+1,..., C(i)+nClasses-2 in
   bX corresponding to c_i, when x(j,i)=k,  k=0,1,..,nClass-2
   bX(j,C(i)+k) = 1 ;  When x(j,i)=nClasses-1, 
   bX(j,C(i)+m) = 0,  for any m, m=0,1,...,nClasses-2
   return([bX,1]) ;
*/
void FeatBinarized(int *X, int nNeighb, int nClasses, int *bX)
{
   int i,j, Ind_c ;
   
   Ind_c = nNeighb*(nClasses-1) +1 ;
   for(i=0; i<Ind_c; i++) bX[i] = 0 ;
   
   for(i=0; i<nNeighb; i++) {
   	if(X[i]<(nClasses-1)) {
   		j = (nClasses-1)*i + X[i] ;
   		bX[j] = 1 ;
   	}
   }
   bX[Ind_c-1] = 1 ;
}



