#include "CPT.h"
#include "CPT_global.h"

static int debug=0 ;

/*****************************************************************************/
/* Generate a not at parent  */                 
/*****************************************************************************/
static struct TreeNode* TreeGenNode(
  struct TreeNode* parentPt,
  int i0, int i1,
  struct Tree_Info *treeInfoPt,
  char T     /* generate left "L" child or right "R" child */
)
{
    struct TreeNode *newNode ;
    int i, RangeF, nClasses, nNeighb ;
    static nodeNumber=0 ;
    
    RangeF = treeInfoPt->RangeF ;
    nNeighb = treeInfoPt->dimen ;
    nClasses = treeInfoPt->nclasses ;
              
    if(vlevel >100) printf("n0=%d, n1=%d, Rf=%d\n", i0,i1,RangeF) ;
    nodeNumber ++ ;
    
    newNode=(struct TreeNode *) get_spc(1,sizeof(struct TreeNode));
    
    newNode->left=NULL;
    newNode->right=NULL;

    newNode->n0=i0 ;
    newNode->n1=i1 ;
    newNode->n2=-1 ;
    newNode->nodeNumber = nodeNumber ;

    newNode->costDrop=-1 ;
    newNode->Prior = (double *) get_spc(RangeF, sizeof(double)) ;
    newNode->Hist = (double **) get_mtrx(RangeF, nClasses, sizeof(double)) ;
    newNode->testCostLR = (double *) get_spc(RangeF, sizeof(double)) ; 
    newNode->MP_Class = (int *) get_spc(RangeF, sizeof(int)) ; 
    for(i=0; i<RangeF; i++) (newNode->MP_Class)[i]=-1 ;
    
    newNode->entropy = -1 ;
    newNode->nLeavesBelow = 0 ;

    newNode->T = (double**) get_mtrx(nNeighb, nClasses-1, sizeof(double)) ;     
    /* Linear partition function  T[nNeighb][nClasses-1] */
    /* double T0 ;     constant term of the linear partition function */    
    

    if(parentPt==NULL) {    /* The tree is empty */
        newNode->parent=NULL;
        newNode->level = 0 ;

    } else { /* the tree is not empty */
        newNode->parent=parentPt ;
        newNode->level = (parentPt->level) +1 ;
        
        if(((parentPt->left)==NULL) && (T=='L')) newNode->IsLeftChild=1 ;
        else if(((parentPt->right)==NULL) && (T=='R')) newNode->IsLeftChild=0 ;
        else {
            fprintf(stderr, "Parent has the child, Cannt generate the node.\n") ;
            /* fprintf(stderr, "L%x, R%x, %x, Ch=%c\n",parentPt->left,parentPt->right,parentPt,T) ;*/  /* line changed to the following by GF 3/25/02 */

 fprintf(stderr, "L%p, R%p, %p, Ch=%c\n",parentPt->left,parentPt->
right,parentPt,T);

            exit(-1) ;
        }
        
    }
    return(newNode) ;
}


    
    
/* Generate a node at parent  */                 
static void TreeFreeNode(struct TreeNode* node)
{        
    if(node->testCostLR != NULL) free((void*) (node->testCostLR)) ;
    if(node->MP_Class != NULL) free((void*) (node->MP_Class)) ;
    if(node->Prior != NULL) free((void*) (node->Prior)) ;
    if(node->Hist != NULL) free_mtrx((void**) node->Hist) ;
    if(node->T != NULL) free_mtrx((void**) node->T) ;
    if(node != NULL) free((void*) node) ;
}


    
    
/* Generate a node at parent  */                 
void FreeNodeBelow(struct TreeNode* node)
{
    struct TreeNode *child ;
                
    child = node->left ;
    if(child != NULL) FreeNodeBelow(child) ;
         
    child = node->right ;
    if(child != NULL) FreeNodeBelow(child) ;
         
    TreeFreeNode(node) ;         
}


    
    
void DecTreeInitial( struct TreeNode **root, int nL, int nH,
    struct Tree_Info *treeInfoPt ) 
/*  nL,nH ; */    /* lower and upper indices of samples for this node */
{    
    if(vlevel > 1) printf("\tInitializing decision tree ... %d, %d\n", nL, nH);
    
    (*root) = TreeGenNode(NULL,nL,nH,treeInfoPt,'R') ;
    if( (*root)==NULL ) {
        fprintf(stderr,"Fail to generate tree node\n") ;
        exit(-1) ;
    }
}




/****************************************************************************/
/****************************************************************************/
static int BuildingNode(
  struct TreeNode *node,
  int iChild,    /* the i-th component of the desired output */
  struct Tree_Info *treeInfoPt,
  struct Tree_Data *DataPt 
)
{ 
  if( ! LinearSplit(node, DataPt, treeInfoPt, iChild )){
      if(vlevel >1 && debug) fprintf(stderr, "ERROR: LinearSplit() failed.\n");
      return(-1) ;
  }
  if( ! Partition(node, iChild, treeInfoPt, DataPt)) {
      if( debug ) fprintf(stderr, "Cannt partition \n") ;
      return(0);  /* not success, return */
  }
  
  return(1) ;
}    
      
      




/****************************************************************************/
/****************************************************************************/
static void PrintNodePool(struct TreeNode *NodePool) 
{
  struct TreeNode *node ;    
  int i=0 ;
  
  fprintf(stdout, "Ouput NodePool:\n") ;
  node = NodePool ;
  for(;node!=NULL;) {
      i++ ;
      fprintf(stdout, "    %d,  costDrop=%lf, ",i,node->costDrop) ;
      fprintf(stdout, "ent=%lf  ",node->entropy) ;
      fprintf(stdout, "%lf, n0=%d, n1%d, n2=%d\n",
          node->T0,node->n0,node->n1,node->n2) ;
      
      node = node->right ;
  }
}
  


/****************************************************************************/
/****************************************************************************/
static void Put2Pool(    /* put the root into node pool */
  struct TreeNode **NodePool ,    /* the root of the NodePool */
  struct TreeNode *node     /* the current best node */
)
{
  struct TreeNode *current, *Parent ;    /* the current best node */
  double key ;    
  
  key = node->costDrop ;    /* cost or award */
  
  if((*NodePool)==NULL) {    /* empty pool */
      (*NodePool) = node ;
      node->right = NULL ;    /* end of the node pool */
      node->left = NULL ;    /* head of the node pool */
    return ;
  } 
  
  /* when NodePool is not empty */
  current = (*NodePool) ;
  for(; (current->costDrop) >= key; )  {/* the node is worse than current */
  
      if(current->right != NULL) {    /* not the tail */
          current = current->right ;
      } else {            /* the tail */
          current->right = node ;
          node->left = current ;
          node->right = NULL ;
          return ;
      }
  }
      
  Parent = current->left ;

  if(Parent != NULL) {
      node->left = Parent ;
      Parent->right = node ;
      node->right = current ;
      current->left = node ;
              
  } else {    /* this node should be the new head */
      node->right = (*NodePool) ;    /* end of the node pool */
      node->left = NULL ;    /* head of the node pool */
       (*NodePool) = node ;
  }
}





/****************************************************************************/
/****************************************************************************/
static struct TreeNode* GetFromPool(    /* Get a node from NodePool */
  struct TreeNode **NodePool ,    /* the root of the NodePool */
  int *Leaves 
)
{
  struct TreeNode *node ;    /* the current best node */
  struct TreeNode *parent ;
  
  if((*NodePool)==NULL) {
      return(NULL) ;    /* Pool is empty */
      
  } else {        /* Get the first from the Pool */
      node = (*NodePool) ;
      (*NodePool) = node->right ;
      if((*NodePool)!=NULL) (*NodePool)->left=NULL ;    /* the head has no father */

      node->left = NULL ; /* reset pointers for the one just got from the pool */
      node->right = NULL ;
      
      /* adjust the number of leaves and put the node into decTree */
      parent = node->parent ;
    
    if(parent==node) {
        fprintf(stderr, "Error:  Parent==Node\n") ;
        exit(-1) ;
    }
  
      if(parent!=NULL) {        /* tree is not empty */
          
          /* insert the node as a child of its parent */
          if((node->IsLeftChild) && ((parent->left)==NULL)) {
              parent->left = node ;
          } else if((!(node->IsLeftChild)) && ((parent->right)==NULL)) {
              parent->right = node ;
          } else {
              fprintf(stderr, "Something is wrong. Parent may have") ;
              fprintf(stderr, "a child already. IsLeft=%d,", (node->IsLeftChild)) ;
          /*    fprintf(stderr, " L%x, R%x\n",(parent->left),(parent->right)) ; */ /* changed to following GF 3/25/02 */

                fprintf(stderr, " L%p, R%p\n",(parent->left),(parent->right));

              exit(-1) ;
          }
      } 

      
      (*Leaves) ++ ;
      return(node) ;
  }
}



/****************************************************************************/
/****************************************************************************/
static void BestFirstGrow(    /* return number of leaves used */
  struct Trees *treePt, 
  struct TreeNode **nodePool,
  int iChild,     /* process i-th child, i-th component of desired output Y */
                  /* the i-th child, 00-0, 01-1, 10-2, 11-3 */
  int *nLeaves,
  double *TotalCost
)
{
  struct TreeNode *bestNode ;    /* the current best node */
  struct TreeNode *child ;
  struct Tree_Data *DataPt ;
  struct Tree_Info *infoPt ;
  int  Success,MaxLeaves ;
  int  n0,n1,n2 ;
  int  i=0 ;
    
  DataPt = &(treePt->data) ;
  infoPt = treePt->infoPt ;
  
  MaxLeaves = (treePt->infoPt)->MaxLeaves ;
  while((*nLeaves) < MaxLeaves) {
      
      i++; 
  
      bestNode = GetFromPool(nodePool, nLeaves) ;
      if(bestNode==NULL) {
          /* fprintf(stdout, "No more nodes available. Tree construction stops!\n"); */
          break ;
      }
      *TotalCost -= bestNode->costDrop ;
      
      if(vlevel>20) {
         printf("%d, cstDrp=%lf,  ", i, bestNode->costDrop) ;
         printf("TotalCost=%lf, nLeaves=%d\n", *TotalCost,(*nLeaves));
      }

      if(vlevel>100) {
          fprintf(stdout, "\nGot a node from Pool\n") ;
          PrintNodePool((*nodePool)) ;
      }
    
      
      
      /* Generate children of the bestNode and put them into NodePool */      
      /* initialization */
      n0 = bestNode->n0 ;
      n1 = bestNode->n1 ;
      n2 = bestNode->n2 ;
      if(vlevel>30) fprintf(stdout, "    n0=%d, n1=%d, n2=%d\n", n0,n1,n2) ;
  
      if(n2<0) {
          fprintf(stderr, "This node hasn't been splitted. n2=%d, ",n2) ;
          fprintf(stderr, "ConstructDecTree, DecTree_u.c\n") ;
          exit(-1) ;
      }
      
      
      if((n2-n0) >infoPt->MinSamples) {  /* enough data left to improve entropy */
      
          child = TreeGenNode(bestNode, n0,n2, infoPt, 'L');
          Success = BuildingNode( child, iChild, infoPt, DataPt );
          if(Success==1) Put2Pool(nodePool, child) ;
      }     
  
      if(vlevel>30) {
          fprintf(stdout, "\nPut left child to Pool\n") ;
          PrintNodePool((*nodePool)) ;
      }
 
      if((n1-n2) > infoPt->MinSamples) {
  
          child = TreeGenNode(bestNode,n2,n1, infoPt,'R');
          Success = BuildingNode(child, iChild, infoPt, DataPt ) ;
          if(Success==1) Put2Pool(nodePool, child) ;
     } 
     
     if(vlevel>30) {
          fprintf(stdout, "\nPut right child to Pool\n") ;
          PrintNodePool((*nodePool)) ;
     }
  }
}





/****************************************************************************/
/****************************************************************************/
static void FreeNodePool(struct TreeNode *NodePool) 
{
  struct TreeNode *child ;    
  
  if(NodePool!=NULL) {
      child=NodePool->right ;
      TreeFreeNode(NodePool) ;
      FreeNodePool(child) ;
  }
}




/****************************************************************************/
/****************************************************************************/

int ConstructDecTree(    /* return number of leaves used */
  struct Trees *treePt,
  int iChild     /* process i-th child, i-th component of desired output Y */
                  /* the i-th child, 00-0, 01-1, 10-2, 11-3 */
)
{
  struct TreeNode *root ;
  struct TreeNode *NodePool ;    /* the root of the NodePool */
  int  nLeaves ;        /* the number of leaves */
  int  Success ;
  struct Tree_Data *DataPt ;
  struct Tree_Info *infoPt ;
  double TotalCost=0;
  int  i,j, nClasses, RangeF ;
    
  DataPt = &(treePt->data) ;
  infoPt = treePt->infoPt ;
  nClasses = infoPt->nclasses ;
  RangeF = infoPt->RangeF ;
  
  NodePool = NULL ;    /* empty pool */
  nLeaves = 1 ;        /* root counts as one leave */
  root = (treePt->roots)[iChild] ;
  Success = BuildingNode(root, iChild, infoPt,DataPt) ;
  
  if(vlevel>100) {
      fprintf(stdout, "\nRoot\n") ;
      PrintNodePool(NodePool) ;
  }

  if(Success==1) {
      Put2Pool(&NodePool, root);  /* put the root into node pool */
  } else if(Success==0 || Success==-1) {
  
/* modified on May 5 */
      root->nodeNumber = -1 ;
      root->T0 = 0.0 ;
      for(i=0; i<infoPt->dimen; i++)
      for(j=0; j<nClasses-1; j++) (root->T)[i][j]=0.0 ;
      
      for(i=0; i<RangeF; i++) (root->Prior)[i] = 1.0 / (double) RangeF ;
      
      for(i=0; i<RangeF; i++) 
      for(j=0; j<nClasses; j++) (root->Hist)[i][j] = 1.0 / (double) nClasses ;
      
      root->left=NULL;
      root->right=NULL;

      for(i=0; i<RangeF; i++) (root->MP_Class)[i]=0 ;
    
      return(-1) ;

/*    fprintf(stderr, "BuildingNode(root) failed. DecTree_u.c. \n") ;
      fprintf(stderr, "No tree generated. Please increase the \n") ;
      fprintf(stderr, "percentage of samples used\n") ;
      exit(-1) ;
*/
  }
  
  if(vlevel>100) {
      fprintf(stdout, "\nRoot\n") ;
      PrintNodePool(NodePool) ;
  }
  
  BestFirstGrow(treePt, &NodePool, iChild, &nLeaves, &TotalCost)  ;

  FreeNodePool(NodePool) ;
  
  return(nLeaves) ;
}





/***************************************************************************/
/* regrow the constructed tree on another sample set and also */
/* initialize the NodePool */
/***************************************************************************/

static void ApplyTreeMkPool( 
  struct TreeNode *node,
  int iChild,     /* process i-th child, i-th component of desired output Y */
                  /* the i-th child, 00-0, 01-1, 10-2, 11-3 */
  struct Tree_Info *treeInfoPt, 
  struct Tree_Data *DataPt, 
  struct TreeNode **nodePool,    /* the root of the NodePool */
  int *nLeaves
)
{
  struct TreeNode *left, *right ;
  int n0,n1,n2, Success, MinS ;
  
  MinS = treeInfoPt->MinSamples ;
  if(node==NULL) NullPt("node, ApplyTreeMkPool") ;
  
  /* apply current node only, partition [n0, n1)*/
  testPartitionOnly(node, iChild, treeInfoPt, DataPt) ;  
  
  left=node->left;  right=node->right;
  n0=node->n0;      n1=node->n1;        n2=node->n2;
  /*if(vlevel>100) printf("n0=%d, n1=%d, n2=%d, l%x, r%x\n", n0,n1,n2,left,right);
*/ /* changed by GF 3/25/02 */
if(vlevel>100) printf("n0=%d, n1=%d, n2=%d, l%p, r%p\n", n0,n1,n2,left,right
);
  
  if(left!=NULL) {    
      left->n0=n0;  left->n1=n2;  left->n2=-1;
      if(((n2-n0)>0)&&(n2>=0)) ApplyTreeMkPool(left, iChild, treeInfoPt, DataPt,
        nodePool, nLeaves) ;
    
  } else {     /* expending leaves */
  
      if((n2-n0) > MinS) {    /* enough data left to improve entropy */
        left = TreeGenNode(node, n0,n2, treeInfoPt, 'L') ;
        Success=BuildingNode(left, iChild, treeInfoPt, DataPt);
        if(Success==1) Put2Pool(nodePool, left) ;
      }     
      (*nLeaves) ++ ;
  }

  if(right!=NULL) {    
      right->n0=n2;  right->n1=n1;  right->n2=-1;
      if(((n1-n2)>0)&&(n2>=0)) ApplyTreeMkPool(right, iChild, treeInfoPt,DataPt,
        nodePool, nLeaves) ;
    
  } else {     /* expending leaves */
  
      if((n1-n2) > MinS) {    /* enough data left to improve entropy */
        right = TreeGenNode(node,n2,n1,treeInfoPt,'R') ;
        Success=BuildingNode(right, iChild, treeInfoPt, DataPt);
        if(Success==1) Put2Pool(nodePool, right) ;
      } 
      (*nLeaves) ++ ;
  }
  
  if(vlevel>100) {
      fprintf(stdout, "\nPut children to Pool\n") ;
      PrintNodePool((*nodePool)) ;
  }

}

  


/***************************************************************************/
/* reset all the cost before grow a new tree */
/***************************************************************************/
static void resetCost(struct TreeNode *node, int *nLeaves) 
{
  struct TreeNode *child ;

  if(node!=NULL) {
      node->n0 = -1 ;
      node->n1 = -1 ;
      node->n2 = -1 ;

      child=node->left ;
      resetCost(child, nLeaves) ;

      child=node->right ;
      resetCost(child, nLeaves) ;  
        
  } else (*nLeaves)++ ;
}





/***************************************************************************/
/***************************************************************************/

int growTree(struct Trees *treePt, int nN0, int nN1, int iChild) 
{
  struct TreeNode *root ;
  struct TreeNode *NodePool ;    /* the root of the NodePool */
  struct Tree_Data *DataPt ;
  struct Tree_Info *infoPt ;
  int nLeaves1=0, nLeaves2=0 ;
  double TotalCost=0 ;
  
  root = (treePt->roots)[iChild] ;
  if(root==NULL) NullPt("root, growTree") ;
  
  DataPt = &(treePt->data) ;
  infoPt = treePt->infoPt ;

  if(vlevel>0) printf("\n\tTree growing ...\n") ;
  resetCost(root, &nLeaves1) ;
  if(vlevel>0) printf("\tOld tree has %d leaves\n", nLeaves1) ;
  

  NodePool = NULL ;
  root->n0=nN0;  root->n1=nN1;  root->n2=-1 ;
  if(vlevel>10) printf("\tGrowTree starts with N0=%d, N1=%d\n", nN0, nN1) ;
  
  /* apply old tree and expend leaves */
  ApplyTreeMkPool(root, iChild, infoPt, DataPt, &NodePool, &nLeaves2) ; 
  if(vlevel>0) printf("\tNew data reaches  %d leaves\n", nLeaves2) ;
  
  /* grow a new tree, starting from leaves of the old tree */
  BestFirstGrow(treePt, &NodePool, iChild, &nLeaves1, &TotalCost)  ;
  if(vlevel>0) printf("\tNew tree has %d leaves\n", nLeaves1) ;

  FreeNodePool(NodePool) ;
  
  return(nLeaves1) ;
}



