/* calculate the linear mean square fit of the data 
   and form the linear split function by calculating 
   the principle direction of the l.m.s predicts
*/
#include "CPT.h"
#include "CPT_global.h"

static int debug=0 ;


int LinearSplit( struct TreeNode *node, struct Tree_Data *DataPt, 
    struct Tree_Info *treeInfoPt, int iChild )
{
   int i,j,k, Ind_r, Ind_c ;
   int IndMax, nL, nH, Len ;
   double EigMax, sum ;
   int **X, *YY, nClasses, nNeighb;
   static double **XX, **XY, **pinvXX, **A, **Cov, *pY, *MpY, *Eig, *T ;
   static int *bX, AllocateFlag=1 ;

   X = DataPt->X ;
   YY = (DataPt->Y)[iChild] ;
   nNeighb = treeInfoPt->dimen ;
   nClasses = treeInfoPt->nclasses ;
   Ind_r = Ind_c = nNeighb*(nClasses-1) +1 ;
   nL=node->n0;	nH=node->n1 ;
   Len = nH-nL ;
   
   if(Len<2) return(0) ;

   if(AllocateFlag) {
   	AllocateFlag = 0 ;
   	bX = (int*) get_spc(Ind_r,sizeof(int)) ;
   	XX = (double**) get_mtrx(Ind_r,Ind_c,sizeof(double)) ;
   	XY = (double**) get_mtrx(nClasses,Ind_c,sizeof(double)) ;
   	pY = (double*) get_spc(nClasses,sizeof(double)) ;
   	MpY = (double*) get_spc(nClasses,sizeof(double)) ;
   	Eig = (double*) get_spc(nClasses,sizeof(double)) ;
   	Cov = (double**) get_mtrx(nClasses,nClasses, sizeof(double)) ;
   	pinvXX = (double**) get_mtrx(Ind_r,Ind_c,sizeof(double)) ;	
   	T = (double*) get_spc(Ind_c,sizeof(double)) ;	
   	A = (double**) get_mtrx(nClasses,Ind_c,sizeof(double)) ; 	/* Y=A[X] 
   								              [1] */
   }
   
      
   FeatBinCor(YY,X,nL,nH,nNeighb,nClasses,XX,XY,0) ;
   
   /* pseudoinverse. if fails, return(0) */
   if(!(pinv(XX, Ind_r, Ind_c, pinvXX, 1))) return(0) ;	
   
   MatrixMultiply(XY,0,nClasses-1,0,Ind_c-1,0,pinvXX,0,Ind_r-1,0,Ind_c-1,0,A) ;

   if( debug ) {
	   printf("XX\n");  PrintMatrix(stdout, XX,0,Ind_r-1,0,Ind_c-1) ;
	   printf("XY\n");  PrintMatrix(stdout, XY,0,nClasses-1,0,Ind_c-1) ;
	   printf("pinvXX\n");  PrintMatrix(stdout, pinvXX,0,Ind_r-1,0,Ind_c-1) ;
	   printf("A\n");  PrintMatrix(stdout,A,0,nClasses-1,0,Ind_c-1) ;
   }
      
   
   /* calculate the mean, covariance matrix of results of the l.m.s estimation */
   for(i=0; i<nClasses; i++) {
   	   MpY[i] = 0.0 ;
   	   for(j=0; j<nClasses; j++) Cov[i][j]=0 ;
   }
   
   for(k=nL; k<nH; k++) {
   	   FeatBinarized(X[k],nNeighb,nClasses,bX) ;
   	
           if( debug ) {
   		   for(i=0; i<Ind_c; i++) printf("%d  ",bX[i]) ;
   		   printf("\n") ;
   	   }
   	
   	   for(i=0; i<nClasses; i++) pY[i]=0.0 ;
   	   for(i=0; i<Ind_c; i++) {
   		   if(bX[i]==1) {
   			   for(j=0; j<nClasses; j++) (pY[j]) += A[j][i] ;
   		   }
   	   }
   	
   	
  	   for(i=0; i<nClasses; i++) {

           if( debug )  printf("%lf  ", pY[i]) ;
  		   MpY[i] += pY[i] ;
  		   for(j=0; j<nClasses; j++) {
  			   Cov[i][j] += pY[i] * pY[j] ;
  		   }
  	   }
       if( debug )  printf("\n") ;
   }
   
   for(i=0; i<nClasses; i++) 
   for(j=0; j<nClasses; j++) {
       Cov[i][j] = (Cov[i][j] - MpY[i] * MpY[j]/Len)/ (Len-1) ;
   }
   
   for(i=0; i<nClasses; i++) MpY[i] /= Len ;

   if( debug ) {
   	fprintf(stdout,"Covariance:\n");  
   	PrintMatrix(stdout, Cov,0,nClasses-1,0,nClasses-1) ;
   	fprintf(stdout,"Mean:\n");  PrintVector(stdout, MpY,0,nClasses-1) ;
   }
   
   if(!(eigen(Cov,Eig,nClasses))) {	/* if fails */ 

     return(0) ;
   }


   
   if( debug ) {
   	printf("Eigen vectors and eigen values \n");
   	PrintMatrix(stdout, Cov,0,nClasses-1,0,nClasses-1) ;
   	PrintVector(stdout, Eig,0,nClasses-1) ;
   }
   
   /* find the principle component */
   IndMax=0;  EigMax=fabs(Eig[0]) ;
   for(i=1; i<nClasses; i++) {
   	if(fabs(Eig[i]) > EigMax) {
   		EigMax = fabs(Eig[i]) ;
   		IndMax = i ;
   	}
   }
   
   for(i=0; i<Ind_c; i++) {
   	sum = 0.0 ;
   	for(j=0; j<nClasses; j++) sum += Cov[j][IndMax]*A[j][i] ;
   	T[i] = sum ;
   }
   
   sum = 0.0 ;
   for(j=0; j<nClasses; j++) sum += Cov[j][IndMax]*MpY[j] ;
   T[Ind_c-1] -= sum ;
   
   if( debug ) {
   	printf("Final T\n") ;
   	PrintVector(stdout, T, 0, Ind_c-1) ;
   }
   
   
   /* save parameters */
   Ind_r = Ind_c = nNeighb*(nClasses-1) +1 ;
   for(i=0; i<nNeighb; i++) 
   for(j=0; j<nClasses-1; j++) {
   	k=i*(nClasses-1) + j ;
   	node->T[i][j] = T[k] ;
   }
   
   node->T0 = T[Ind_c-1] ;    
   return(1) ; 
}  

   
   
   
   
   
