/* partition the sample set using the linear partition function
   calculated from LinearSplit()
*/
#include "CPT.h"
#include "CPT_global.h"

static int debug = 0 ;

/* evaluate the effect of this node according to entropy drop */
static void Cost(struct TreeNode *node, int nSamplesTotal) 
{
  int n0,n1;
  
  n0= node->n0 ;  
  n1= node->n1 ;  
  
  node->costDrop = node->costDrop * (double) (n1-n0) / (double) nSamplesTotal;
}



static void Classifier(struct TreeNode *node, int nClasses, int RangeF)
{
    int i,j, maxIndex ;
    double **Hist, maxH ;
    static First=1 ;


    Hist = node->Hist ;

    /* reset Hist and Prior */
    for (i=0; i< RangeF; i++) {
        maxH = Hist[i][0] ;
        maxIndex = 0 ;
        for(j=1; j<nClasses; j++) {
                if(maxH < Hist[i][j]) {
                        maxH = Hist[i][j] ;
                        maxIndex = j ;
                }  else if((maxH==Hist[i][j]) && (First)) {
                        /* printf("There is a tie \n") ; */
                        First = 0 ;
                }
        }
        node->MP_Class[i] = maxIndex ;

    }
}

 
int localPartition(int *XX, struct TreeNode *node, int nNeighb, int nClasses)
{
   int i ;
   double sum ;
   
   sum = node->T0 ;
   for(i=0; i<nNeighb; i++) {
       if( (XX[i]<0) || (XX[i]>nClasses-1)) {
           fprintf(stderr, "Class out of range nei=%d, cla=%d, nclasses=%d\n",
               i,XX[i],nClasses ) ;
           exit(-10) ;
       }
       
       if(XX[i]<(nClasses-1)) {
           sum += (node->T)[i][(XX[i])] ;
       }
   }
   
   if(sum>=0) return(1) ;
   else return(0) ;
}




int Partition(struct TreeNode *node, int iChild, struct Tree_Info *treeInfoPt,
   struct Tree_Data *DataPt)
{
   int i,j,k ;
   int nL, nH, y,x ;
   int Show=0, GoOn ;
   int nTree, nNeighb, nClasses, RangeF ;
   int **X, **Y, nLen, currentSS ;
   
   nTree = treeInfoPt->nTree ;
   nNeighb = treeInfoPt->dimen ;
   RangeF = treeInfoPt->RangeF ;
   nClasses = treeInfoPt->nclasses ;
   
   X = DataPt->X ;
   Y = DataPt->Y ;
   currentSS = DataPt->CurrentSS ;
   if( currentSS != 0 && currentSS != 1) {
       fprintf(stderr, "ERROR:  invalid currentSS %d, must be 0 or 1\n", 
           currentSS) ;
       exit(-1) ;
   }
   nLen = (DataPt->nlen)[currentSS] ;
       
   nL=node->n0;    nH=node->n1 ;
   if(vlevel>100) printf("nL%d, nH%d\n", nL,nH) ;    
        
   if(vlevel>1 && debug ) {
       for(i=nL; i<nH; i++) {
           x=localPartition(X[i], node, nNeighb, nClasses) ;
           printf("%d,  %d:  ",i,x) ;
           for(j=0; j<nNeighb; j++) printf("%d  ", X[i][j]) ;
           printf("\n") ;
       }
   }

   /* partition Y,X according to the linear partition function, 
      put 1 or 0 together */
   i = nL ;
   j = nH-1 ;
   while((i<j)&&(i<nH)&&(j>=nL)) {
       
       /* first part should be 1 */
       GoOn = 1 ;
       while( (i<nH)&& GoOn) {
          GoOn = (localPartition(X[i], node, nNeighb, nClasses)==1) ;
          if(GoOn) i++ ; 
       }
       
       GoOn = 1 ;
       while((j>=nL)&&GoOn) {
          GoOn = (localPartition(X[j], node, nNeighb, nClasses)==0) ;
         if(GoOn) j-- ; 
       }
      
       if((i<j)&&(i<nH)&&(j>=nL)) {    /* switch values betwee i,j */
          for(k=0; k<nTree; k++) {
              y=Y[k][j]; Y[k][j]=Y[k][i]; Y[k][i]=y;
          }
          
          for(k=0; k<nNeighb; k++) {
              x=X[j][k]; X[j][k]=X[i][k]; X[i][k]=x;
          }
          
          i++; j--;
       }
  }
  
  
  /* double check if the above shuffling is right and also get node->n2 */
  k=0 ;
  for(i=nL; i<nH; i++) {
  
      x=localPartition(X[i], node, nNeighb, nClasses) ;
      if(vlevel>100) printf("%d,  %d\n",i,x) ;
    
      if(k==0) {
          if(x==0) {
              node->n2=i;    /* ending of 1s, starting of 0s in BestF */
              k=1 ;   /* there is a switch from 1 to 0 */
          }
      } else {
          if(x==1) {
              fprintf(stderr,"Error in shuffling BestF!\n") ;
              exit(-1) ;
          }
      }
  }
  if(k==0) node->n2=nH ;    /* all samples belong to '1' class */
  
  
  /* evaluate the effect of this node,  Reissance criterion */
  /* cost difference due to the Reissance criterion of order*/ 
  /* indentification 1/2*K*log2(N)/N */

  if(vlevel>1 && debug) Show=1 ;
  else Show=0 ;

  entropyEntpDrop(Y[iChild],node,nClasses,RangeF,Show) ;
  Classifier(node, nClasses, RangeF) ;   /*  entropy.c */
  Cost(node, nLen) ;    /* calculate the cost for this node */
      
  if(((node->n2)<=nL)||((node->n2)>=nH)||(nH<=nL)) {

    if(vlevel>1 && debug) printf("    No more infor, n0=%d, n1=%d, n2=%d\n",
          nL,nH,(node->n2));
      return(0) ;    /* no more information */
      
  } else {    
      
      if((node->costDrop) < MIN_COST_CHANGE) return(0) ;
      else return(1) ;    /* successful */
  }
}  

   
   
/* Only partition the node,  do not evaluate how good it is */  
void testPartitionOnly(struct TreeNode *node, int iChild,
    struct Tree_Info *treeInfoPt, struct Tree_Data *DataPt)
{
   int i,j,k ;
   int nClasses, nNeighb, nTree ;
   int nL, nH, y,x, **Y, **X ;
   int GoOn ;
   
   
   X = DataPt->X ;
   Y = DataPt->Y ;
   nClasses = treeInfoPt->nclasses ;
   nNeighb = treeInfoPt->dimen ;
   nTree = treeInfoPt->nTree ;
   
   nL=node->n0;    nH=node->n1 ;
   if(vlevel>100) printf("nL%d, nH%d\n", nL,nH) ;    
        
   if(vlevel>100)  {
       for(i=nL; i<nH; i++) {
          x=localPartition(X[i],node, nNeighb, nClasses) ;
          printf("%d,  %d\n",i,x) ;
       }
   }
   
   /* partition Y,X according to the linear partition function, 
      put 1 or 0 together */
   i = nL ;
   j = nH-1 ;
   while((i<j)&&(i<nH)&&(j>=nL)) {
       
       /* first part should be 1 */
       GoOn = 1 ;
       while( (i<nH)&& GoOn ) {
          GoOn = (localPartition(X[i],node, nNeighb, nClasses)==1) ;
          if(GoOn) i++ ; 
       }
       
       GoOn = 1 ;
       while((j>=nL)&&GoOn) {
           GoOn = (localPartition(X[j],node,nNeighb, nClasses)==0) ;
           if(GoOn) j-- ; 
       }
      
       if((i<j)&&(i<nH)&&(j>=nL)) {    /* switch values betwee i,j */
          for(k=0; k<nTree; k++) {
              y=Y[k][j]; Y[k][j]=Y[k][i]; Y[k][i]=y;
          }
          
          for(k=0; k<nNeighb; k++) {
              x=X[j][k]; X[j][k]=X[i][k]; X[i][k]=x;
          }
          
          i++; j--;
       }
  }
  
  
  /* double check if the above shuffling is right and also get node->n2 */
  k=0 ;
  for(i=nL; i<nH; i++) {
  
      x=localPartition(X[i],node,nNeighb, nClasses) ;
      if(vlevel>100) printf("%d,  %d\n",i,x) ;
    
      if(k==0) {
          if(x==0) {
              node->n2=i;    /* ending of 1s, starting of 0s in BestF */
              k=1 ;          /* there is a switch from 1 to 0 */
          }
      } else {
          if(x==1) {
              fprintf(stderr,"Error in shuffling BestF!\n") ;
              exit(-1) ;
          }
      }
  }
  if(k==0) node->n2=nH ;    /* all samples belong to '1' class */
}  

   
   
   
  
