#include "CPT.h"
#include "CPT_global.h"

static int debug=0 ;


/*****************************************************************************/
/*****************************************************************************/
static void showClassify1Cost(FILE *fd, struct TreeNode *x, int depth)
{
  int i;
	
  if (x!= NULL)  {		/* recursively print tree in DESENDING ORDER */
	showClassify1Cost(fd, x->right,depth+1);
	for(i=0;i<depth*4;i++) fprintf(fd, " ");
			/* PRINT BLANK SPACES, so that the printed order
					 creates a virual tree shape, 90 degrees 
					 rotated with parent as a pivot */
	fprintf(fd, "%d, ",x->nodeNumber);
	fprintf(fd, "%6.0lf   ",x->testCostBelow);
	fprintf(fd, "(node L%6.0lf, R%6.0lf)\n",x->testCostLR[1],x->testCostLR[0]);
	showClassify1Cost(fd, x->left,depth+1);
  }
}



/*****************************************************************************/
/* Use Decision Tree to find out the classification of the sample */
/*****************************************************************************/

void Classify1_TestSample(int YY, int *XX, struct TreeNode *TrNode,
  int nNeighb, int nClasses)
/*  int *XX, YY -- known classification of coarse neighbors, desired class */
{
  int True ;
  struct TreeNode *child ;	/* TreeNode -- decision tree structure */
  
  
  if(TrNode != NULL) {
  	True = localPartition(XX, TrNode, nNeighb, nClasses) ;
  	
  	if(True) {	/* >=0 */
  		if(vlevel>100) printf("L(X)>=0 ") ;
  		child = TrNode->left ;
  		if((TrNode->MP_Class)[1] != YY)  (TrNode->testCostLR)[1] ++ ;

  	} else {
  		if(vlevel>100) printf("==False ") ;
  		child = TrNode->right ;
  		if((TrNode->MP_Class)[0] != YY)  (TrNode->testCostLR)[0] ++ ;
 	
  	}
  		
  	if(child != NULL) {
  		Classify1_TestSample(YY, XX, child, nNeighb, nClasses) ;
  	}
  
  } else {
  	fprintf(stderr,"Null Decision Tree pointer! ContexProb\n") ;
  	exit(-1) ;
  }
}
  



/*****************************************************************************/
/*****************************************************************************/
static void ZeroSetTestCost(struct TreeNode *node, int RangeF)   
{
  int i ;
  struct TreeNode *child ;	/* TreeNode -- decision tree structure */
  
  if(node != NULL) {
  	
  	node->testCost = 0 ;
  	node->testCostBelow = 0 ;
  	for(i=0; i<RangeF; i++) (node->testCostLR)[i]=0 ;
  	
  	child = node->left ;
  	if(child!=NULL) ZeroSetTestCost(child, RangeF) ;
  	
  	child = node->right ;
   	if(child!=NULL) ZeroSetTestCost(child, RangeF) ;
 	
  } else {
  	fprintf(stderr,"Null Decision Tree pointer! ZeroSetTestCost\n") ;
  	exit(-1) ;
  }
}
	





/*****************************************************************************/
/*****************************************************************************/
static void TreeTest(struct Trees *treePt, int iChild, int KK) 
{
  struct Tree_Data *DataPt ;
  struct Tree_Info *infoPt ;
  struct TreeNode *root ;
  int i, nNeighb, nClasses, RangeF ;
  int Method ;
  
  Method = 1 ;
  
  root = (treePt->roots)[iChild] ;
  DataPt = &(treePt->data) ;
  infoPt = treePt->infoPt ;
  nNeighb = infoPt->dimen ;
  RangeF = infoPt->RangeF ;
  nClasses = infoPt->nclasses ;
  
  ZeroSetTestCost(root, RangeF) ;   
  for(i=(DataPt->nL)[KK]; i<(DataPt->nH)[KK]; i++)  {
  	switch(Method) {
  	  case 1:
  	  	Classify1_TestSample( (DataPt->Y)[iChild][i], (DataPt->X)[i], root,
  	  	    nNeighb, nClasses ) ;
  	  	break ;
  	  default:
  	  	fprintf(stderr, "Invalid method, TreeTest\n") ;
  	  	exit(-1) ;
  	}
  }
  
  if(vlevel>1 && debug) showClassify1Cost(stdout, root, 0) ;
}  



/*****************************************************************************/
/*****************************************************************************/

static void NodePruning(struct TreeNode *x, int depth, char LR)
{
  float costBelow, costThisNode ;
  int i, index_left, index_right ;
  struct TreeNode *child ;
	
  
  index_left = 1;	/* left means true==1 */
  index_right = 0;

  switch (LR) {
    case 'L':	/* left child */
    	costThisNode = x->testCostLR[index_left] ;
    	child = x->left ;
    	if(child != NULL) costBelow = child->testCostBelow ;
    	else NullPt("NodePruning, L") ;
  
	   if(vlevel>100) {
  		   for(i=0;i<depth*4;i++) printf(" ");
  		   printf("%d, L, below %f,  (node %f) \n",x->nodeNumber,  
  			   costBelow, costThisNode) ;
	   }

	   if(costThisNode <= costBelow) {	/* get rid of branches below */

  		   if(vlevel>100) {
  			   for(i=0;i<depth*4;i++) printf(" ");
  			   printf("%d, L, Prune nLbelow %d \n", x->nodeNumber, 
  				   child->nLeavesBelow) ;
  		   }

  		   FreeNodeBelow((x->left)) ;
  		   x->left = NULL ;
	   }
  	break ;
    

    case 'R' :	/* right child */
    	costThisNode = x->testCostLR[index_right] ;
    	child = x->right ;
    	if(child != NULL) costBelow = child->testCostBelow ;
    	else NullPt("NodePruning, R") ;
  
  		if(vlevel>100) {
  			for(i=0;i<depth*4;i++) printf(" ");
  			printf("%d, R, below %f,  (node %f) \n",x->nodeNumber,  
  		    	costBelow, costThisNode) ;
  		}

  		if(costThisNode <= costBelow) {	/* get rid of branches below */

  			if(vlevel>1 && debug) {
  				for(i=0;i<depth*4;i++) printf(" ");
  				printf("%d, R, Prune nLbelow %d \n", x->nodeNumber, 
  			    	child->nLeavesBelow) ;
  			}

  			FreeNodeBelow((x->right)) ;
  			x->right = NULL ;
  		}
  	break ;
  	
    default:
    	fprintf(stderr, "Unknown func parameter %c\n", LR) ;
    	exit(-10) ;
  }
}




/*****************************************************************************/
/*****************************************************************************/
void TreePruning(FILE *fd, struct TreeNode *x, int depth)
{
  struct TreeNode *child ;
  int index_left, index_right ;
	
  
  index_left = 1;	/* left means true==1 */
  index_right = 0;
	
  if ((depth==0) && (vlevel>0)) printf("\n\tTree pruning ...\n") ;
  
  if (x!= NULL)  {		
	x->testCostBelow = 0.0 ;
	x->nLeavesBelow = 0 ;
	  	
  	child = x->right ;
  	if(child != NULL)  {
  		TreePruning(fd, child, depth+1) ;
  		NodePruning(x,depth, 'R') ;
  	}
  	
  	child = x->right ;	/* x->right may be reset after NodePruning() */
  	if(child != NULL)  {
  		(x->nLeavesBelow) += child->nLeavesBelow ;
  		(x->testCostBelow) += child->testCostBelow ;
  	
  	} else {
  		(x->nLeavesBelow) ++ ;
  	 	(x->testCostBelow) += x->testCostLR[index_right] ;
  	}
  	


  	child = x->left ;
  	if(child != NULL)  {
  		TreePruning(fd, child, depth+1) ;
  		NodePruning(x,depth, 'L') ;
  	}
  	
  	child = x->left ;
  	if(child != NULL)  {
  		(x->nLeavesBelow) += child->nLeavesBelow ;
  		(x->testCostBelow) += child->testCostBelow ;
  	
  	} else {
  		(x->nLeavesBelow) ++ ;
  	 	(x->testCostBelow) += x->testCostLR[index_left] ;
  	}
  	
  	x->testCost =  x->testCostLR[index_right] +  x->testCostLR[index_left] ;
  }
}
			



/*****************************************************************************/
/* tree-pruning algorithm proposed by Saul Gelfand,  1993  */
/*****************************************************************************/
void construct_CPT(struct Trees *treePt) 
{
  struct Tree_Data *DataPt ;
  struct Tree_Info *InfoPt ;
  struct TreeNode **roots ;
  int i,j,k, nLeaves, nLeavesPre ;
  int *nL, *nH, *nlen, nTree ;
  int Dimen, nSamples ;
  int **X, **Y, tmp ;

  
  DataPt = &(treePt->data) ;
  InfoPt = treePt->infoPt ;
  
  /* partition samples into two sets */
  i = DataPt->nSamples ;
  (DataPt->nL)[0]=0;    (DataPt->nL)[1]=i/2 ;
  (DataPt->nH)[0]=i/2;  (DataPt->nH)[1]=i ;
  (DataPt->nlen)[0] = (DataPt->nH)[0] - (DataPt->nL)[0] ;
  (DataPt->nlen)[1] = (DataPt->nH)[1] - (DataPt->nL)[1] ;


  /* since samples are saved sequentially from sample image */
  /* to sample image, we need to shuffle the samples, in order */
  /* to get better covergency and better estimation  */
/************* not used
  Dimen = InfoPt->dimen ;
  nSamples = DataPt->nSamples ;
  X = DataPt->X ;
  Y = DataPt->Y ;
  for(j=0; j< nSamples/2; j+=2 ) {
      for(k=0; k<Dimen; k++) {
          tmp = X[j][k] ;
          X[j][k] = X[nSamples-1-j][k] ;
          X[nSamples-1-j][k] = tmp ;
      }
      for(k=0; k<nTree; k++) {
          tmp = Y[k][j] ;
          Y[k][j] = Y[k][nSamples-1-j] ;
          Y[k][nSamples-1-j] = tmp ;
      }
  }
************ donot use this */          
      
  
  roots = treePt->roots ;
  nL = DataPt->nL ;
  nH = DataPt->nH ;
  nlen = DataPt->nlen ;

  nTree = InfoPt->nTree ;
  for(i=0; i<nTree; i++)   {

      nLeavesPre = -1 ;
      nLeaves = 0 ;
      for(j=0;  nLeavesPre<nLeaves && j<10; j++) {
          printf("\tTree %d,  iteration %d\n ",i, j) ;

          k=j%2 ;
          DataPt->CurrentSS = k ;
          nLeavesPre = nLeaves ;

          if(j==0) {
              DecTreeInitial(&((treePt->roots)[i]), nL[k],nH[k], InfoPt);
              nLeaves = ConstructDecTree(treePt, i) ;
              if(nLeaves<0) break ;    /* no tree generated */
              printf("\tThere are %d leaves\n", nLeaves) ;
          } else {
              /* apply old tree and expend leaves */
              nLeaves=growTree(treePt, nL[k], nH[k], i) ;
              printf("\tThere are %d leaves\n", nLeaves) ;
          }
          /* SaveTree(roots[i], MaxLeaves, fn, i) ; */
          if(vlevel>50) ShowCurrentResult( roots[i], InfoPt, DataPt ) ;


          k=(j+1)%2 ;  /* change to test set */
          DataPt->CurrentSS = k ;
          TreeTest(treePt, i, k) ;
          if(vlevel>50) {
              printf("\nOn Test set \n") ;
              showClassify1Cost(stdout, roots[i], 0) ;
          }

          TreePruning(stdout, roots[i], 0) ;
          printf("\tAfter pruning %d leaves\n", (roots[i])->nLeavesBelow);
          nLeaves = (roots[i])->nLeavesBelow ;

          if(vlevel>50) {
               printf("On Test set \n") ;
               showClassify1Cost(stdout, roots[i], 0) ;
          }

          printf("\tnLeavesPre %d, nLeaves %d\n", nLeavesPre, nLeaves) ;
          if(vlevel>50) ShowCurrentResult( roots[i], InfoPt, DataPt ) ;
      }
  }
}

		



