#include "CPT.h"
#include "CPT_global.h"

    /* procedure definitions  of the basic  bin. search tree operations */

    
    /* Various tree operations */    
void showTreeEntropy(FILE *fd, struct TreeNode *x, int depth)
{
  int i;
    
  if (x!= NULL)  {        /* recursively print tree in DESENDING ORDER */
    showTreeEntropy(fd, x->right,depth+1);
    for(i=0;i<depth*4;i++) fprintf(fd, " ");
            /* PRINT BLANK SPACES, so that the printed order
                     creates a virual tree shape, 90 degrees 
                     rotated with parent as a pivot */
    fprintf(fd, "%lf   ",x->entropy);
    fprintf(fd, "(%lf)\n",x->costDrop);
    showTreeEntropy(fd, x->left,depth+1);
  }
}
            


void showTreeSampleSize(FILE *fd, struct TreeNode *x, int nSamplesTotal, 
    int depth)
{
  int i, n0,n1,n2 ;
  double percent ;
    
  if (x!= NULL)  {        /* recursively print tree in DESENDING ORDER */
    showTreeSampleSize(fd, x->right, nSamplesTotal, depth+1);
    for(i=0;i<depth*4;i++) fprintf(fd," ");
            /* PRINT BLANK SPACES, so that the printed order
            creates a virual tree shape, 90 degrees 
            rotated with parent as a pivot */
    n0 = x->n0 ;
    n1 = x->n1 ;
    n2 = x->n2 ;
    i = n1-n0 ;
    percent = (i+.0) / nSamplesTotal ;
    fprintf(fd,"%d -- %d  (%5.3f) [%d, %d, %d)\n",x->nodeNumber,i,percent,n0,n2,n1);
    showTreeSampleSize(fd, x->left, nSamplesTotal, depth+1);
  }
}
            


void showTreeHistogram(FILE *fd, struct TreeNode *x, int nClasses, int depth)
{
  int i,j,k ;
  double **Hist ;
  
    
  if (x!= NULL)  {        /* recursively print tree in DESENDING ORDER */
    showTreeHistogram(fd, x->right, nClasses, depth+1);
    
    k = x->n2;
    if(k>0) {
    
        Hist = x->Hist ;
        for(i=0;i<depth*4;i++) fprintf(fd," ");
            /* PRINT BLANK SPACES, so that the printed order
            creates a virual tree shape, 90 degrees 
            rotated with parent as a pivot */
        fprintf(fd,"%d ( <0)", x->nodeNumber) ;
        for(j=0; j<nClasses; j++) fprintf(fd," %d=%5.4f", j, Hist[0][j]);
        fprintf(fd, " MP %d", (x->MP_Class)[0]) ;
        fprintf(fd,"\n") ;
            
        for(i=0;i<depth*4;i++) fprintf(fd," ");
        fprintf(fd,"%d (>=0)", x->nodeNumber) ;
        for(j=0; j<nClasses; j++) fprintf(fd," %d=%5.4f", j, Hist[1][j]);
        fprintf(fd, " MP %d", (x->MP_Class)[1]) ;
        fprintf(fd,"\n") ;
    }
    
    showTreeHistogram(fd, x->left, nClasses, depth+1);
  }
}
            



            

static double localBayesErr(double **Hist,double *Prior, int nClasses, 
    int Index)
/*  int Index ;   */ /* the column selected, 1=left=true */
{
  double sum, Max, BayesErr ;
  int i ;
  
  Max = Hist[Index][0] ;
  sum = Max ;
  
  for(i=1; i<nClasses; i++) {
      sum += Hist[Index][i] ;
      if(Max < Hist[Index][i]) Max=Hist[Index][i] ;
  }
  
  if(fabs(sum-1)>.000001) {
        fprintf(stderr, "Probabilities must sum to 1, %lf.\n",sum) ;
        exit(-1) ;
  }

  BayesErr = 1-Max ;
  BayesErr = BayesErr * Prior[Index] ;
  if(vlevel>100) printf("\tlocal BayesErr=%lf, prior=%lf\n",BayesErr, 
      Prior[Index]);
  
  return(BayesErr);
}





static double totalBayesErr(
  struct TreeNode *x,
  struct Tree_Info *treeInfoPt,
  struct Tree_Data *DataPt,
  int *nleaves,
  double *test     /* test if all the samples added together is correct */
)
{
  struct TreeNode *left, *right ;
  double **Hist, *Prior, Percent;
  double BayesError ;
  double test_left, test_right ;
  int index_left, index_right ;
  int nClasses, nLen, currentSS ;
  
  nClasses = treeInfoPt->nclasses ;
  currentSS = DataPt->CurrentSS ;
  if( currentSS != 0 && currentSS != 1) {
       fprintf(stderr, "ERROR:  invalid currentSS %d, must be 0 or 1\n", 
           currentSS) ;
       exit(-1) ;
  }
  nLen = (DataPt->nlen)[currentSS] ;

  index_left = 1;    /* left means true==1 */
  index_right = 0;
  
  test_left = 0.0 ;
  test_right = 0.0 ;
  BayesError = 0.0 ;
  
  if (x!= NULL)  {        
      
      left = x->left ; 
      right = x->right ;
      Hist = x->Hist ;
      Prior = x->Prior ;
      Percent = (double) ((x->n1) - (x->n0)) / nLen ;    
      
      if((left != NULL) && (right != NULL)) {
           BayesError = totalBayesErr(left, treeInfoPt, DataPt, 
               nleaves, &test_left) ;
           BayesError += totalBayesErr(right, treeInfoPt, DataPt, 
               nleaves, &test_right) ;
           (*test) = test_left + test_right ;
           
      } else if((left == NULL) && (right != NULL)) {

           BayesError = localBayesErr(Hist,Prior,nClasses,index_left)*Percent ;
           (*nleaves)++;
           BayesError += totalBayesErr(right, treeInfoPt, DataPt,
               nleaves, &test_right) ;
           (*test) = Percent * Prior[index_left] + test_right ;
        
      } else if((left != NULL) && (right == NULL)) {

           BayesError = localBayesErr(Hist,Prior,nClasses,index_right)*Percent;
           (*nleaves)++;
           BayesError += totalBayesErr(left, treeInfoPt, DataPt,
               nleaves, &test_left) ;
           (*test) = Percent * Prior[index_right] + test_left ;
        
      
      } else if(((left == NULL) && (right == NULL))) {
      /* no siblings or reaches the level we want */
      
        BayesError = localBayesErr(Hist,Prior,nClasses,index_left)*Percent ;
        (*nleaves)++;
        BayesError += localBayesErr(Hist,Prior,nClasses,index_right)*Percent ;
        (*nleaves)++;
        (*test) = Percent ;
    }
    
    if(vlevel >100) printf("BayesErr=%lf test=%lf per %lf\n", BayesError,
        (*test), Percent ) ;
    return(BayesError) ;
    
  } else {    /* NULL point */
      (*test) = 0 ;
      return(0.0) ;
  }
}

            


static double localEntropy(double **Hist, double *Prior, int nClasses, 
    int Index)
/*  int Index ;  */  /* the column selected, 1=left=true */
{
  double sum ;
  int i ;
  
  sum=0.0;
  for (i=0; i<nClasses; i++)  {
      if(Hist[Index][i] >1) {
         fprintf(stderr, "Probability cannt be greater than 1, %lf.\n",
             Hist[Index][i]) ;
         exit(-1) ;
      }
      if(Hist[Index][i] > 0) {
          sum -= Hist[Index][i] * (double) log((double) Hist[Index][i]) ;
      } 
  }
  
  sum = sum * Prior[Index] ;
  if(vlevel>100) printf("    local entropy=%lf, prior=%lf\n",sum,Prior[Index]);
  
  return(sum);
}



static double totalEntropy(
  struct TreeNode *x,
  struct Tree_Info *treeInfoPt,
  struct Tree_Data *DataPt,
  int *nleaves,
  double *test    /* test if all the percentage added up to 1 */
)
{
  struct TreeNode *left, *right ;
  double **Hist, *Prior, Percent;
  double Entropy ;
  double test_left, test_right ;
  int index_left, index_right ;
  int nClasses, nLen, currentSS ;
  

  nClasses = treeInfoPt->nclasses ;
  currentSS = DataPt->CurrentSS ;
  if( currentSS != 0 && currentSS != 1) {
       fprintf(stderr, "ERROR:  invalid currentSS %d, must be 0 or 1\n", 
           currentSS) ;
       exit(-1) ;
  }
  nLen = (DataPt->nlen)[currentSS] ;
  
  index_left = 1;    /* left means true==1 */
  index_right = 0;
  
  Entropy = 0.0 ;
  test_left = 0.0 ;
  test_right = 0.0 ;
  
  if (x!= NULL)  {        /* recursively print tree in DESENDING ORDER */
      
      left = x->left ; 
      right = x->right ;
      Hist = x->Hist ;
      Prior = x->Prior ;
      Percent = (double)((x->n1)-(x->n0))/ (double) nLen ;   
      
      if((left != NULL) && (right != NULL)) {
           Entropy += totalEntropy(left, treeInfoPt, DataPt, nleaves,
               &test_left) ;
           Entropy += totalEntropy(right, treeInfoPt, DataPt, nleaves,
               &test_right) ;
           (*test) = test_left + test_right ;
           
      } else if((left == NULL) && (right != NULL)) {

           Entropy = localEntropy(Hist,Prior,nClasses,index_left) * Percent ;
           (*nleaves) ++ ;
           Entropy += totalEntropy(right, treeInfoPt, DataPt, nleaves,
               &test_right) ;
           (*test) = Percent * Prior[index_left] + test_right ;
        
      } else if((left != NULL) && (right == NULL)) {

           Entropy = localEntropy(Hist,Prior,nClasses,index_right) * Percent ;
           (*nleaves) ++ ;
           Entropy += totalEntropy(left, treeInfoPt, DataPt, nleaves, 
               &test_left) ;
           (*test) = Percent * Prior[index_right] + test_left ;
        
      
      } else if((left == NULL) && (right == NULL)) {
      /* no siblings or reaches the level we want */
      
        Entropy = localEntropy(Hist,Prior,nClasses,index_left) * Percent ;
           (*nleaves) ++ ;
        Entropy += localEntropy(Hist,Prior,nClasses,index_right) * Percent ;
           (*nleaves) ++ ;
        (*test) = Percent ;
    }
    
    if(vlevel >100) printf("entropy=%lf test=%lf per%lf\n", Entropy,(*test),
        Percent ) ;
    return(Entropy) ;
    
  } else {    /* NULL point */
      (*test) = 0 ;
      return(0.0) ;
  }
}

            
void ShowCurrentResult(struct TreeNode *root, struct Tree_Info *treeInfoPt,
    struct Tree_Data *DataPt) 
{
  
  double Entp, BaysE, ctest ;
  int nLeaves ;
  
  printf("Current Sample Distribution:\n") ;
  showTreeSampleSize(stdout, root, DataPt->nSamples, 0) ;
  printf("\n\n") ;

  printf("Class Distribution at Each Node:\n") ;
  showTreeHistogram(stdout, root, treeInfoPt->nclasses, 0) ;
  printf("\n\n") ;
    
  nLeaves = 0 ;
  Entp=totalEntropy(root, treeInfoPt, DataPt, &nLeaves, &ctest) ;
  printf("\n\tnL=%d  entropy=%lf  ctest=%lf\n", nLeaves, Entp,ctest);
  
  nLeaves = 0 ;
  BaysE= totalBayesErr(root, treeInfoPt, DataPt, &nLeaves, &ctest) ;
  printf("\tnL=%d  Bayes error=%lf  ctest=%lf\n",nLeaves, BaysE,ctest) ;

}

