/*   	Construct a decision tree using linear classification tree.  
	X is a sample set.    We will split a node according to the
	principle axis of the l.m.s prediction.
 
	Jun 17, 1997	Hui Cheng
*/

#include "CPT.h"


int DebugFlag=0 ;    /* default no debug */    
int X0,Y0 ;
int vlevel ;
int defaultClass=2 ;

/* to time all subroutines */
/* struct tms time1, time2, timeStart, timeEnd ; */
time_t TimeGrow,  TimePrune,  TimeInit ;

void error(char *name) ;
void RandInitial() ;

main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  struct CP_Trees CPT ;
  struct Trees *treePt ;
  FILE *fd;
  int  i,j;  		/* indexes for band and column */
  char *infn;        	/* input image file name, Low or High resolution */
  char *otfn;                 /* output segmentation file name */
  int  N, nNeighb, nClasses, nTree ;
  int  iflag,oflag,Cflag, Nflag, Bflag ; 	
  int  **X, **Y ;
  double Tr_Entropy ;
  

  /* malloc_debug(2); */
/*
  TimeInit= TimeGrow = TimePrune = 0 ;
  times(&timeStart) ;
*/

  vlevel = 1 ;
  oflag=iflag=Nflag=Bflag=Cflag=0;
  
  while( (c=getopt(argc,argv,"h,o:,v:,i:,N:,B:,C:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      case 'v':
        if(0==sscanf(optarg,"%d",&vlevel)) error(argv[0]);
        if(vlevel<0) error(argv[0]);
        break;
      case 'N':
        if(0==sscanf(optarg,"%d",&N)) error(argv[0]);
        if(N<=0) error(argv[0]);
        Nflag = 1 ;
        break;
      case 'C':
        if(0==sscanf(optarg,"%d",&nClasses)) error(argv[0]);
        if(nClasses<=0) error(argv[0]);
        Cflag = 1 ;
        break;
      case 'B':
        if(0==sscanf(optarg,"%d",&nNeighb)) error(argv[0]);
        if(nNeighb<=0) error(argv[0]);
        Bflag = 1 ;
        break;
      default:
        error(argv[0]);
        break;
    }
  }


  /* Check for inconsistant inputs */   
  if(!(oflag && iflag && Nflag && Bflag && Cflag)) {
    fprintf(stderr, "input,output,maximal # of leaves, the number of data, ") ;
    fprintf(stderr, " # of neighb, and nclasses are required ");
    fprintf(stderr, " \n");
    error(argv[0]);
  }
   

  /* initial CPT parameters */ 
  CPT.nlevels = 2 ;  
  CPT.info.nclasses = nClasses ;
  CPT.info.dimen = nNeighb ;
  CPT.info.nTree = nTree = 4 ;
  CPT.info.RangeF = 2 ;
  CPT.info.MaxLeaves = 100;
  CPT.info.MinSamples = 2 ;
  CPT.info.Margin = (int) ( sqrt((double) CPT.info.dimen) ) ;
  if(CPT.info.Margin * CPT.info.Margin != CPT.info.dimen) {
      fprintf(stderr, "ERROR: CPT dimen %d must be the square of an integer\n",
          CPT.info.dimen) ;
      exit(-1) ;
  } else {
      CPT.info.Margin = (CPT.info.Margin-1) /2 ;
  }
    
  InitializeCPT( &CPT ) ; 
  treePt = &((CPT.trees)[1]) ;
  AllocateCPT_Data(treePt, N);
  
  X = (treePt->data).X ;
  Y = (treePt->data).Y ;
  
    
  if(vlevel >0) printf("\treading input file %s ...... \n", infn) ;
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;

  for(i=0; ((!feof(fd))&&(i<N)); i++) {
  	for(j=0; j<nTree; j++) {
  		if(feof(fd)) {
  			fprintf(stderr, "Not enough data, read %d !\n",i) ;
  			exit(-1) ;
  		}
  		fscanf(fd, "%d", &(Y[j][i])) ;	/* desired output */
  	}
  	
  	for(j=0; j<nNeighb; j++) {
  		if(feof(fd)) {
  			fprintf(stderr, "Not enough data, read %d !\n",i) ;
  			exit(-1) ;
  		}
  		/* NOTE:  when we save it, there are 2 spaces after an integer */
  		fscanf(fd, "%d  ", &(X[i][j])) ;
  	}
  }
  fclose(fd);
  N=i;  printf("N=%d samples are read.\n\n",N); 
  (treePt->data).nSamples = N ;


  Tr_Entropy = 0 ;
/*  
  for(i=0; i<nTree; i++)  printf("%lf\n", entropy(Y[i],0,N)) ;
  for(i=0; i<nTree; i++)  Tr_Entropy += entropy(Y[i],0,N) ;
  printf("Initial entropy %f\n",Tr_Entropy) ;
*/
  
/*
  times(&time1) ;
  TimeInit = time1.tms_utime - timeStart.tms_utime ;
*/  

  GelfandPruning( treePt ) ;

  SaveCPT_Par( &CPT, otfn ) ;


/*
  times(&timeEnd) ;
  printf("Timewhole %d\n", timeEnd.tms_utime - timeStart.tms_utime) ;
  printf("TimeInit %d\n", TimeInit) ;
  printf("TimeGrow %d\n", TimeGrow) ;
  printf("TimePrune %d\n", TimePrune) ;
*/
}




void error(name)
char *name;
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i sample -- input file, nNeigb+1 column format ;\n");
  fprintf(stderr,"     -o output -- Segmentation image in pgm format\n");
  fprintf(stderr,"     -N data   -- Approximate number of data, \n");
  fprintf(stderr,"		  N should larger than real number of data N>0\n");
  fprintf(stderr,"     -B number   -- nNeighb, B>0\n");
  fprintf(stderr,"     -C number   -- nClasses, B>0\n");
  fprintf(stderr,"     -v level  -- Verbose operation; level>=0\n");

  exit(1);
}

