#ifndef CPT_DEF
#define CPT_DEF

/******************************************************************************/
/* TreeNode is used in feature extraction */
/******************************************************************************/
struct TreeNode 
{
    int level;	/* the level of this node */
    double **Hist;	/* histogram according to min entropy [RangeF][nClasses]*/
    double **T ;	/* Linear partition function */
		    /* T[nNeighbors][nClasses-1] */
    double T0 ;	/* constant term of the linear partition function */

    struct TreeNode *left;	/* left sibling 
    			   or the predicessor if used in NodePool */
    struct TreeNode *right;	/* right sibling 
    			   or the child if used in NodePool */
    struct TreeNode *parent;	/* parent */


    int nodeNumber ;	/* to distinguish each node */
    int IsLeftChild;	/* this node is the Left child */
    int n0 ;	/* starting and ending indices in sample vector */
    int n1 ;	/* n0 to n1-1, the length is n1-n0 */
    int n2 ;	/* after a boolean function is choosen,
		       the sample will be partitioned into two parts:
		       (n0 to n2-1) and (n2 to n1-1) */
    double entropy ; /* the entropy after this partition */
    double costDrop; /* the drop of cost */

    double *Prior;	/* prior according to min entropy*/


    /* for pruning */
    double entpBelow ;	/* entropy due to the leaves below, for pruning */
    int nLeavesBelow ;	/* number of leaves below this node, used for pruning */
    double testCost ;
    double testCostBelow ;
    double *testCostLR ;
    int *MP_Class ;
	
} ;


/*****************************************************************************/
/*****************************************************************************/
struct Trees
{
	struct Tree_Info      /* General parameters for CPT's */
	{
    	    int nclasses ;    
    	    int dimen ;       /* dimension of input vectors */
    	    int Margin ;
    	    int nTree ;       /* number of trees */
    	    int RangeF ;      /* number of child of a tree node */
    	                      /* for binary tree, it is 2 */

    	    /* for tree construction */
    	    int MaxLeaves ;   /* max number of leaves of a tree */
    	    int MinSamples ;  /* min number of samples for a node */

	}  *infoPt ;
	
	struct TreeNode **roots;  /* root[nTree] */
	int *nLeaves ;            /* nLeaves[nTree] */
	
	struct Tree_Data
	{
	    int nSamples ;
	    int **Y ;	      /* Y[nTree][nSamples] desired output */
            int **X ;         /* X[nSamples][dimen] input */

            int nL[2] ;       /* partition sample set into two parts */
            int nH[2] ;       /* for Gelfand pruning */
            int nlen[2] ;     /* nH[.]-nL[.] */
            int CurrentSS ;   /* current sample set: 0 or 1 */
	} data ;
} ;


/*****************************************************************************/
/*****************************************************************************/
struct CP_Trees
{
    int nlevels ;        
	struct Tree_Info info;   /* General parameters for CPT's */
	struct Trees *trees;     /* Tree[level] */
} ;

#endif  /* CPT_DEF */





