/*****************************************************************************/
/* This program will refine the groundtruth images segmented by hand. */
/* Both color map and class labels will be justed to make sure that */
/* there is no inconsistency. */
/*****************************************************************************/
 

#include "../../utilities/src/util.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <time.h>


void error(char *name);
static int valid_seg(eisl_img *seg_map, uint16 *class_label, int *nclass_here) ;
static int standarlize_hs( char *hs_fn, char *out_fn, int nClasses) ;
static int Init_color_map(eisl_img *img) ;


int main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd;
  char *infn;       	      /* input image file */
  char *otfn;                 /* output segmentation file name */
  char filename[256], filename2[256] ;
  int  nFiles ;		          /* number of input files */
  int  oflag,iflag;
  int  i, nClasses ;


  iflag=oflag=0;
  
  while( (c=getopt(argc,argv,"h,i:,o:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      default:
        fprintf(stderr, "Invalid char %c\n",c) ;
        error(argv[0]);
        break;
    }
  }


  /* Check for inconsistant inputs */   
  if(!(oflag&&iflag)) {
    fprintf(stderr, "Input,output are required!\n",argv[0]);
    error(argv[0]);
  }


  /* read input image */
  printf("\treading input file %s \n", infn) ;
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;
  fscanf(fd, "nFiles=%d\n", &nFiles) ;
  fscanf(fd, "nClasses=%d\n", &nClasses) ;
 

  /* decimate original images */
  for(i=0; i<nFiles; i++) {
      fscanf(fd, "%s", filename) ;   /* name of org images */
      fscanf(fd, "%s", filename) ;   /* name of ground truth images */
      sprintf(filename2, "%s.hs%d", otfn, i) ;
      if( ! standarlize_hs(filename, filename2, nClasses) ) {
          fprintf(stderr, "ERROR: standarlize_hs()\n") ;
          goto Bad ;
      }
  }


Good:
  exit(0) ;  /* use this to ingore un-de-allocated memory */

Bad:
  exit(-1) ;  
}




void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");

  exit(1);
}


/*****************************************************************************/
/*  Read in orginal image and ground truth segmentation */
/*  Check for any inconsistency and store their dimensions */
/*****************************************************************************/

static int standarlize_hs( char *hs_fn, char *out_fn, int nClasses )
{
  char filename[256] ;
  int  i,j, ii,jj ;
/*  int  nclass_here ; */
  unsigned long nRowIn, nColIn, nRowOut, nColOut ;
  uint16 *rmap, *gmap, *bmap ;
  eisl_img img_out ;
  eisl_img img_in ;                     /* input image, tiff or pxm format,*/
                                        /* ../../utilities/src/eisl_img.h*/
/*  uint16 class_label[256];  */
  /*   class_label[0] :  TWOCOLOR   */
  /*   class_label[1] :  IMG        */
  /*   class_label[2] :  ONECOLOR   */
  /*   class_label[3] :  HIGH       */
  /*   class_label[4] :  SWEEP      */

  
  /* orginal image */
  printf("\treading ground truth image %s .\n", hs_fn) ;
  if( ! read_img (hs_fn, &img_in) ) {
      fprintf(stderr,"ERROR: Cannot read image %s!\n", hs_fn);
      goto Bad ;
  }
  sprintf(filename, "%s.tiff", out_fn) ;
  
  
  /* allocate image_out */
  nRowIn = nRowOut = (unsigned long) img_in.height ;
  nColIn = nColOut = (unsigned long) img_in.width ;
  
  printf("\tnRowIn %d, nColIn %d, nRowOut %d, nColOut %d, %c\n",
      nRowIn, nColIn, nRowOut, nColOut, img_in.img_type ) ;

  if(! allocate_img(&img_out, nRowOut, nColOut, 'p') ) {
      fprintf(stderr, "ERROR: cannot allocate output image %s\n",
          filename) ;
      goto Bad ;
  }
 
  if(! Init_color_map(&img_out) ) {
      fprintf(stderr, "ERROR:  Cannot initial color map\n") ;
      goto Bad ;
  }

/*  if(! valid_seg(&img_in, class_label, &nclass_here) ) {
      fprintf(stderr, "ERROR: valid_seg()\n") ;
      goto Bad ;
  } 
  if( nclass_here > nClasses ) {
      fprintf(stderr, "ERROR: # classes found %d > # classes read in %d \n",
          nclass_here, nClasses) ;
      goto Bad ;
  } 
 
  printf("# of classes found: %d\n", nclass_here) ;
*/  

  rmap = img_in.rmap ;
  gmap = img_in.gmap ;
  bmap = img_in.bmap ;

  for(i=0;  i<nRowOut;  i++)
  for(j=0;  j<nColOut;  j++) {
      
      ii = (img_in.mono)[i][j] ;

      
      if ( rmap[ii]==(uint16) 255  && gmap[ii]==(uint16) 0  &&     
    	      bmap[ii]==(uint16) 0 ) {
    	      (img_out.mono)[i][j] = 0 ;       /* text */
      } else 
      if ( rmap[ii]==(uint16) 0  && gmap[ii]==(uint16) 255  &&     
    	      bmap[ii]==(uint16) 0 ) {
    	      (img_out.mono)[i][j] = 1 ;       /* img */
      } else 
      if ( rmap[ii]==(uint16) 0  && gmap[ii]==(uint16) 0  &&     
    	      bmap[ii]==(uint16) 255 ) {
    	      (img_out.mono)[i][j] = 2 ;       /* bkgd */
      } else 
      if ( rmap[ii]==(uint16) 255  && gmap[ii]==(uint16) 255  &&     
    	      bmap[ii]==(uint16) 255 ) {
    	      (img_out.mono)[i][j] = 3 ;       /* heading */
      } else 
      if ( rmap[ii]==(uint16) 255  && gmap[ii]==(uint16) 0  &&     
    	      bmap[ii]==(uint16) 255 ) {
    	      (img_out.mono)[i][j] = 4 ;       /* bkgd */
      } else 
      if ( rmap[ii]==(uint16) 0  && gmap[ii]==(uint16) 255  &&     
    	      bmap[ii]==(uint16) 255 ) {
    	      (img_out.mono)[i][j] = 5 ;       /* bkgd */
      } else {
          fprintf(stderr, "Invalid color %d, %d, %d, at %d, %d\n", 
              rmap[ii], bmap[ii], gmap[ii], i,j ) ;
          exit(-10) ;
      }
  }

  /* write out the decimated image */ 
  
  if(! write_img(filename, &img_out, 't', "lzw:2") ) {
      fprintf(stderr, "ERROR: can not write out %s\n", filename) ;
      goto Bad ;
  }

  free_eisl_img( &img_in ) ; 
  free_eisl_img( &img_out ) ; 
  return(TRUE) ;
  
Bad :
  free_eisl_img( &img_in ) ;
  free_eisl_img( &img_out ) ; 
  return(FALSE) ;
  
}


/******************************************************************************/
/*  check for valid segmentation map */
/*  find out the corresponding indices for each class */
/******************************************************************************/
static int valid_seg(eisl_img *seg_map, uint16 *class_label, int *nclasses)
  /*   class_label[0] :  TWOCOLOR   */
  /*   class_label[1] :  IMG        */
  /*   class_label[2] :  ONECOLOR   */
  /*   class_label[3] :  HIGH       */
  /*   class_label[4] :  SWEEP      */
{
	uint32 nrS, ncS, i,j, MaxColor ;
	uint16 *rmap, *gmap, *bmap ;
	uint8  **mono, m ;
	int32 tmp, nH, nOne, nTwo, nI, nSwp ;
	
	MaxColor = 10 ;
	
	if( seg_map == NULL ) {
		fprintf(stderr, "input pt to seg_map is NULL\n") ;
		return(FALSE) ;
	}
	
	if( seg_map->img_type != 'p') {
		fprintf(stderr, "Need a palette image, not type %c\n", 
			seg_map->img_type ) ;
		return(FALSE) ;
	}
	
	nrS = seg_map->height ;
	ncS = seg_map->width ;
	
	
	/* check for color map */
	
	rmap = seg_map->rmap ;
	gmap = seg_map->gmap ;
	bmap = seg_map->bmap ;

	printf("CMAP of seg map \n") ;
	for(i=0; i<MaxColor; i++) {
    	    printf("%d,   %d, %d, %d\n", i, rmap[i], gmap[i], bmap[i]) ;
	}
	printf("\n") ;

	for(i=0; i<MaxColor; i++) class_label[i] = 255 ;
	*nclasses = 0 ;

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 0  &&     
    	 	    bmap[i]==(uint16) 0 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[0] = (uint16) tmp ;	
    	    printf("TEXT color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No TEXT color found !\n") ;
	}

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 0  && gmap[i]==(uint16) 255  &&     
    	 	    bmap[i]==(uint16) 0 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[1] = (uint16) tmp ;	
    	    printf("IMG color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No IMG color found !\n") ;
	}

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 0  && gmap[i]==(uint16) 0 &&     
    	 	    bmap[i]==(uint16) 255 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[2] = (uint16) tmp ;	
    	    printf("BKGD color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No BKGD color found !\n") ;
	}


	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 255  &&     
    	 	    bmap[i]==(uint16) 255 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[3] = (uint16) tmp ;	
    	    printf("HEADING color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No HEADING color found !\n") ;
	}


	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 0  &&     
    	 	    bmap[i]==(uint16) 255 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[4] = (uint16) tmp ;	
    	    printf("COLOR_BKGD color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No COLOR_BKGD color found !\n") ;
	}

	printf("color indices: \n") ;
	printf("  TWOCOLOR %d \n", class_label[0]) ;
	printf("  IMG %d\n", class_label[1]) ;
	printf("  ONECOLOR %d \n", class_label[2]) ;
	printf("  HEADING %d \n", class_label[3]) ;
	printf("  COLOR_BKGD %d \n", class_label[4]) ;
        printf("  # of classes found :  %d\n", *nclasses) ;
	

	/* check color indices */
	mono = seg_map->mono ;
	nH = nI = nOne = nTwo = nSwp = 0 ;
	for(i=0; i<nrS; i++) 
	for(j=0; j<ncS; j++) {
		
		m = mono[i][j] ;
		if( m==class_label[1] ) nI ++ ;
		else if ( m==class_label[3] ) nH ++ ;
		else if ( m==class_label[2] ) nOne ++ ;
		else if ( m==class_label[0] ) nTwo ++ ;
		else if ( m==class_label[4] ) nSwp ++ ;
		else {
			fprintf(stderr, "Invalid class labe %d\n", mono[i][j]) ;
			return(FALSE) ;
		}
	}
	
	printf("nI %d, nHeading %d, nOne %d, nTwo %d, nColorBKGD %d\n", 
		nI, nH, nOne, nTwo, nSwp) ;
	return(TRUE) ;
}



/******************************************************************************/
/*  check for valid segmentation map */
/*  find out the corresponding indices for each class */
/******************************************************************************/
static int valid_seg_comp(eisl_img *seg_map, uint16 *class_label, int *nclasses)
  /*   class_label[0] :  TWOCOLOR   */
  /*   class_label[1] :  IMG        */
  /*   class_label[2] :  ONECOLOR   */
  /*   class_label[3] :  HIGH       */
  /*   class_label[4] :  SWEEP      */
{
	uint32 nrS, ncS, i,j, MaxColor ;
	uint16 *rmap, *gmap, *bmap ;
	uint8  **mono, m ;
	int32 tmp, nH, nOne, nTwo, nI, nSwp ;
	
	MaxColor = 10 ;
	
	if( seg_map == NULL ) {
		fprintf(stderr, "input pt to seg_map is NULL\n") ;
		return(FALSE) ;
	}
	
	if( seg_map->img_type != 'p') {
		fprintf(stderr, "Need a palette image, not type %c\n", 
			seg_map->img_type ) ;
		return(FALSE) ;
	}
	
	nrS = seg_map->height ;
	ncS = seg_map->width ;
	
	
	/* check for color map */
	
	rmap = seg_map->rmap ;
	gmap = seg_map->gmap ;
	bmap = seg_map->bmap ;

	printf("CMAP of seg map \n") ;
	for(i=0; i<MaxColor; i++) {
    	    printf("%d,   %d, %d, %d\n", i, rmap[i], gmap[i], bmap[i]) ;
	}
	printf("\n") ;

	for(i=0; i<MaxColor; i++) class_label[i] = 255 ;
	*nclasses = 0 ;

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 0  && gmap[i]==(uint16) 255  &&     
    	 	    bmap[i]==(uint16) 0 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[1] = (uint16) tmp ;	
    	    printf("IMG color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No IMG color found !\n") ;
	}

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 255  &&     
    	 	    bmap[i]==(uint16) 0 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[3] = (uint16) tmp ;	
    	    printf("HIGH color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No HIGH color found !\n") ;
	}


	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 0  && gmap[i]==(uint16) 0  &&     
    	 	    bmap[i]==(uint16) 255 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[2] = (uint16) tmp ;	
    	    printf("ONECOLOR color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No ONECOLOR color found !\n") ;
	}


	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 0  &&     
    	 	    bmap[i]==(uint16) 0 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[0] = (uint16) tmp ;	
    	    printf("TWOCOLOR color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No TWOCOLOR color found !\n") ;
	}

	for(i=0, tmp=-1 ; i<MaxColor; i++) {
    	    if ( rmap[i]==(uint16) 255  && gmap[i]==(uint16) 0  &&     
    	 	    bmap[i]==(uint16) 255 ) {
    		    tmp = (int32) i ;
    		    break ;
    	    }
	}
	if( tmp != -1 ) {
    	    class_label[4] = (uint16) tmp ;	
    	    printf("SWEEP color index found\n") ;
            (*nclasses) ++ ;
	} else {
    	    printf("No SWEEP color found !\n") ;
	}

	printf("color indices: \n") ;
	printf("  IMG %d\n", class_label[1]) ;
	printf("  HIGH %d \n", class_label[3]) ;
	printf("  ONECOLOR %d \n", class_label[2]) ;
	printf("  TWOCOLOR %d \n", class_label[0]) ;
	printf("  SWEEP %d \n", class_label[4]) ;
        printf("  # of classes found :  %d\n", *nclasses) ;
	

	/* check color indices */
	mono = seg_map->mono ;
	nH = nI = nOne = nTwo = nSwp = 0 ;
	for(i=0; i<nrS; i++) 
	for(j=0; j<ncS; j++) {
		
		m = mono[i][j] ;
		if( m==class_label[1] ) nI ++ ;
		else if ( m==class_label[3] ) nH ++ ;
		else if ( m==class_label[2] ) nOne ++ ;
		else if ( m==class_label[0] ) nTwo ++ ;
		else if ( m==class_label[4] ) nSwp ++ ;
		else {
			fprintf(stderr, "Invalid class labe %d\n", mono[i][j]) ;
			return(FALSE) ;
		}
	}
	
	printf("nI %d, nH %d, nOne %d, nTwo %d, nSwp %d\n", 
		nI, nH, nOne, nTwo, nSwp) ;
	return(TRUE) ;
}



static int Init_color_map(eisl_img *img) 
{
  int i, KK ;
  
  img->bitspersample = 8 ;
  KK = (1<<(img->bitspersample)) ;    /* 2^bitspersample */
  img->rmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  img->gmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  img->bmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  if( (img->rmap == NULL) || (img->gmap == NULL) ||
      (img->bmap == NULL) ) {
    fprintf(stderr, "ERROR: Can not allocate color map, write_segment\n");
    return(FALSE) ;
  }

  /* Has to be initialied.  o.w. may cause confusion between */
  /* 8-bit and 16-bit color map,  randomly initialization */
  for(i=0; i<KK; i++) {  
      img->rmap[i] = rand() % KK ;
      img->gmap[i] = rand() % KK ;
      img->bmap[i] = rand() % KK ;
  }

  /* Since the number of classes are usually less than 10 */
  /* manually set the first 8 colors */
  img->rmap[0]=255;  img->gmap[0]=  0;  img->bmap[0]=  0;  
  img->rmap[1]=  0;  img->gmap[1]=255;  img->bmap[1]=  0;  
  img->rmap[2]=  0;  img->gmap[2]=  0;  img->bmap[2]=255;  
  img->rmap[3]=255;  img->gmap[3]=255;  img->bmap[3]=255;  
  img->rmap[4]=255;  img->gmap[4]=  0;  img->bmap[4]=255;  
  img->rmap[5]=  0;  img->gmap[5]=255;  img->bmap[5]=255;  
  img->rmap[6]=255;  img->gmap[6]=255;  img->bmap[6]=  0;  
  img->rmap[7]=  0;  img->gmap[7]=  0;  img->bmap[7]=  0;  
  
  return(TRUE) ;
}
