/* Use SVD to calculate the pseudoinverse of a matrix
   Notice: svdcmp() starts indices from 1 instead of 0
   

   Jun 13, 1997,		Hui Cheng
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../../utilities/src/nrutil.h"
#include "../../utilities/src/allocate.h"

#define TINY 1.0e-15

extern int vlevel ;
/* inserted following 4 lines -GF*/
void PrintMatrix(FILE *fd, double **a, int rl, int rh, int cl, int ch);
void ePrintVector(FILE *fd, double *a, int nl, int nh) ;
void  MatrixMultiply(
  double **In1, int rl1, int rh1, int cl1, int ch1, int T1,
  double **In2, int rl2, int rh2, int cl2, int ch2, int T2,
  double **Out

) ;
int svdcmp(double **a, int m, int n, double *w, double **v);	

int pinv(double **a, int m, int n, double **b, int ChangeIndex)
  /* a=input, b=output matrices */
  /* m, n -- number of row, col */
  /* If ChangeIndex>0, change index. Index starts from 1 instead of from 0 */
{
  double tol ;
  static double **v, *w, **u, **out ;
  static int M=-1, N=-1, NN=-1, flag1=0, flag2=0 ;
  int i,j ;

  if(N!=n) {
  	if(0) {
  		free_vector(w,1,n) ;
  		free_matrix(v,1,n,1,n) ;
  	}
  	v = (double **) matrix(1,n,1,n) ;
  	w = vector(1,n) ; 
  	N=n;  flag1 = 1 ; 
  }
  
  if(ChangeIndex) {
  	if((M!=m) || (NN!=n) || (flag2==0))  {
  		if(0 && flag2) {
  			free((char*) (u+1)) ;
  			free((char*) (out+1)) ;
  		}
  		u = (double **) get_spc(m,sizeof(double*)) ;
		u--;  M=m ;

  		out = (double **) get_spc(n,sizeof(double*)) ;
		out--;  NN=n;
	}
  	for(i=1; i<=m; i++) u[i]=a[i-1]-1 ;
  	for(i=1; i<=n; i++) out[i]=b[i-1]-1 ;
  	
  	flag2 = 1 ;

  } else {
  	if(0 && flag2) {
  		free((char*) (u+1)) ;
  		free((char*) (out+1)) ;
  	}
  	u = a;
	out = b ;
	flag2=0 ;
  }

  if(vlevel>100)  PrintMatrix(stdout,u,1,m,1,n) ;
  if(vlevel>100)  printf("%d  %d \n", m,n) ;
  
  
  if(!(svdcmp(u,m,n,w,v))) {
	fprintf(stderr, "svd u:\n") ;
   	PrintMatrix(stderr,u,1,m,1,n) ;
  	return(0) ;	/* ~hui/bin/num_recipes */
  }
 
  
  if(vlevel>100)  printf("svd u:\n") ;
  if(vlevel>100)  PrintMatrix(stdout,u,1,m,1,n) ;
  if(vlevel>100)  printf("svd d:\n") ;
  if(vlevel>100)  ePrintVector(stdout,w,1,n) ;
  if(vlevel>100)  printf("svd v:\n") ;
  if(vlevel>100)  PrintMatrix(stdout,v,1,n,1,n) ;
  if(vlevel>100)  printf("u*d*v_t\n") ;
  

  /* calculate condition number, from matlab pinv.m */
  tol=fabs(w[1]) ;
  for(i=2; i<=n; i++) 
  	if(fabs(w[i])>tol) tol=fabs(w[i]) ;
  if(m>n) tol*=m ;
  else tol*=n ;
  tol *= TINY ;

  /* multiply v * diag(1/w_i) */
  for(i=1; i<=n; i++) {
  	if(fabs(w[i])>tol) for(j=1; j<=n; j++) v[j][i] /= w[i] ;
  	else if(fabs(w[i])>0) {
  		for(j=1; j<=n; j++) v[j][i]=0.0 ;
  		if(vlevel>100) fprintf(stderr, "\tWarning:  Near Zero, but not Zero eigenvalues\n") ;
  	} else for(j=1; j<=n; j++) v[j][i]=0.0 ;
  }
  
  MatrixMultiply(v,1,n,1,n,0,u,1,m,1,n,1,out) ;
  
  if(vlevel>100)  printf("pinvXX, out\n") ;
  if(vlevel>100)  PrintMatrix(stdout, out,1,n,1,m) ;

  return(1) ;
}



#undef TINY
