/*****************************************************************************/
/* This program will manager all the training processes fo the */
/* Multiscale Contextual Segmentation Package */
/*   */
/* Hui Cheng,   Jan 27, 1998  */
/*****************************************************************************/

#include "EM_Dec.h"

extern int vlevel ;

/*****************************************************************************/
/*  Calculate the  max scales allowed for an image */
/*****************************************************************************/
static int levels(int nrow, int ncol)
{
    int dd;

    dd=0;
    while((nrow>2)&&(ncol>2)) {
      dd++;
      nrow = nrow/2 + nrow%2;
      ncol = ncol/2 + ncol%2;
      }
    return(dd);
}



/*****************************************************************************/
/*  Read in orginal image and ground truth segmentation */
/*  Check for any inconsistency and store their dimensions */
/*****************************************************************************/

int read_training_images( FILE *fd, int nClasses, int nFiles, char **org_fn, 
    char **hs_fn, int *nRow, int *nCol )
{
  char filename[256] ;
  int  i,ii,jj ;
  eisl_img img_in ;                       /* input image, tiff or pxm format,*/
                                          /* ../../utilities/src/eisl_img.h  */
  unsigned char **hs_seg, seg, M ;

  /* read in training images, check for inconsistancy */
  
  M = (unsigned char) nClasses ;
  
  for(i=0; i<nFiles; i++) {
  
      /* orginal image */
      
      fscanf(fd, "%s", filename) ;
      if(vlevel >0) printf("\treading original image %s .\n", filename) ;
	  if( ! read_img_header (filename, &img_in) ) {
          fprintf(stderr,"ERROR: Cannot read header of %s!\n", filename);
          goto Bad ;
      }
      if( img_in.img_type != 'g' ) {
          fprintf(stderr,"ERROR: orginal image must be gray level image,") ;
          fprintf(stderr," %s, read_training_images\n", filename) ;
          goto Bad ;
      }
      nCol[i] = img_in.width ;	 
      nRow[i] = img_in.height ;
      sprintf(org_fn[i], "%s", filename) ;
      


      /* ground truth image */
    
      fscanf(fd, "%s", filename) ;
      sprintf(hs_fn[i], "%s", filename) ;
      if(vlevel >0) printf("\treading ground truth segmentation %s .\n", 
          filename) ;
	  if( ! read_img (filename, &img_in) ) {
          fprintf(stderr,"ERROR: Cannot read image %s!\n", filename);
          goto Bad ;
      }
      if( img_in.img_type != 'g' && img_in.img_type != 'p' ) {
          fprintf(stderr,"ERROR: ground truth must be a gray level or") ;
          fprintf(stderr," a pallete image\n") ;
          goto Bad ;
      }
      
      if( (nCol[i] != img_in.width) || (nRow[i] != img_in.height) ) {
          fprintf(stderr,"ERROR: dimension conflicts between orginal and") ;
          fprintf(stderr," ground truth segmentation %s.\n", filename) ;
          goto Bad ;
      }
      nCol[i] = img_in.width ;	   /* EM_Dec mainly works with ground truth */
      nRow[i] = img_in.height ;    /* This makes sure that the dimension of */
                                   /* ground truth segmentation is used     */


      /* check if all pixel values of ground truth segmentation */
      /* are in the range {0,1, ... , nClasses-1} */
      
      hs_seg = img_in.mono ;
      for(ii=img_in.height -1; ii>=0; ii--)
      for(jj=img_in.width -1;  jj>=0; jj--) {
          seg = hs_seg[ii][jj] ;
          if( seg >= M ) {
              fprintf(stderr, "ERROR: class label out of range\n") ;
              fprintf(stderr, "\trow %d, col %d,  seg %d,", ii,jj,seg);
              fprintf(stderr, "  nclasses %d \n", nClasses );
              goto Bad ;
          }
      }
      free_eisl_img( &img_in ) ; 
      
  }

  return(TRUE) ;
  
Bad :
  free_eisl_img( &img_in ) ;
  return(FALSE) ;
  
}



/*****************************************************************************/
/*  Calculate the min ( max scales allowed for each image ) */
/*****************************************************************************/
int number_resolutions_allowed( int DD, int *nRow, int *nCol, int nFiles ) 
{
    int i, i0, min, tmp ;

    if( DD>0 ) {
        i0 = 0 ;
        min = DD ;
        
    } else {
        i0 = 1 ;
        min = levels(nRow[0], nCol[0]) ;
    }
    
    for(i=i0; i<nFiles; i++) {
        tmp = levels(nRow[i], nCol[i]) ;
        if(min > tmp) min = tmp ;
    }
    
    return( min ) ;
}



