/*****************************************************************************/
/* This program will upsample a segmentation 2x2. */
/*****************************************************************************/
 

#include "../../utilities/src/util.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <time.h>


void error(char *name);
static int upsample_hs( char *org_fn, char *hs_fn, char *out_fn, int nClasses);
static int Init_color_map(eisl_img *img) ;


int main(argc,argv)
int     argc;
char    *argv[];
{
  extern char   *optarg;
  extern int    optind;
  char          c;
 
  FILE *fd;
  char *infn;       	      /* input image file */
  char *otfn;                 /* output segmentation file name */
  char fn_hs[256], fn_org[256], filename[256], filename2[256] ;
  int  nFiles ;		          /* number of input files */
  int  oflag,iflag;
  int  i, nClasses ;


  iflag=oflag=0;
  
  while( (c=getopt(argc,argv,"h,i:,o:")) !=EOF ) {
    switch(c) {
      case 'h':		/* help */
        error(argv[0]);
        break;
      case 'i':		/* input original image file name */
        iflag=1;
        infn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(infn,optarg);
        break;
      case 'o':		/* output file name */
        oflag=1;
        otfn = (char *)mget_spc((strlen(optarg)+1),sizeof(char));
        strcpy(otfn,optarg);
        break;
      default:
        fprintf(stderr, "Invalid char %c\n",c) ;
        error(argv[0]);
        break;
    }
  }


  /* Check for inconsistant inputs */   
  if(!(oflag&&iflag)) {
    fprintf(stderr, "Input,output are required!\n",argv[0]);
    error(argv[0]);
  }


  /* read input image */
  printf("\treading input file %s \n", infn) ;
  fd = fopen (infn, "r");  if (fd == NULL) FileOpenErr(infn) ;
  fscanf(fd, "nFiles=%d\n", &nFiles) ;
  fscanf(fd, "nClasses=%d\n", &nClasses) ;
 

  /* decimate original images */
  for(i=0; i<nFiles; i++) {
      fscanf(fd, "%s", fn_org) ;   /* name of org images */
      fscanf(fd, "%s", fn_hs ) ;   /* name of ground truth images */
      sprintf(filename2, "%s.hs%d", otfn, i) ;
      if( ! upsample_hs(fn_org, fn_hs, filename2, nClasses) ) {
          fprintf(stderr, "ERROR: standarlize_hs()\n") ;
          goto Bad ;
      }
  }


Good:
  exit(0) ;  /* use this to ingore un-de-allocated memory */

Bad:
  exit(-1) ;  
}




void error(char *name)
{
  fprintf(stderr,"\nUsage: %s\n",name);
  fprintf(stderr,"     -help -- produce this help menu\n");
  fprintf(stderr,"     -i input -- the file must be in the following format\n");
  fprintf(stderr,"        nFiles=# of training images\n");
  fprintf(stderr,"        nClasses=# of classes\n");
  fprintf(stderr,"        filename_original_image_no_1\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_1\n");
  fprintf(stderr,"        ......\n");
  fprintf(stderr,"        filename_original_image_no_nFiles\n");
  fprintf(stderr,"        filename_ground_truth_segment_no_nFiles\n");
  fprintf(stderr,"        image files can be either TIFF or PXM\n");
  fprintf(stderr,"     -o output -- filename of output parameter files\n");
  fprintf(stderr,"        file extension will be provided by the program\n");

  exit(1);
}


/*****************************************************************************/
/*  Read in orginal image and ground truth segmentation */
/*  Check for any inconsistency and store their dimensions */
/*****************************************************************************/

static int upsample_hs( char *org_fn, char *hs_fn, char *out_fn, int nClasses )
{
  char filename[256] ;
  int  i,j, ii,jj ;
  int  nclass_here ;
  unsigned long nRowIn, nColIn, nRowOut, nColOut ;
  eisl_img img_org ;
  eisl_img img_out ;
  eisl_img img_in ;                     /* input image, tiff or pxm format,*/
                                        /* ../../utilities/src/eisl_img.h*/
  
  /* orginal image */
  printf("\treading ground truth image %s .\n", hs_fn) ;
  if( ! read_img_header (org_fn, &img_org) ) {
      fprintf(stderr,"ERROR: Cannot read image header %s!\n", org_fn);
      goto Bad;
  }

  /* org hs image */
  printf("\treading ground truth image %s .\n", hs_fn) ;
  if( ! read_img (hs_fn, &img_in) ) {
      fprintf(stderr,"ERROR: Cannot read image %s!\n", hs_fn);
      goto Bad ;
  }
  sprintf(filename, "%s.tiff", out_fn) ;
  
  
  /* allocate image_out */
  nRowIn = (unsigned long) img_in.height ;
  nColIn = (unsigned long) img_in.width ;
  nRowOut = (unsigned long) img_org.height ;
  nColOut = (unsigned long) img_org.width ;
  
  printf("\tnRowIn %d, nColIn %d, nRowOut %d, nColOut %d, %c\n",
      nRowIn, nColIn, nRowOut, nColOut, img_in.img_type ) ;
  if( nRowIn != ((nRowOut+1)/2) || nColIn != ((nColOut+1)/2) ) {
      fprintf(stderr, "ERROR: dimension conflict !") ;
      exit(-10) ;
  } 

  if(! allocate_img(&img_out, nRowOut, nColOut, 'p') ) {
      fprintf(stderr, "ERROR: cannot allocate output image %s\n",
          filename) ;
      goto Bad ;
  }
 
  if(! Init_color_map(&img_out) ) {
      fprintf(stderr, "ERROR:  Cannot initial color map\n") ;
      goto Bad ;
  }

  for(i=0;  i<nRowOut;  i++)
  for(j=0;  j<nColOut;  j++) {
      
      ii = (img_in.mono)[i/2][j/2] ;

      if (ii>=0 && ii<nClasses)  {
          (img_out.mono)[i][j] = (unsigned char) ii ;
          
      } else {
          fprintf(stderr, "ERROR:  class label out of range %d\n", ii) ;
          fprintf(stderr, "ERROR:  i %d, j %d \n", i,j) ;
          exit(-1) ;
      }
  }

  /* write out the decimated image */ 
  
  if(! write_img(filename, &img_out, 't', "lzw:2") ) {
      fprintf(stderr, "ERROR: can not write out %s\n", filename) ;
      goto Bad ;
  }

  free_eisl_img( &img_in ) ; 
  free_eisl_img( &img_out ) ; 
  return(TRUE) ;
  
Bad :
  free_eisl_img( &img_in ) ;
  free_eisl_img( &img_out ) ; 
  return(FALSE) ;
  
}




static int Init_color_map(eisl_img *img) 
{
  int i, KK ;
  
  img->bitspersample = 8 ;
  KK = (1<<(img->bitspersample)) ;    /* 2^bitspersample */
  img->rmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  img->gmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  img->bmap = (unsigned short *) get_spc(KK, sizeof(unsigned short));
  if( (img->rmap == NULL) || (img->gmap == NULL) ||
      (img->bmap == NULL) ) {
    fprintf(stderr, "ERROR: Can not allocate color map, write_segment\n");
    return(FALSE) ;
  }

  /* Has to be initialied.  o.w. may cause confusion between */
  /* 8-bit and 16-bit color map,  randomly initialization */
  for(i=0; i<KK; i++) {  
      img->rmap[i] = rand() % KK ;
      img->gmap[i] = rand() % KK ;
      img->bmap[i] = rand() % KK ;
  }

  /* Since the number of classes are usually less than 10 */
  /* manually set the first 8 colors */
  img->rmap[0]=255;  img->gmap[0]=  0;  img->bmap[0]=  0;  
  img->rmap[1]=  0;  img->gmap[1]=255;  img->bmap[1]=  0;  
  img->rmap[2]=  0;  img->gmap[2]=  0;  img->bmap[2]=255;  
  img->rmap[3]=255;  img->gmap[3]=255;  img->bmap[3]=255;  
  img->rmap[4]=255;  img->gmap[4]=  0;  img->bmap[4]=255;  
  img->rmap[5]=  0;  img->gmap[5]=255;  img->bmap[5]=255;  
  img->rmap[6]=255;  img->gmap[6]=255;  img->bmap[6]=  0;  
  img->rmap[7]=  0;  img->gmap[7]=  0;  img->bmap[7]=  0;  
  
  return(TRUE) ;
}
