/*****************************************************************************/
/*   general image input, output utilities which are independent from*/
/*   specific image format*/
/*   */
/*   Hui Cheng,                Mar 3, 1998*/
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "eisl_img.h"
#include "allocate.h"
#include "general_u.h"


#define   FALSE  0
#define   TRUE   1


/*****************************************************************************/
int allocate_img (eisl_img *img, unsigned long ht, unsigned long wd, 
    char img_type)
{
    int i, k ;
    
    img->bitspersample = 8 ;
    img->img_type = img_type ;
    img->height = ht;
    img->width = wd;
    sprintf((img->compress_type), "none") ;
    
    switch ( img_type ) {
      case 'p':
          k = (1<<(img->bitspersample)) ;    /* 2^bitspersample */
          img->rmap = (unsigned short *) get_spc(k,sizeof(unsigned short));
          img->gmap = (unsigned short *) get_spc(k,sizeof(unsigned short));
          img->bmap = (unsigned short *) get_spc(k,sizeof(unsigned short));
          if( (img->rmap == NULL) || 
              (img->gmap == NULL) ||
              (img->bmap == NULL) ) {
            fprintf(stderr, "Can not allocate momery!  Allocate_img 3\n");
            return(FALSE) ;
          }
          /* Has to be initialied.  ow. may cause confusion between */
          /* 8-bit and 16-bit color map */
          for(i=0; i<k; i++) {    
              img->rmap[i] = 0 ;
              img->gmap[i] = 0 ;
              img->bmap[i] = 0 ;
          }
          /* NO break */
      case 'b':
      case 'g':
      case 'l':
          img->mono = (unsigned char **) get_img(wd, ht,sizeof(unsigned char));
          break ;
      case 'c':
      case 'y':
      case 'L':
          img->color = 
              (unsigned char ***) get_spc(3, sizeof(unsigned char **)) ;
          for(i=0; i<3; i++) {
              img->color[i] = (unsigned char **) get_img(wd, ht, 
                  sizeof(unsigned char)) ;
          }
          break ;
      default:
          fprintf(stderr, "ERROR: unknown eisl_img type %c \n", img_type);
          return(FALSE);
    }
    
    return(TRUE) ;
}


#ifdef _SUPPORT_PXM_

/*****************************************************************************/
static int 
write_pxm_img (char *fname, eisl_img *img)
{
  FILE *fp ;
  struct pxm_img  img_pxm ;
  int status ;

  img_pxm.width = img->width ;
  img_pxm.height = img->height ;
  img_pxm.max_gray = 255 ;

  switch( img->img_type ) {
     case 'b':
     case 'g':
         img_pxm.mono = img->mono ;
         img_pxm.pxm_type = img->img_type ;
         break ;
     case 'c':
         img_pxm.color = img->color ;
         img_pxm.pxm_type = 'p' ;           /* ppm */
         break ;
     default:
         fprintf(stderr, "ERROR: pxm does not support %c image type\n", 
             img->img_type ) ;
         return(FALSE) ;
  }
  
  fp = fopen(fname, "w");  if (fp == NULL) FileOpenErr(fname) ;
  status = write_pxm (fp, &img_pxm);
  if (status) {
	  fprintf(stderr,"ERROR: Cannot write out pxm file %s\n",fname);
	  return(FALSE) ;
  }
  fclose(fp) ;
  
  return(TRUE) ;
}


/*****************************************************************************/
static int 
read_pxm_img_header (char *fname, eisl_img *img)
{
  FILE *fp ;
  struct pxm_img  img_pxm ;
  int status ;

  fp = fopen(fname, "r");
  if(fp == NULL) FileOpenErr(fname) ;
  status = read_pxm_header(fp,&img_pxm) ;
  fclose(fp) ;
  if (status) {
      fprintf(stderr,"ERROR: Cannot read pxm header of file %s !!\n",fname);
      return(FALSE) ;
  }
  
  img->height = img_pxm.height ;
  img->width = img_pxm.width ;
  img->img_format = 'p' ;
  switch( img_pxm.pxm_type ) {
    case 'b':
        img->bitspersample = 1 ;
        img->img_type = 'b' ;
        break ;
    case 'g':
        img->bitspersample = 8 ;
        img->img_type = img_pxm.pxm_type ;
        break ;
    case 'p':
        img->bitspersample = 8 ;
        img->img_type = 'c' ;
        break ;
    default:
        fprintf(stderr, "ERROR: unkown pxm type %c\n", img_pxm.pxm_type ) ;
        return(FALSE) ;
  }
  
  return(TRUE) ;
}

/*****************************************************************************/
static int 
read_pxm_img (char *fname, eisl_img *img)
{
  FILE *fp ;
  struct pxm_img  img_pxm ;
  int status ;

  fp = fopen (fname, "r");
  if(fp == NULL) FileOpenErr(fname) ;
  status = read_pxm(fp,&img_pxm) ;
  fclose(fp) ;
  if (status) {
      fprintf(stderr,"ERROR: Cannot read pxm image from file %s !!\n",fname);
      return(FALSE) ;
  }
  
  img->height = img_pxm.height ;
  img->width = img_pxm.width ;
  img->img_format = 'p' ;
  switch( img_pxm.pxm_type ) {
    case 'b':
        img->bitspersample = 1 ;
        img->mono = img_pxm.mono ;
        img->img_type = 'b' ;
        break ;
    case 'g':
        img->bitspersample = 8 ;
        img->mono = img_pxm.mono ;
        img->img_type = img_pxm.pxm_type ;
        break ;
    case 'p':
        img->bitspersample = 8 ;
        img->color = img_pxm.color ;
        img->img_type = 'c' ;
        break ;
    default:
        fprintf(stderr, "ERROR: unkown pxm type %c\n", img_pxm.pxm_type ) ;
        return(FALSE) ;
  }
  
  return(TRUE) ;
}

#endif  /* _SUPPORT_PXM_ */





#ifdef _SUPPORT_TIFF_

/*****************************************************************************/
static int 
write_TIFF_img (char *fname, eisl_img *img, char *compression)
{
  TIFF_img  *img_tiff ;

  /* set up structure for output image */
  img_tiff = New_TIFF_Img("img_tiff") ;
  img_tiff->height = img->height ;
  img_tiff->width = img->width ;
  img_tiff->TIFF_type = img->img_type ;
  img_tiff->bitspersample = 8 ;
  sprintf(img_tiff->compress_type, compression) ;

/*  right now,  there is no need to re-allocate 

  if( ! Allocate_New_TIFF(img_tiff) ) {
      fprintf(stderr, "ERROR: Can not allocate img_tiff \n") ;
      return(FALSE) ;
  }
*/

  switch( img->img_type ) {
     case 'p':
         img_tiff->rmap = img->rmap ;
         img_tiff->gmap = img->gmap ;
         img_tiff->bmap = img->bmap ;
         img_tiff->mono = img->mono ;
         break ;
     case 'b':
     case 'g':
         img_tiff->mono = img->mono ;
         break ;
     case 'c':
         img_tiff->color = img->color ;
         break ;
     default:
         fprintf(stderr, "ERROR: right now does not support %c image type\n", 
             img->img_type );
         fprintf(stderr, "ERROR: you can modify  write_TIFF_img()\n") ; 
         return(FALSE) ;
  }
  
  if ( ! Write_TIFF_Img(fname, img_tiff, (char *) NULL) ) {
      fprintf ( stderr, "ERROR: error writting file %s\n", fname );
      return(FALSE) ;
  }

  /* added to fix memory leak.  JLH  18 Dec 02 */
  /* note this does NOT free img->rmap, etc. */
  /* this just frees the TIFF_img structure */
  free ( (void *)img_tiff );
  /* end of addition */
  
  return(TRUE) ;
}


/*****************************************************************************/
static int 
read_TIFF_img_header (char *fname, eisl_img *img)
{
  TIFF_img img_tiff ;
  TIFF* in;

  in = TIFFOpen(fname, "r");
  if (in == NULL) {
      fprintf(stderr, "ERROR: can not open %s.",fname) ;
      return(FALSE);
  }

  if(! Read_TIFF_Img_Header(in, &img_tiff) ) {
      fprintf(stderr, "ERROR: can not read tiff header %s\n", fname) ;

      /*added in attempt to fix memory leak.   17 Dec 02 JLH */
      (void) TIFFClose(in);
      /* end of addition */

      return(FALSE) ;
  }

	/*added in attempt to fix memory leak.   17 Dec 02 JLH */
	(void) TIFFClose(in);
	/* end of addition */

  img->height = img_tiff.height ;
  img->width = img_tiff.width ;
  img->bitspersample = img_tiff.bitspersample ;
  img->img_type = img_tiff.TIFF_type ;
  img->img_format = 't' ;
  
  return(TRUE) ;
}


/*****************************************************************************/
static int 
read_TIFF_img (char *fname, eisl_img *img)
{
  TIFF_img *img_tiff ;

  img_tiff = Read_TIFF_Img(fname, 0) ;
  if (img_tiff == NULL) {
      fprintf(stderr, "ERROR: can not read tiff image %s.",fname) ;
      return(FALSE);
  }

  img->height = img_tiff->height ;
  img->width = img_tiff->width ;
  img->bitspersample = img_tiff->bitspersample ;
  img->img_type = img_tiff->TIFF_type ;
  img->img_format = 't' ;
  
  switch( img_tiff->TIFF_type ) {
    case 'g':
    case 'b':
        img->mono = img_tiff->mono ;
        break ;
    case 'p':
        img->mono = img_tiff->mono ;
        img->rmap = img_tiff->rmap ;
        img->gmap = img_tiff->gmap ;
        img->bmap = img_tiff->bmap ;
        break ;
    case 'c':
        img->color = img_tiff->color ;
        break ;
    default:
        fprintf(stderr, "ERROR:   eisl_img does not support image type %c\n",
            img_tiff->TIFF_type) ;

        /*added in attempt to fix memory leak.   17 Dec 02 JLH */
        free( (void *)img_tiff );
        /* end of addition */

        return(FALSE) ;
  }

  /* added in attempt to fix memory leak.  17 Dec 02  JLH */
  free( (void *)img_tiff );
  /* end of addition */

  return(TRUE) ;
}

#endif  /* _SUPPORT_TIFF_ */




/*****************************************************************************/
int write_img (char *fname, eisl_img *img, char img_format, char *compression)
{
  switch( img_format ) {

#ifdef _SUPPORT_PXM_
    case 'p':         /* pxm format */
        if (! write_pxm_img (fname, img) )  goto Bad ;
        break ;
#endif 
  
#ifdef _SUPPORT_TIFF_
    case 't':         /* TIFF format */
        if (! write_TIFF_img (fname, img, compression) )  goto Bad ;
        break ;
#endif 
  
    default:
        fprintf(stderr, "ERROR: unkown or unsupported image format %c\n",
            img->img_format ) ;
        return( FALSE ) ;
  }

  return(TRUE) ;
  
Bad:
  fprintf(stderr, "ERROR: can not write %s in %c format\n", fname, img_format) ;
  return(FALSE) ;
}



/******************************************************************************/
/*   Determine the image format.  The rules using right now are*/
/*        (1)  if the first letter of the image file is 'M',  it is TIFF*/
/*        (2)  if the first letter of the image file is 'P',  it is PXM*/
/* */
/*   Hui Cheng,                      Mar 3, 1998 */
/******************************************************************************/
static int determine_image_format (char *img_format, char *fname) 
{
  FILE *fp ;
  char ch ;
  
  fp = fopen(fname, "r");  
  if (fp == NULL) FileOpenErr(fname) ;
  fscanf(fp, "%c", &ch) ;
  fclose(fp) ;
  
  switch( ch ) {
    case 'M':
printf("\tinput image is in tiff format\n") ;
        *img_format = 't' ;
        return(TRUE) ;
   case 'I':
printf("\tinput image is in tiff format\n") ;
        *img_format = 't' ;
        return(TRUE) ;
   case 'P':

printf("\tinput image is in pxm format\n") ;
        *img_format = 'p' ;
        return(TRUE) ;
    default:
        fprintf(stderr, "ERROR: cannot determine input image formate!") ;
        fprintf(stderr, " The first letter of the file is %c\n", ch) ;
        return(FALSE) ;
  }
  return(FALSE) ;
}



/******************************************************************************/
int read_img_header (char *fname, eisl_img *img)
{
  if( ! determine_image_format(&(img->img_format), fname) ) {
      fprintf(stderr, "ERROR: image formate can not be determined %s\n",
          fname) ;
      return(FALSE) ;
  }

  switch( img->img_format ) {

#ifdef _SUPPORT_PXM_
    case 'p':         /* pxm format */
        if (! read_pxm_img_header (fname, img) )  goto Bad ;
        break ;
#endif 
  
#ifdef _SUPPORT_TIFF_
    case 't':         /* TIFF format */
        if (! read_TIFF_img_header (fname, img) )  goto Bad ;
        break ;
#endif 
  
    default:
        fprintf(stderr, "ERROR: unkown or unsupported image format %c\n",
            img->img_format ) ;
        return( FALSE ) ;
  }

  return(TRUE) ;
  
Bad:
  fprintf(stderr, "ERROR: can not read image %s\n", fname) ;
  return(FALSE) ;

}



/******************************************************************************/
int read_img (char *fname, eisl_img *img)
{
  if( ! determine_image_format(&(img->img_format), fname) ) {
      fprintf(stderr, "ERROR: image formate can not be determined %s\n",
          fname) ;
      return(FALSE) ;
  }

  switch( img->img_format ) {

#ifdef _SUPPORT_PXM_
    case 'p':         /* pxm format */
        if (! read_pxm_img (fname, img) )  goto Bad ;
        break ;
#endif 
  
#ifdef _SUPPORT_TIFF_
    case 't':         /* TIFF format */
        if (! read_TIFF_img (fname, img) )  goto Bad ;
        break ;
#endif 
  
    default:
        fprintf(stderr, "ERROR: unkown or unsupported image format %c\n",
            img->img_format ) ;
        return( FALSE ) ;
  }

  return(TRUE) ;
  
Bad:
  fprintf(stderr, "ERROR: can not read the header of %s\n", fname) ;
  return(FALSE) ;
}



/******************************************************************************/
void free_eisl_img (eisl_img *img)
{
  int i;

  switch( img->img_type ) {
      case 'b' :
      case 'g' :    /* grayscale */
      case 'l' :
          if(img->mono != NULL)  free_img ( ( void ** ) ( img->mono ) );
          break ;
      case 'p' :     /* palette-color */
          if(img->mono != NULL) 
              tiff_free_img ( ( void ** ) ( img->mono ) );
          if(img->rmap != NULL) 
              free ( ( void * ) ( img->rmap ) );
          if(img->gmap != NULL) 
              free ( ( void * ) ( img->gmap ) );
          if(img->bmap != NULL) 
              free ( ( void * ) ( img->bmap ) );
          break ;
      case 'c' :    /* full color */
      case 'y' :
      case 'L' :
          if(img->color != NULL) {
               for ( i = 0; i < 3; i++ ) {
                   if(img->color[i]!= NULL) 
                       free_img ( ( void ** ) ( img->color[i] ) );
               }
               free( (void *) img->color ) ;
           }
           break ;
      default :
          fprintf(stderr, "ERROR: unkown img_type %d\n", img->img_type) ;
          exit(-1) ;
  }
}
