/*****************************************************************************/
/*   general image type which are independent from specific image format*/
/*   */
/*   Hui Cheng,                Mar 3, 1998 */
/*****************************************************************************/
#include <stdio.h>

#define _SUPPORT_TIFF_
#define _SUPPORT_PXM_


#ifdef _SUPPORT_TIFF_
#include "../../tiff_lib/tiff-v3.4/libtiff/Tiff_img.h"
#endif


#ifdef _SUPPORT_PXM_
#include "pxm.h"
#endif


#ifndef _EISL_IMG_
#define _EISL_IMG_

typedef struct {
  unsigned long         height;
  unsigned long         width;
  int                   bitspersample ;
  char                  img_type;       /* 'g' = grayscale;               */
                                        /* 'b' = bilevel image;           */
                                        /* 'p' = palette-color;           */
                                        /* 'c' = RGB full color           */
                                        /* 'L' = CIEL*a*b* full color     */
                                        /* 'l' = CIEL*a*b* monochrome     */
  unsigned char         **mono;         /* monochrome data, or indices    */
                                        /* of color-map; indexed as       */
                                        /* mono[row][col]                 */
  unsigned short        *rmap;          /* RGB color map                  */
  unsigned short        *gmap;
  unsigned short        *bmap;

  unsigned char         ***color;       /* full-color RGB data; indexed   */
                                        /* as color[plan][row][col],      */
                                        /* with planes being red,  green, */
                                        /* and blue, respectively         */
  char                  img_format;     /* 't' = tiff                     */
                                        /* 'p' = pxm                      */

  char                  compress_type[256] ;
  /*                                                                      */
  /*    lzw[:opts]      compress output with Lempel-Ziv & Welch encoding  */
  /*    zip[:opts]      compress output with deflate encoding             */
  /*    jpeg[:opts]compress output with JPEG encoding                     */
  /*    packbits        compress output with packbits encoding            */
  /*    g3[:opts]       compress output with CCITT Group 3 encoding       */
  /*    g4              compress output with CCITT Group 4 encoding       */
  /*    none    use no compression algorithm on output                    */
  /*                                                                      */
  /*    Group 3 options:                                                  */
  /*            1d      use default CCITT Group 3 1D-encoding             */
  /*            2d      use optional CCITT Group 3 2D-encoding            */
  /*            fill    byte-align EOL codes                              */
  /*            For example, "g3:2d:fill" to get G3-2D-encoded            */
  /*                    data with byte-aligned EOLs                       */
  /*                                                                      */
  /*    JPEG options:                                                     */
  /*            #       set compression quality level (0-100, default 75) */
  /*            r       output color image as RGB rather than YCbCr       */
  /*            string  file which contains the application supported     */
  /*                    quantization tables.                              */
  /*            For example, -c jpeg:r:50:MyQTables  to get JPEG-encoded  */
  /*                    RGB data with 50% comp. quality, using q_tables   */
  /*                    stored in MyQTables                               */
  /*            NB: The order of parameters are fixed as listed in the    */
  /*                    example.                                          */
  /*                                                                      */
  /*    LZW and deflate options:                                          */
  /*            #       set predictor value                               */
  /*            For example, -c lzw:2 to get LZW-encoded data with        */
  /*                    horizontal differencing                           */
  
}  eisl_img ;


int allocate_img (eisl_img *img, unsigned long ht, unsigned long wd, 
    char img_type);
int write_img (char *fname, eisl_img *img, char img_format, char *compression);
int read_img_header (char *fname, eisl_img *img);
int read_img (char *fname, eisl_img *img);
void free_eisl_img (eisl_img *img);


#endif  /* _EISL_IMG_ */
