/* Utilities for matrices manipulations
   	
   Jun 16, 1997,		Hui Cheng
*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "util.h"



void PrintMatrix(FILE *fd, double **a, int rl, int rh, int cl, int ch) 
{
  int i,j;
  

  fprintf(fd, "\n") ;
  for(i=rl; i<=rh; i++) {
  	for(j=cl; j<=ch; j++) fprintf(fd, "%lf  ",a[i][j]) ;
  	fprintf(fd, "\n") ;
  }
  fprintf(fd, "\n") ;
}




void PrintVector(FILE *fd, double *a, int nl, int nh) 
{
  int i ;
  
  for(i=nl; i<=nh; i++) fprintf(fd, "%lf  ", a[i]) ;
  fprintf(fd, "\n") ;
}



void ePrintMatrix(FILE *fd, double **a, int rl, int rh, int cl, int ch) 
{
  int i,j;
  

  for(i=rl; i<=rh; i++) {
  	for(j=cl; j<=ch; j++) fprintf(fd, "%6.4e  ",a[i][j]) ;
  	fprintf(fd, "\n") ;
  }
  fprintf(fd, "\n") ;
}




void ePrintVector(FILE *fd, double *a, int nl, int nh) 
{
  int i ;
  
  for(i=nl; i<=nh; i++) fprintf(fd, "%6.4e  ", a[i]) ;
  fprintf(fd, "\n") ;
}



void iPrintMatrix(FILE *fd, int **a, int rl, int rh, int cl, int ch) 
{
  int i,j;
  

  for(i=rl; i<=rh; i++) {
  	for(j=cl; j<=ch; j++) fprintf(fd, "%d  ",a[i][j]) ;
  	fprintf(fd, "\n") ;
  }
  fprintf(fd, "\n") ;
}




void iPrintVector(FILE *fd, int *a, int nl, int nh) 
{
  int i ;
  
  for(i=nl; i<=nh; i++) fprintf(fd, "%d  ", a[i]) ;
  fprintf(fd, "\n") ;
}


void fPrintMatrix(FILE *fd, float **a, int rl, int rh, int cl, int ch) 
{
  int i,j;
  

  for(i=rl; i<=rh; i++) {
  	for(j=cl; j<=ch; j++) fprintf(fd, "%f  ",a[i][j]) ;
  	fprintf(fd, "\n") ;
  }
  fprintf(fd, "\n") ;
}




void fPrintVector(FILE *fd, float *a, int nl, int nh) 
{
  int i ;
  
  for(i=nl; i<=nh; i++) fprintf(fd, "%f  ", a[i]) ;
  fprintf(fd, "\n") ;
}

void  MatrixMultiply( 
  double **In1, int rl1, int rh1, int cl1, int ch1, int T1,
  double **In2, int rl2, int rh2, int cl2, int ch2, int T2,
  double **Out
  /* if T1==1, then In1 needs to be transposed first */
)
{
  double s ;
  int i,j,k ;
  double **tmp ;	/* Tmp array to save the output */
 
/*printf("  MatrixMultiply In1\n") ;
  PrintMatrix(stdout, In1,rl1,rh1,cl1,ch1) ;
printf("In2\n") ;
  PrintMatrix(stdout, In2,rl2,rh2,cl2,ch2) ;
*/
 
  if(!T1 && !T2) {	/* no transpose */
  	
  	tmp = (double **) matrix(rl1,rh1, cl2,ch2) ;
  	
  	/* dimension test */
  	if((ch1-cl1)!=(rh2-rl2)) {
  		fprintf(stderr,"Dimension conflicts in MatrixMultiply\n") ;
  		exit(-1) ;
  	}
  	
  	for(i=rl1; i<=rh1; i++) 
  	for(j=cl2; j<=ch2; j++) {
  		s = 0.0 ;
  		for(k=0; k<=(ch1-cl1); k++) s += In1[i][k+cl1]*In2[k+rl2][j] ;
  		tmp[i][j] = s ;
  	}
  	
  	for(i=rl1; i<=rh1; i++) 
  	for(j=cl2; j<=ch2; j++) {
  		Out[i][j] = tmp[i][j] ;
  	}
  	
  	free_matrix(tmp, rl1,rh1, cl2,ch2) ;
  	
  	
  } else if(T1 && !T2) {	/* transpose In1 */
  	
  	tmp = (double **) matrix(cl1,ch1, cl2,ch2) ;
  	
  	/* dimension test */
  	if((rh1-rl1)!=(rh2-rl2)) {
  		fprintf(stderr,"Dimension conflicts in MatrixMultiply\n") ;
  		exit(-1) ;
  	}
  	
  	for(i=cl1; i<=ch1; i++) 
  	for(j=cl2; j<=ch2; j++) {
  		s = 0.0 ;
  		for(k=0; k<=(rh1-rl1); k++) s += In1[k+rl1][i]*In2[k+rl2][j] ;
  		tmp[i][j] = s ;
  	}

  	for(i=cl1; i<=ch1; i++) 
  	for(j=cl2; j<=ch2; j++) {
  		Out[i][j] = tmp[i][j] ;
  	}
  	
  	free_matrix(tmp, cl1,ch1,cl2,ch2) ;
  	
  	
  } else if(!T1 && T2) {	/* transpose In2 */
  
  	tmp = matrix(rl1,rh1,rl2,rh2) ;
  	
  	/* dimension test */
  	if((ch1-cl1)!=(ch2-cl2)) {
  		fprintf(stderr,"Dimension conflicts in MatrixMultiply\n") ;
  		exit(-1) ;
  	}
  	
  	for(i=rl1; i<=rh1; i++) 
  	for(j=rl2; j<=rh2; j++) {
  		s = 0.0 ;
  		for(k=0; k<=(ch1-cl1); k++) s += In1[i][k+cl1]*In2[j][k+cl2] ;
  		tmp[i][j] = s ;
  	}
  	
  	for(i=rl1; i<=rh1; i++) 
  	for(j=rl2; j<=rh2; j++) {
  		Out[i][j] = tmp[i][j] ;
  	}
  	
  	free_matrix(tmp, rl1,rh1,rl2,rh2) ;
  	
  } else if(T1 && T2) {	/* transpose In1 and In2 */

  	tmp = matrix(cl1,ch1,rl2,rh2) ;

  	/* dimension test */
  	if((rh1-rl1)!=(ch2-cl2)) {
  		fprintf(stderr,"Dimension conflicts in MatrixMultiply\n") ;
  		exit(-1) ;
  	}
  	
  	for(i=cl1; i<=ch1; i++) 
  	for(j=rl2; j<=rh2; j++) {
  		s = 0.0 ;
  		for(k=0; k<=(rh1-rl1); k++) s += In1[k+rl1][i]*In2[j][k+cl2] ;
  		tmp[i][j] = s ;
  	}
  	
  	for(i=cl1; i<=ch1; i++) 
  	for(j=rl2; j<=rh2; j++) {
  		Out[i][j] = tmp[i][j] ;
  	}
  	
  	free_matrix(tmp, cl1,ch1,rl2,rh2) ;
  }  
}






