#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "power.h"
static void *power_get_spc(int num, size_t size);



/* This is an example main program which can be used to exercise */
/* the subroutines in this file. To try them, simply delete the  */
/* comment marks and compile.                                    */

/* 
main()
{
    int samples = 32768;
    double max_val = 256;
    double exp_1, exp_2;
    struct power_LUT *power_LUT_1;
    struct power_LUT *power_LUT_2;

    double data[101];
    float output[101];
    int i,N=101;

    exp_1 = 1.2;
    exp_2 = 0.2;

    for(i=0; i<N; i++) {
      data[i] = 260.0*((double)i/N);
    }

    power_LUT_1 = build_power_LUT(exp_1, max_val, samples);
    power_LUT_2 = build_power_LUT(exp_2, max_val, samples);

    for(i=0; i<N; i++) output[i] = data[i];
    power(output, N, power_LUT_1 );

    for(i=0; i<N; i++) {
      printf("Exact value = %f;  Approximate value = %f\n",
        pow(data[i],exp_1), output[i]);
    }


    printf("\n\n\n\n\n");

    distroy_power_LUT(power_LUT_1); 


    for(i=0; i<N; i++) output[i] = data[i];
    power(output, N, power_LUT_2 );

    for(i=0; i<N; i++) {
      printf("Exact value = %f;  Approximate value = %f\n",
        pow((double)data[i],exp_2), output[i]);
    }


}

*/
    
    



/* This subroutine builds a look-up-table structure which can be used   */
/* with the power() function to perform fast (approximate) floating     */
/* point exponential functions. 					*/

struct power_LUT 
*build_power_LUT(double exponent, double max_val, double samples) 
{
    int    i;
    double period;
    struct power_LUT *power_LUT;


    /* allocate memory for power_LUT structure */
    power_LUT = (struct power_LUT *)power_get_spc(1,sizeof(struct power_LUT));

    /* set LUT length */
    power_LUT->samples = samples;

    /* compute sampling period and rate of LUT */
    period = max_val/(samples-1);
    power_LUT->rate = 1.0/period;

    /* allocate value LUT and fill in entries */
    power_LUT->value_lut = (float *) power_get_spc(samples,sizeof(float));
    for (i=0; i<samples; i++) {
      power_LUT->value_lut[i] = pow(i*period,exponent);
    }

    /* allocate difference LUT and fill in entries */
    power_LUT->diff_lut = (float *) power_get_spc(samples,sizeof(float));
    for (i=0; i<samples-1; i++) {
      power_LUT->diff_lut[i] 
        = power_LUT->value_lut[i+1] - power_LUT->value_lut[i];
    }

    return(power_LUT);
}





/* This subroutine computes a look-up-table approximation to the pow() */
/* for the n elements stored in the array data_array. It passes the    */
/* computed values back in the same array.                             */

void power(float *data_array, int n, struct power_LUT *power_LUT)
{
    static float  fraction;
    static int    i, index;

    /* Compute the approximation to the power function for each element	*/
    /* of realarray by linearly interpolating between the exact values	*/
    /* of the power functon in the LUT just below and just above the    */
    /* element.								*/

    for (i=0; i<n; i++) {
        fraction = power_LUT->rate*data_array[i];
        index = fraction;
        if(index>=power_LUT->samples) {
          printf("Error: Overflow in power function\n");
          data_array[i] = power_LUT->value_lut[power_LUT->samples-1];
        }
        else {
          data_array[i] = power_LUT->value_lut[index] 
                          + (fraction-index)*power_LUT->diff_lut[index];
        }
    }
}


void distroy_power_LUT(struct power_LUT *power_LUT) 
{
    free( (void *)power_LUT->value_lut);
    free( (void *)power_LUT->diff_lut);
    free( (void *)power_LUT);
}



static void *power_get_spc(int num, size_t size)
{
	void *pt;

	if( (pt=calloc((size_t)num,size)) == NULL ) {
		fprintf(stderr, "==> calloc() error\n");
		exit(-1);
		}
	return(pt);
}


