#include <math.h>

#include "allocate.h"

#define SAMPLES	65536	/* Number of uniformly spaced samples in LUT */
#define MAX_VAL	256.0	/* Maximum value to passed to power function */


/* Power approximates an exponentiation operation on an array of real	*/
/* numbers, replacing the entries with the results in place.  Each	*/
/* element of realarray is approximately raised to exponenet with the	*/
/* use of a Look-up-table and linear interpolation. It is assumed that	*/
/* each element of realarray is nonnegative and less MAX_VAL.		*/

void power(float *realarray, int n, float exponent) {

static int first = 1;			/* initialization flag		*/
static float rate;			/* distance between samples	*/
static float *value_lut, *diff_lut;	/* look up talbes		*/
static float  constant, fraction;
static int    i, index;

    /* The first time power is called, two LUTs are generated: one for	*/
    /* the values of the power functon, and one for the difference in 	*/
    /* successive values, used for the linear interpolation.          	*/

    if (first) {
        first = 0;
        rate = (float) MAX_VAL / SAMPLES;
        constant = 1.0/rate;
        value_lut = (float *) get_spc(SAMPLES,sizeof(float));
        for (i=0; i<SAMPLES; i++) {
            value_lut[i] = pow(i*rate,exponent);
        }
        diff_lut = (float *) get_spc(SAMPLES,sizeof(float));
        for (i=0; i<SAMPLES-1; i++) {
            diff_lut[i] = value_lut[i+1] - value_lut[i];
        }
        diff_lut[SAMPLES-1] = pow(MAX_VAL,exponent) - value_lut[SAMPLES-1];
    }

    /* Compute the approximation to the power function for each element	*/
    /* of realarray by linearly interpolating between the exact values	*/
    /* of the power functon in the LUT just below and just above the    */
    /* element.								*/

    for (i=0; i<n; i++) {
        fraction = constant*realarray[i];
        index = fraction;
        if(index>=SAMPLES) {
          printf("Error: Overflow in power function\n");
          realarray[i] = value_lut[SAMPLES-1];
        }
        else {
          realarray[i] = value_lut[index] + (fraction - index)*diff_lut[index];
        }
    }
}
